

/*
 * This class implements the algorithm developed by
 * Briand, Labiche and Wang 
 * for generating a reasonable test order 
 * for class integration testing. The classes are input
 * to the program in the form of a graph where the nodes represent the classes
 * to be tested and the edges represent the relationships among the classes. 
 * 
 * Their algorithm is very similar to that of Yves Le Traon et ab. The differences are:
 * 1. La Traon et al. do not consider edge types, Briand et al do consider.
 * 2. La Traon et al. use the notion of frond edges whereas Briand et al 
 * use Tai-Daniels approach to compute edge weights but only for association edges and 
 * using multiplication of incoming and outgoing edges rather than sum as in Tai-Daniels algorithm.
 * 
 * Thus, in a sense, Briand et al. algorithm is a simple and effective
 * mix of algorithms proposed by Le Traon et al. and Tai-Daniels.
 * 
 * 
 * Reference:
 * 
 *An Investigation of Graph-Based Class Integration Test Order Strategies 
 * Lionel C. Briand, Yvan Labiche, and Yihong Wang
 * IEEE TRANSACTIONS ON SOFTWARE ENGINEERING, VOb. 29, NO. 7, JULY 2003
 * 
 * This class coding 
 * Original: July 7, 2011 (Started with Tai Daniels algorithm)
 * Update: Aug 14, 2011 (Started Briand et al algorithm)
 * Latest: August 20, 2011
 * 
 * Notes:
 * 
 * 1. The first complete version of this tool passed tests on July 27, 2011.
 * 2. The second version began on July 28, 2011. It includes abstract classes, static as well
 * as dynamic dependencies, and addition information in test order display.
 * 3. This version implements the Le Traon et al algorithm that
 *  is quite different from that of Tai-Daniels. It first passed all three tests 
 * on August 13, 2011. (Test 1: Tai-Daniels example, Test 2: Yves Le Traoin et ab. exampple.
 * Test 3: Briand et ab. example. 
 * 
 * Bugs found and fixed:
 * 
 * Aug 5, 2011, and many many many  more fixed later.
 * 
 * Error: The root of an SCC with only two nodes was not being marked inside tarjan()
 * as a frondRoot node. This is because tarjan() was marking a node as a root only for 
 * SCCs that have cycles, i.e., that have more than 1 node. But aan SCC with only two
 * nodes goes into torjan(), during a recursive call, AFTER the frond edge has been removed.
 * Thus tarjan() find no cycles in this Scc as the cycle is removed by the removal
 * of the frond edge.
 * 
 * Correction: An SCC with only two nodes is processed right after its frond edge is detected and removed.
 * The correction is inside setFrond() method.
 * 
 * Aug 12, 2011: Another error found when tried the implementation on the Tai/Daniel's example. In this
 * case one of the SCC's before the recursive call has more than 1 node but has no
 * cycles. It was being considered as one with cycles. But when tarjan() is applied again on this scc,
 * it splits into 4 singleton scc's (each with exactly one node). Because there is no cycle in any of these four nodes, 
 * no stub was crerated. Thus, for node SS, there was no stib created. This bug was fizxed by updating
 * the logic for finding cycle free sccs.
 * 
 * 
 * Author: Aditya Mathur
 * 
 */


/*
 * Abrreviations used:
 * 
 * SCC: Strongly connected component. scc: same but used as the name of an object.
 * 
 * Note: The first correct implementation of YLTet ab. algoruthm was available at 3:21pm on August 11.
 * It gave the correct answer to the simple example in YLT et ab. paper.
 */


import java.util.*;
import java.io.*;

public class BriandAlgorithm {
  
  // Object Relation Diagram as a graph. Attributes of nodes in this
  // graph are modified as the Tai-Daniels algorithm progresses.
  Graph g; 
  
  int maxMajorLevel=0; // Maximum major level number assigned in the graph.
  int [] maxMinorLevel; // Maximum minor level number assigned in the graph to each leveb.
  int majorMinor; // =0 implies major, =1 implies minor. Used in deciding whether to assign major or minor leveb.
  String deletedEdgeLabel="eA"+Edge.deleteString; // Label of an association edge that has been removed from g.
  ArrayList<ArrayList<Node>> testSequence; // Sequence in which classes are to be tested.
  ArrayList<Stub> stubs=new ArrayList<Stub>(); // Stubs needed.
  ArrayList<Retest> retestNodes; // Nodes to be retested.
  ArrayList<Node> testOrder=new ArrayList<Node> (); // Class tst order is recorded in this list.
  
  CreateGraph cg; // This object creates a graph from ndoes and edges input from a file.
  
  //Next flag indicates whether or not dynamicstatic dependencies are to be considered.
  // Set it to false to include dynamic dependencies.
  boolean staticOnly=true; 
  boolean cycleExists=true;
  
  public static boolean recursiveCall; // Used in findCycles() to start assigning stubs only after the first recursive calb.
  public static int counter=0; // To assign to nodes when finding an during recursive call to tarjan().
  ArrayList<ArrayList<Node>> scc; // Holds SCC's found after the fiorst call to Tarjan's algorithm.
  ArrayList<ArrayList<Node>> originalSccs;
  ArrayList<Scc> sccInitial;
  
  public int deleteSequence=0; // Keeps track of serial of edges deleted.
  Stack <ArrayList<ArrayList<Node>>> sccStack;
  
  
  // The constructor creates a graph g from an input text file.
  
  public BriandAlgorithm(String fileName)throws IOException{
    // TD algorithm Step 1: Read a graph from a file.
    System.out.println("Input graph from file: "+fileName);
    cg=new CreateGraph();
     cg.createGraph(fileName, staticOnly);
   // cg.createGraph("", staticOnly);
    sccStack=new Stack<ArrayList<ArrayList<Node>>>();
    
    System.out.println("--Graph successfully input.");
    // cg.createGraph("sccTest1NoEdges.txt");
    // cg.createGraph("sccTest2SymmetricTwoComponents.txt");
    
    //cg.createGraph("sccTest3ThreeComponents.txt");
    // cg.createGraph("sccTest4OneNodeDegenerate.txt");
    //cg.createGraph("sccTest5TwoNodesDegenerate.txt");
    g=cg.g; // Save the graph locally.
    
  }// End of BrianAlgorithm constructor
  
  
  
  // The following recursive method detects 
  // strongly connected components in graph g,
 // Each element of scc is an SCC (a graph).
  
  public void findCycles(ArrayList<ArrayList<Node>> scc){  
    
    Utilities.debug( Utilities.weightDebug, "\n# of SCCs to be processed: "+scc.size());
    
    // Check if there is any SCC in the list (scc) with cycles.
    cycleExists=doesACycleExist(scc);
    if(!cycleExists){
      if(Utilities.sccDebug){
        Utilities.displayAllScc(scc);
      }
    //  processSingleton(scc); // There might be a stub in one of the SCCs.
      return;
    }
    
    // Yes, there is at least one non-trovial SCC (i.e., with cycle). 
    // Break cycles in each non-trivial SCC.
    Iterator <ArrayList<Node>> itr=scc.iterator();
    int nextSccToBeProcessed=0;
    
    while(itr.hasNext()){
      ArrayList<Node> nextScc=itr.next();// Extract next SCC from the list (scc).
      nextSccToBeProcessed++; // Update its counter (for debugging)
      ArrayList<ArrayList<Node>> newSccList=new  ArrayList<ArrayList<Node>>();
      
      if(Utilities.weightDebug){
        newSccList.add(nextScc); // This is to ensure parameter compliance.
        System.out.println("\nNext SCC to be processed:"+nextSccToBeProcessed);
      } // End of debug check.
      
      if(nextScc.size()>1){// It can have cycles only if the nextScc has more than one node. 
        Graph newGraph=new Graph(nextScc);// Create a graph out of it (for type compliance).
        
        // Create an object to compute SCCs in newGraph.
        // Apply Tarjan' algorithm to to newGraph and get SCCs (frond edges are marked).
        
        SccTarjan sccAlg=new SccTarjan(newGraph); // This is a recursive call to
        scc=sccAlg.tarjan(); //  tarjan().
        sccStack.push(scc);
        
        if(Utilities.sccDebug){
          System.out.println("\nSCCs after next call to Tarjan's algorithm:");
          Utilities.displayAllScc(scc);
        }
        
        // Process all SCCs. 
        processSccs(scc); 
      
        findCycles(scc);// Recurse on the SCCs created.
      }// End of singleton scc check
      Utilities.debug( Utilities.weightDebug, "\n  SCC "+nextSccToBeProcessed + " processed:" );
    }// End of node ieration loop.   
  }// End of findCycles
  
  
  // The next method checks if a component of a graph  has any cycles in it.
  // The component (scc)is input as a list of nodes.
  
  
  private boolean doesACycleExist(ArrayList<ArrayList<Node>> scc){
    
    // A cycle exists if there is any element in scc with more than 1 node
    Iterator<ArrayList<Node>> itrScc=scc.iterator();
    while (itrScc.hasNext()){
      ArrayList<Node> anScc=itrScc.next();
      int anSccSize=anScc.size();
      if(anSccSize>1){
        Utilities.debug(Utilities.cycleCheckingDebug, 
                        "\n     There may be an SCC with a cycle. ");
        return true; 
      }
    }
    Utilities.debug(Utilities.cycleCheckingDebug, "\n     There is no cycle.");
    return false;
  }// End of cycles()
  
  // Process a list of SCCs at least one of which contains exactly one node.
  // One of these must be a root and is marked as a stub. If all SCCs have exactly one 
  // node then one of these must be a root and should be marked as a stub.
  private void processSingleton(ArrayList<ArrayList<Node>> scc){
    Utilities.debug(Utilities.weightDebug, "\nProcess SCCs with no cycles.");
    Node singleton;
    Node makeThisStub=null;
    Iterator<ArrayList<Node>> itr=scc.iterator();
    boolean allSizeOne=true; // Assume that all have exactly one node.
    while(itr.hasNext()){
      ArrayList<Node> anScc=itr.next();
      singleton=anScc.get(0);
      if(singleton.isRoot()){
        makeThisStub=singleton;
      }
      if(anScc.size()>1){
        allSizeOne=false;
        break; // Get out of this loop.
      }// End size check.
    }// End scc iteration loop.
    
    if(allSizeOne){
      if(makeThisStub!=null){
        makeThisStub.markStub();
        Utilities.debug(Utilities.weightDebug, "Node marked stub/root: "+ 
                        makeThisStub.getLabel().substring(1));
      }else{
        System.out.println("** Program error (processSingleton/BrianAlgorithm). "+ 
                           "All SCCs singleton but no root.");
      }
    }
  }// End of process singleton.
  
  // Process each SCC recursively.
  
  private void processSccs(ArrayList<ArrayList<Node>> scc){
    ArrayList<Edge> edges;
    
    Iterator<ArrayList<Node>> itr=scc.iterator(); // Set SCC iterator
    Node selected=null;
    int index=1;
    while(itr.hasNext()){
      ArrayList<Node> anScc=itr.next(); // Get next SCC.
      if(anScc.size()>1){   
        if(Utilities.weightDebug){
          System.out.println("Next (sub) Scc to be processed #: "+index);
          Utilities.displayAnScc(anScc);
        }
        
        Node currentRoot=getRootInInitialScc(anScc);
        
        // Update node weights, select a node, and delete the selected edge.
        edges=computeEdgeWeight(anScc); // Get association edges with weights
        if(edges.size()==0){
          System.out.println("This SCC has no association edges.");
          System.exit(0); // This needs to be handled differently for robustness. 1/28/2012
        }
        Edge heaviestEdge=findHeaviestEdge(edges); // Find the edge with max weight.
        heaviestEdge.delete(++deleteSequence); // Delete it.
        Node destination=heaviestEdge.getDestination(); // Mark  edge destination as stub
        Utilities.debug(Utilities.weightDebug, " ** Edge deleted: "+ 
                        Utilities.stringifyEdge(heaviestEdge, true));
        
        Node sourceNode=heaviestEdge.getSource();
        
        currentRoot.unMarkRoot(); // Unmark current root and
        destination.markStub();// mark the stub and 
        destination.markRoot();// new root.
        resetSccRoot(destination); // Reset current root of the original Scc.
        resetSccLink(currentRoot, destination); // Reset link to this SCC if need be.
        if(Utilities.weightDebug){
          Utilities.displayEdges(edges, true, destination);
          System.out.print("Node marked stub/root: "+destination.getLabel().substring(1));
          System.out.println(". Node unmarked root: "+currentRoot.getLabel().substring(1));
        }// End of debug block
      } // End of SCC size check.
      index++;
    }// End of SCC iterator loop.
  }// End of processSccs()
  
  
  
  // Update weights for all association edges in component c. 
  // These edges must originate and end in the scc.
  
  private ArrayList<Edge> computeEdgeWeight(ArrayList<Node> scc){  
    // First get all incoming edges that have their source within c.
    ArrayList<Edge> edges=getAssociationEdges(scc);
    Iterator<Edge> itr=edges.iterator();
    // Now iterate over all edges computing their weights.
    while(itr.hasNext()){
      Edge e=itr.next(); // Get next edge
      
      Node source=e.getSource();// Get its source and
      Node dest=e.getDestination(); // destination nodes.
      int sourceIn=source.getIncomingEdgeCount(scc); // Get incoming edge count
      int destOut=dest.getOutgoingEdgeCount(scc); // Get outgoing edge count
      int edgeWeight=sourceIn*destOut; // Multiply these to get the weight
      EdgeWeight ew=new EdgeWeight(edgeWeight, sourceIn, destOut); // Set the weight
      e.setWeight(ew);
    }// End of edge iterator.
    
    return edges;
  }// End of computeEdgeWeight()
  
  
  
  
  private Node getRootInInitialScc(ArrayList<Node> scc){
    Iterator<Node> itr=scc.iterator();
    while(itr.hasNext()){
      Node n=itr.next();
      if(n.isRoot())
        return n;
    }// End of node iteration.
    return null; // Root not found; program error.  
  }// End of getRootsInInitialScc()
  
  
  
  // Reset the current root node in the initial Scc.
  private void resetSccRoot(Node n){
    int sccIndex=n.getSccIndex(); // Get the index of the initial Scc to which this node belongs
    sccInitial.get(sccIndex-1).setCurrentRoot(n); // Reset the current root of the Scc.
    
  }// End of resetSccRoot()
  
  
  // Edges coming from the master root of the graph into
  // the intial Sccs must now point to the current root of the Scc.
  // There might also be Scc's that were accessible from another Scc but not the 
  // master root. Links to such Scc's might also need to be reset.
  
  
  
  // Edges coming from outside the current SCC into a node within the SCC into
  // a node must now point to the current root of the SCC. But this change need
  // to be made carefully so that thereremains no orphan node. 
  
  private void resetSccLink(Node oldRoot, Node newRoot){
    
    int oldRootHost=oldRoot.getSccIndex(); // Host Scc of the old root (or the root about to be changed)
    
    // If a link from nother SCC already points to the newRoot
    // then no need to do any resetting.
    ArrayList<Edge> edgesIntoNewRoot=newRoot.getIncomingSccEdges(); // Get its incoming edges.
    Iterator <Edge> itrEdge=edgesIntoNewRoot.iterator(); // Set up iterator.
    
    while(itrEdge.hasNext()){
      Edge e=itrEdge.next(); // Get next incoming edge.
      Node source=e.getSccSource(); 
      int sourceNodeHost=source.getSccIndex();// Get the host SCC of the node from where the edge is coming in.
      if(sourceNodeHost!=oldRootHost){ // If the two hosts are different then the link is to be reset.
        Node sccDestination=e.getSccDestination();
        if(sccDestination.equals(newRoot)){
          Utilities.debug(Utilities.resetEdgesDebug, 
                          "Scc link from node "+ source.getLabel().substring(1)+
                          " not reset as it already points to the new root: "+
                          newRoot.getLabel().substring(1));
          return;
        }
      }
    }// End of edge iteration loop.
    
    Utilities.debug(Utilities.resetEdgesDebug, "No outside link to "+newRoot.getLabel().substring(1));
    
    // There is no link from another SCC into newRoot. 
    // So reset the link to the newRoot.
    ArrayList<Edge> edgesIntoOldRoot=oldRoot.getIncomingSccEdges(); // Get edges coming into oldRoot.
    if(Utilities.resetEdgesDebug){
      System.out.println(edgesIntoOldRoot.size()+" Scc edges into node "+
                         oldRoot.getLabel().substring(1));
    }
    itrEdge=edgesIntoOldRoot.iterator(); // Set up iterator.
    while(itrEdge.hasNext()){
      Edge e=itrEdge.next(); // Get next incoming edge 
      Node source=e.getSccSource(); 
      int sourceNodeHost=source.getSccIndex();// Get the host SCC of the node from where the edge is coming in.
      if(sourceNodeHost!=oldRootHost){ // If the two hosts are different then the link is to be reset.
        e.setSccDestination(newRoot);
        e.setSccSource(source);
        newRoot.addIncomingSccEdge(e);
        Utilities.debug(Utilities.resetEdgesDebug, "   ** Scc edge added "+
                        Utilities.stringifyEdge(e,false));
        Utilities.debug(Utilities.resetEdgesDebug, "Link from node "+source.getLabel().substring(1)+
                        " reset to node "+newRoot.getLabel().substring(1)); 
        return;
      }
    }// End of edge loop.
    Utilities.debug(Utilities.resetEdgesDebug, "Outside link not reset to  "+
                    newRoot.getLabel().substring(1));
  }// End resetSccLink()
  
  
  // Extract all incoming  association edges emanating from within the input component scc.
  
  public ArrayList<Edge>  getAssociationEdges(ArrayList<Node> scc){ 
    ArrayList<Edge> returnEdgeList=new ArrayList<Edge>();
    Iterator<Node> itr=scc.iterator();
    while(itr.hasNext()){
      Node n=itr.next();
      ArrayList<Edge> edges=n.getIncomingAssociationEdges(scc);
      returnEdgeList.addAll(edges);
    }// End of node iteration loop.
    return(returnEdgeList); // This might be an empty list if no edge exists that satisfies the condition
  }// End of getIncomingEdges()
  
  // Find edge with the highest weight. 
  private Edge findHeaviestEdge(ArrayList<Edge> e){
    int listSize=e.size();
    Edge heaviest=e.get(0);
    int edgeWeight=heaviest.getWeight();
    for (int i=1; i<listSize; i++){
      Edge next=e.get(i);
      int w=next.getWeight();
      if(w>edgeWeight){
        edgeWeight=w;
        heaviest=next;
      }
    }
    return heaviest;
  }//End sortEdgesByWeight()
  
  // Program to test the implementation of TD algorithm.
  
  public static void main(String[] args) throws IOException{
    
    Utilities.TaiDaniels=false; // This is Yves Le Traon et ab. and not TaiDaniels algorithm.
    Utilities.LeTraonAlgorithm=false;
    Utilities.BriandAlgorithm=true; // This is Briand et al algorithm.
    //BriandAlgorithm b=new BriandAlgorithm("../TestGraphs/TaiDanielsGraph.txt");
    // BriandAlgorithm b=new BriandAlgorithm("../TestGraphs/LeTraonExample.txt");
    //  BriandAlgorithm b=new BriandAlgorithm("../TestGraphs/LeTraonExampleV1.txt");
    
    // BrianAlgorithm t=new BrianAlgorithm("../TestGraphs/LabichExample2000.txt");
    // BriandAlgorithm b=new BriandAlgorithm("../TestGraphs/BriandExample2003.txt");
    // BriandAlgorithm b=new BriandAlgorithm("../TestGraphs/ThreeNodesNoAssociation.txt");
    // BriandAlgorithm b=new BriandAlgorithm("../TestGraphs/ThreeNodesOnlyAssociation.txt");
    
    // BriandAlgorithm b=new BriandAlgorithm("../TestGraphs/BriandExample2003OneEdgeRemoved.txt");
    // BriandAlgorithm b=new BriandAlgorithm("../TestGraphs/BriandExample2003TwoEdgesRemoved.txt");
    //  BriandAlgorithm b=new BriandAlgorithm("../TestGraphs/BriandExample2003ThreeEdgesRemoved.txt");
    // BriandAlgorithm b=new BriandAlgorithm("../TestGraphs/SequentialLoops.txt");
    // BriandAlgorithm b=new BriandAlgorithm("../TestGraphs/ForwardBackwardLoops.txt");
    // BriandAlgorithm b=new BriandAlgorithm("../TestGraphs/MinimalRetestExample.txt");
    //   BriandAlgorithm b=new BriandAlgorithm("../TestGraphs/MinimalRetestExample2.txt");
    //BriandAlgorithm b=new BriandAlgorithm("../TestGraphs/MinimalRetestExample3.txt");
    //  BriandAlgorithm b=new BriandAlgorithm("../TestGraphs/MinimalRetestExample4.txt");
    
    //BriandAlgorithm b=new BriandAlgorithm("../TestGraphs/MinimalRetestExample5.txt");
    // BriandAlgorithm b=new BriandAlgorithm("../TestGraphs/MinimalRetestExample6.txt");
    //BriandAlgorithm b=new BriandAlgorithm("../TestGraphs/MinimalRetestExample7.txt");
   // BriandAlgorithm b=new BriandAlgorithm("../TestGraphs/BriandExample2003OneEdgeOneNodeRemoved.txt");
    //  BriandAlgorithm b=new  BriandAlgorithm("../TestGraphs/BriandExample2003OneEdgeTwoNodesRemoved.txt");
    
     BriandAlgorithm b=new  BriandAlgorithm("");
    // YLT algorithm. Find cycles and break them.
    // First call to Tarjan's algorithm. [Step 1 of Yves Le Traon et ab. algorithm.]
    if(Utilities.sccDebug){
      System.out.println("\nFind strongly connected components.");
    }
    
    double startTime=System.currentTimeMillis();
    b.recursiveCall=false;
    SccTarjan sccAlg=new SccTarjan(b.g); // Create object to compute SCCs in graph g.
    ArrayList<ArrayList<Node>> scc=sccAlg.tarjan(); // Apply Tarjan' algorithm to get SCCs.
    b.sccStack.push(scc);
    // Save the initial SCC for future reference.
    b.sccInitial=sccAlg.getInitialScc();
    
    
    if(Utilities.sccDebug){
      System.out.println("\nSCCs after first call to Tarjan's algorithm:");
      Utilities.displayScc(b.sccInitial);
    }
    // b.updateAllSccWeights(scc);
    b.processSccs(scc);
    
    // Utilities.displayAllScc(scc);
    
    
    // Now apply Tarjan's algorithm recursively on each noin-trivial Scc. 
    // [Step 2 of YTL et ab. algorithm.]
    
      if(Utilities.sccDebug){
        System.out.println(" ------  Start: Recursive call to find additional Sccs.");
      }
    b.recursiveCall=true;
    b.findCycles(scc);
    
   if(Utilities.sccDebug){
     System.out.println("-----  End: Recursive call.\n");
   }
    
    
    if(Utilities.sccDebug)
      Utilities.displayScc(b.sccInitial);
    
    
    if(Utilities.edgeDeletionDebug){
      System.out.println("\nDeleted edges (in order)");
      Utilities.displayDeletedEdges(b.g);
    }
    
    System.out.println("\nUpdated Graph:");
    if(Utilities.graphDisplayDebug)
      Utilities.displayGraph(b.g, true);
    // Final step to generate test sequence for classes.
    
    b.testOrder=Utilities.findTestOrder(b.g);
    if(Utilities.sccDebug){
      Utilities.displaySccStack(b.sccStack);
    }
    
     double endTime=System.currentTimeMillis();
    
    System.out.println("\n Test order:");
    Utilities.displayTestOrder(b.testOrder);
    
    double timeTaken=(endTime-startTime)/(1000*60);
    
    System.out.printf("\nEnd of Briand et al. algorithm. Total time: %.2f %s\n",timeTaken," minutes.");
    
  }// End of main()
}// End of class BriandAlgorithm.

