

/*
 * Class CreateGraph is used in the algorithm for generating
 * sequence of classes (nodes in a digraph) in integration testing
 * using Tai-Daniels algorithm. It reads graph 
 * data from a file and creates the Graph class. 
 * 
 * Format of the file containing graph nodes/edges follows.
 * 
 * The file containing the graph must be a plain text file.
 * Such a file can be created using, for example, 
 * Microsoft Word and saving it in plain text format.
 * 
 * Data for each node or edge occupies one line.
 * Lines are separated by CRLF.
 * Each node contains the following information: node label
 * There is exactly one line for each nodel label.
 *
 * Node-label
 * 
 * The line for each node is followed by as many lines 
 * as there are edges going out of that node.
 * 
 * edge-label destination-node-label 
 * 
 * An & indicates the end of edges for a node. It
 * is placed by iteself on a line. It indeicates the end
 * of all edges for a node. A node label by iteself on a line 
 * and followed on the next line by an & indicates that this node has no edges.
 * 
 * Label of a node must be prefixed by an "n". For example
 * label nB indicates that B is a node. Similarly an edge
 * label must be prefixed by an "e". For
 * example, label eF indicates an edge labelled F. A node can be labelled
 * as Abstract by using the tage tA following the node name.
 * 
 * Example input:
 * nBL
 * eA nB
 * eA nY
 * &
 * B
 * eA nSS
 * eF nE
 * eF nBS
 * &
 * nX
 * tA
 * eI nY
 * &
 * nY
 * &
 * 
 * The above (hypothetical) data is for a graph with  nodes labeled BL, B, SS, E,  BS, and X.
 * Node BL has one outgoing edge, node B has three outgoing edges and
 * the remaining nodes have no outgoing edges. Node X is abstract.
 * 
 * The obvious: Each node must have a label and each edge must have a label.
 * 
 * 
 * July 7, 2011.
 * Latest update: Aug 11, 2011
 * 
 * January 28, 2012. Added randomGraph() method to generate a random ORD.
 * Aditya Mathur
 * */

import java.util.*;
import java.io.*;

public class CreateGraph{
    
    Graph g;
    private ArrayList <Node> nodeList=new ArrayList<Node>(); // For internal use.
    int nodeCount=0;
    int edgeCount=0;
    int nextLine; // Used for displaying error in input file.
    boolean staticDynamic=true; // true implies dependency is static, otherwise this is false.
    char nodePrefix='n';  // Prefix for node label.
    char edgePrefix='e';  // Prefix for edge  label.
    char nodeTypePrefix='t';  // Prefix for node type.
    boolean inheritance=true; // Indicates weather or not inheritance edges are to be generated.
    
    // Probabilities associated with edges.
    double edgeProbability; // Probability that two nodes are connected by an edge.
    final static double inheritanceProbability=0.0;
    final static double aggregationProbability=0.0;
    final static double associationProbability=1.0;
   
// Graph metrics for randmly generated graph.
     int nodesCreated=0;
     int edgesCreated=0;
     int inheritanceEdges=0;
     int aggregationEdges=0;
     int associationEdges=0;
    
    // If filename is an empty string then a a graph is generated randomly.
    public void createGraph(String filename, boolean staticOnly) {
      
      if(filename.equals("")){
        generateRandomGraph();
        
        return;
      }
        
        // Set node, edge and type prefixes.
      
        Scanner in=null;
        try{
            FileReader f=new FileReader(filename);
            in= new Scanner(f);
        }catch(IOException e){
            System.out.println("File: "+filename+" not found.");
            System.exit(0);
        }// End of try-catch for4 file open 
        g=new Graph(); // Create an empty graph.
        // Read node and edge data.
        nextLine=0; 
        while(in.hasNext()){
            String sourceNodeLabel=in.next(); // Get label of a possibly new node. 
            nextLine++; // Increment line number of error display 
            checkLabel(sourceNodeLabel, nodePrefix);
            if(!sourceNodeLabel.equals("&")){
                Node sourceNode=isNodeInGraph(sourceNodeLabel);// Check if node alreadt in graph.
                if(sourceNode==null){// If not then 
                    sourceNode=new Node(sourceNodeLabel, 0, 0); // create new node with label and levels (=0).
                    g.addNode(sourceNode); // Add it to the graph if it does not exist.
                    nodeList.add(sourceNode); // Also add it to the local list of nodes.
                    if(Utilities.graphInputDebug)
                        System.out.println("Node: "+sourceNodeLabel.substring(1));
                    nodeCount++;
                }
                getEdges(sourceNode, in ); // Get possible abstract indicator, outgoing edges, update sourceNode.  
            }
        } // End of while for reading inout from  a file.  
        computeStaticDependence();
        computeInheritanceChains();
        if(!staticOnly)
            computeDynamicDependence();  
    }// End of createGraph()
    
    /*
     * getEdges() reads edge label and destination node label.
     * It then creates an edge and adds it to node n.
     * */
    
    private void getEdges(Node sourceNode, Scanner in){
        while (in.hasNext()){
            String edgeLabel=in.next(); // Get edge label or abstract type; it might be an & implying no more edges.
            // Check if this is the end of data related o sourceNode. 
            if(edgeLabel.equals("&")){
                return; // No more outgoing edges for this node (source).
            } 
            
            // If not then  an abstract node indicator.
            if(edgeLabel.charAt(0)==nodeTypePrefix){
                sourceNode.setAbstract();
            }else{ // It is not a type identifier, hence check for edge.
                if(edgeLabel.charAt(0)==edgePrefix){
                    String destinationNodeLabel=in.next(); // There is an edge. Get label of destination node.
                    nextLine++; // Update line number.
                    checkLabel(destinationNodeLabel, nodePrefix);
                    
                    // Create new node from the destination label only if
                    // such a node does not exist, else use 
                    // the existing node as the destination.
                    
                    Node destinationNode=createNode(destinationNodeLabel);
                    
                    Edge e=new Edge(sourceNode,destinationNode, edgeLabel);
                    if(Utilities.graphInputDebug){
                        System.out.println("Node-Edge-Node: "+sourceNode.getLabel().substring(1)+"-"+
                                           edgeLabel.substring(1)+"-"+destinationNode.getLabel().substring(1));
                    }
                    sourceNode.addOutgoingEdge(e); 
                    edgeCount++; 
                }else{ // This is an invalid entry.
                    System.out.println("At line: "+nextLine+ " invalid coding of edge, node, or node type: "+
                                       edgeLabel+" Prefix expected: e or t");
                }
            }
        }// End of edge loop
    } // End of getEdges()    

  /*
   * createNode() creates a node with label
   * l if such a node does not exist.
   */
  
  private Node createNode(String l){
    Node n=isNodeInGraph(l); // Check if there is a node with label l.
    if(n!=null){
      return n; // Yes, it does, hence no need to create it again!
    }
    
    // Node with label l does not exist. Hence create a new node.
    
    Node newNode=new Node(l, 0, 0);
    g.addNode(newNode);
    nodeCount++;
    nodeList.add(newNode);
    if(Utilities.graphInputDebug){
        System.out.println("Node: "+l.substring(1));
    }
    return newNode;
  } // End createNode()
  
  
  // Check if a node with label l is already in the graph constructed so far.
  private Node isNodeInGraph(String l){
    Iterator itr=nodeList.iterator();
    while(itr.hasNext()){
      Node next=(Node)itr.next();
      String nodeLabel=next.getLabel();
      if(l.equals(nodeLabel)){
        return next; // Node in graph, return it.
      }
    }// End of while
    return null; // Node not in graph, return null.
  }// End of isNodeInGraph()
  
  public Graph getGraph(){ // Return the graph created.
    return g;
  }
  
  // THis method checks if the label of a node or an
  // edge is correctly prefixed (n for node an e for edge).
  
  private void checkLabel(String label, char prefix){
      if(label.equals("&"))
             return;
         
      if(!(label.charAt(0)==prefix))
          System.out.println("At line: "+nextLine+ " invalid coding of edge, node, or node type: "+
                             label+" Prefix expected: "+prefix);
      
  }// End of checkLabel()
  
  public static void main(String [] args) throws IOException{
    CreateGraph cg=new CreateGraph();
   // cg.createGraph("TaiDanielsGraph.txt");
    cg.createGraph("sccTest1NoEdges.txt", false);
    System.out.println("Nodes: "+cg.nodeCount+" Edges: "+cg.edgeCount);
    if(Utilities.graphDisplayDebug){
        Utilities.displayGraph(cg.g, false);
    }
  }// End of main()
  
  
  
   // Find static dependence for all nodes in the graph.
    private void computeStaticDependence(){
        ArrayList<Node> dependsOn;
        int chainType;
        ArrayList<Node> nodes=g.getNodes();
        Iterator<Node> itr=nodes.iterator();
        while(itr.hasNext()){
            Node n=itr.next();
            g.resetNodeAndEdgeMarks(nodes); // Reset all node and edge marks to not visited.
            dependsOn=new ArrayList<Node>(); // Reset dependsOn.
            chainType=Utilities.staticChain; // We need to find static dependencies.
            DFS(dependsOn, n, chainType); // Find them baby!
            n.setStaticDependence(dependsOn);
        }
        
    } // End of updateStaticDependence()
    
    private void  computeDynamicDependence(){
        
        ArrayList<Node> nodes=g.getNodes(); // Get all nodes in the graph.
        ArrayList<Node> destNodesAdded; // Contains list of nodes to which an edge has already been added.
        g.resetNodeAndEdgeMarks(nodes); // Set all nodes to not visited.
        Iterator<Node> itrNode=nodes.iterator(); // Set up node iterator.
        while(itrNode.hasNext()){
            Node chainRootNode=itrNode.next(); // Get the next root node of an inheritence chain.
                ArrayList<Node> inheritanceChain=chainRootNode.getInheritanceChain();
                Iterator <Node> itrChain=inheritanceChain.iterator(); // Set up chain iterator.
                
                while(itrChain.hasNext()){
                    Node chainNode=itrChain.next();// Get next node on the chain.
                     destNodesAdded=new ArrayList<Node>();
                    
                    ArrayList<Edge> chainNodeEdges=chainNode.getIncomingEdges(); // Get incoming edges.
                    Iterator<Edge>  itrEdges=chainNodeEdges.iterator();// Set up edge iterator.
                    while(itrEdges.hasNext()){
                        Edge chainNodeEdge=itrEdges.next(); // Get next incoming edge.
                        if(!chainNodeEdge.getLabel().equals("eI")){// Process only if this is not inheritence. 
                            Node source=chainNodeEdge.getSource(); // Get source of this edge.
                            if(!destNodesAdded.contains(chainRootNode)){
                                source.updateDynamicDependence(chainRootNode); // Add chainRootNode as a dependent node
                                Edge newEdge=new Edge(source, chainRootNode, "eA");
                                source.addOutgoingEdge(newEdge);
                                destNodesAdded.add(chainRootNode); // Mark node an avoid duplicate link.
                            }// End of chainRootNode added test
                        }     
                    }// End of edge iteration  
                }// End of chain node iteration.
       
        }// End of iterator chain.
    }// End of  computeDynamicDependence();
    
    
     // Find inheritance chains for all nodes in the graph.
     private void computeInheritanceChains(){
        ArrayList<Node> dependsOn;
        int chainType;
        ArrayList<Node> nodes=g.getNodes();
        Iterator<Node> itr=nodes.iterator();
        while(itr.hasNext()){
            Node n=itr.next();
            g.resetNodeAndEdgeMarks(nodes); // Again reset all node and edge marks to not visited.
            dependsOn=new ArrayList<Node>(); // Reset dependsOn.
            chainType=Utilities.inheritanceChain; // Now we need to determine inheritance chain.
            DFS(dependsOn, n, chainType); // Find them baby!
            n.setInheritanceChain(dependsOn);
        }
        
    } // End of computeInheritanceChains()
     
     // The following method does a depth first search 
     // of a graph starting at a given node.  
     // Static (chain) dependence considers all edge types (As, Ag, and I).
     // Inheritence chains consider only the "I" type edges that correspond
     // to the inheritance relation.
     
    
   public static void DFS(ArrayList<Node> dependsOn, Node n, int chainType){
        n.setVisitStatus(true);
        ArrayList<Edge> edges=n.getOutgoingEdges();
        
        int numEdges=edges.size();
        for (int i=0; i<numEdges; i++){
            Edge e=edges.get(i);
            if(!ignoreEdge(e, chainType)){
                if(!e.getVisited()){
                    Node nextNode=edges.get(i).getDestination();
                    if(nextNode!=null){
                        if(!nextNode.getVisitStatus()){
                            edges.get(i).markVisited();
                            Utilities.debug(Utilities.dependenceCreationDebug, "--> "+nextNode.getLabel().substring(1));
                            dependsOn.add(nextNode);
                            DFS(dependsOn, nextNode, chainType);
                        }else{
                             Utilities.debug(Utilities.dependenceCreationDebug, " Back Edge: "+n.getLabel().substring(1)+"-"+
                                               edges.get(i).getLabel().substring(1)+"-"+
                                               edges.get(i).getDestination().getLabel().substring(1));
                        }
                    }// End of node check
                } // End of edge visited check
            }// End of chainType check.
        } // End of for  
    }// End of DFS() 
    
   private static boolean ignoreEdge(Edge e, int chainType){
       if(chainType==Utilities.frondEdge)
           return e.isFrond();
       
        if(chainType==Utilities.staticChain)
            return false; // Do not ignore any edge.
        if(chainType==Utilities.inheritanceChain && e.getLabel().equals("eI"))
            return false;
        else 
            return true;
    }// End of ignoreEdge()
   
   // Get number of nodes in a randomly generated graph.
   private int getNodeCount(Scanner s){
     System.out.print("How many nodes in the graph? ");
     int n=s.nextInt();
     if(n<=0){
       n=(int)(Math.random()*(double)(Utilities.maxNodes-Utilities.minNodes)+Utilities.minNodes);
       System.out.println("Generating graph for "+n+" nodes.");
     }
     if(n>Utilities.maxNodes){
       System.out.println("Generating graph for "+Utilities.maxNodes+" nodes.");
       n=Utilities.maxNodes;
     }
    return n;
   }// End of getNodeCount()
        

   // Get number of edges  in a randomly generated graph.
   private int getEdgeCount(Scanner s){
     
     int maxEdges=nodeCount*(nodeCount-1);
     System.out.print("How many edges (max edges="+ maxEdges+") ");
     edgeCount=s.nextInt();
     if(edgeCount>maxEdges){
       System.out.println("Sorry, but edge count is set to its maximum.");
       edgeCount=maxEdges;
       inheritance=false;
     }
     edgeProbability=(double)edgeCount/maxEdges; // Probability of two nodes being connected.
     return edgeCount;
   }// End of getEdgeCount()
   
   private String getEdgeType(){
     double r=Math.random();
     if(r<inheritanceProbability){ 
       return Utilities.InheritanceEdgeLabel; // Inheritence edge
     }else if(r>=inheritanceProbability && r<(aggregationProbability+inheritanceProbability)){
       aggregationEdges++;
       return Utilities.AggregationEdgeLabel; // Inheritence edge
     }
     associationEdges++;
     return Utilities.AssociationEdgeLabel;
   }// End of getEdgeType()

   // Create an edge from source to destination node.
   // 
   private void createEdge(Node source, Node destination){
     double r=Math.random();// Random number between 0 and 1.
     if(!source.isParent(destination)){// Is the source  parent of destination (i.e, there is an inheritance edge)?
       if(r<edgeProbability){// No. Should an edge be created?
         String edgeType=getEdgeType(); // Yes.  Hence try to create an edge.
         if(edgeType.equals(Utilities.InheritanceEdgeLabel)){ // Is an inheritance edge to be created? 
           if(!destination.isAncestor(source)){// Then source node must not be an ancestor of destination.
             addEdge(source,destination, edgeType);// Add an inheritance edge
           }
         }else{
             addEdge(source,destination, edgeType); // Add a non-inheritance edge.
         }
       }// End of edge probability check.
     }// End of node parent check
   } // Ends of createEdge()
   
   // Add an edge between source and destination node.
   private void addEdge(Node source, Node destination, String edgeType){
     Edge e=new Edge(source, destination, edgeType);
     
     source.addOutgoingEdge(e);
     if(edgeType.equals(Utilities.AggregationEdgeLabel)){
       if(Utilities.isThereACycle(nodeList)){
         source.removeOutgoingEdge(e);
         return;
       }
     }
        
     if(edgeType.equals(Utilities.InheritanceEdgeLabel)){
       destination.addParent(source); // Destination node is the parent of the source.
       inheritanceEdges++;
     }
     edgesCreated++;
   }//End addEdge();
   
   // The next method generates  a random ORD. It gets the number of nodes (say, n) 
   // in the graph from the user. Then it gets the number of edges (say, e) from
   // the user. It then generates an ORD with n nodes and e edges.
   
   private void generateRandomGraph(){
     Scanner s=new Scanner(System.in);
     
     nodeCount=getNodeCount(s); // Get number of nodes in the graph.
     edgeCount=getEdgeCount(s); // Get number of edges in the graph.
     
     // Loop for all nodes randomly connecting them to
     // the remaining nodes. Parent-child nodes
     // are treated especially.
     nodesCreated=0;
     edgesCreated=0;
     inheritanceEdges=0;
     aggregationEdges=0;
     associationEdges=0;
     
     
     g=new Graph(); // Create an empty graph.
     // Create and add nodes to the graph.
     // Each node is labelled as N0, N1, N2, etc.
     
     for(int i=0; i<nodeCount; i++){
       Node n=new Node("nN"+Integer.toString(i), 0, 0);
       nodeList.add(n);
       g.addNode(n);
     }// End of node creation loop.
     
     // Create edges.
     
     for(int i=0; i<nodeCount; i++){
       for(int j=0; j<nodeCount; j++){
         if(i!=j){
           createEdge(nodeList.get(i), nodeList.get(j));
         }
       }// End edge creation loop
     }// End node traersal loop.
  
     System.out.println("\nRandom graph generation complete.");
     checkGraph();
     if(Utilities.isThereACycle(nodeList)){
       System.out.println("This graph has cycle(s)");
     }else{
       System.out.println("This graph has no cycle(s)");
     }
     System.out.print("Nodes: "+nodeCount);
     System.out.print("   Edges requested: "+edgeCount);
     int edgesGenerated=inheritanceEdges+aggregationEdges+associationEdges;
     System.out.println("   Edges generated: "+edgesGenerated);
     System.out.print("\tInheritance edges: "+inheritanceEdges);
     System.out.print("\n\tAggregation edges: "+aggregationEdges);
     System.out.println("\n\tAssociation edges: "+associationEdges);
     if(Utilities.graphDisplayDebug)
       Utilities.displayGraph(g, false);
     
   }// End of generateRandomGraph()
   
   // Check for graph validity. Only
   // a simple inheritance edge check is performed.
   private void checkGraph(){
     boolean graphInvalid=false;
     Iterator<Node> itr=nodeList.iterator();
     while(itr.hasNext()){
       Node next=itr.next();
       if(!next.checkInheritance()){
         graphInvalid=true;
       }
     }// End of node iterator
     if(!graphInvalid){
       System.out.println("Graph is valid.");
     }else{
       System.out.println("Graph is invalid.");
     }
     
   }// End checkGraph();
   
}// End of class CreateGraph;

  