/*
 * This class contains utility methods.
 * Most of these are print methods and useful during debugging the 
 * Tai-Daniels (TD) algorithm as
 * well as other agorithms that it uses.
 * 
 * 
 * Author: Aditya Mathur
 * Original: July 7, 2011

 * Latest update: July 24-28, 2011 Aug 11-25 2011. January 28, 20122
 * 
 */

import java.util.*;
import java.io.*;

public  class Utilities{
    
    // Debugging switches. Set to true to make debugging  active
    // false to make it inactive.
  
  static boolean graphInputDebug=false; // Debug graph input method
  static boolean levelAssignmentDebug=false; // Debug for assigning major/minor levels in TD algorithm.
  static boolean tarjanDebug=false;// Debug Tarjan's SCC algorithm
  static boolean edgeDeletionDebug=false; // Debug edge deletion algorithm
  static boolean cycleBreakingDebug=false; // Debug the cycle breaking portion of the TD algorithm.
  static boolean cycleCheckingDebug=false; // Debug the cycle checking portion of the TD algorithm.
  static boolean majorLevelProcessingDebug=false; // Debug for major level portion of the TD algorithm.
  static boolean minorLevelProcessingDebug=false; // Debug for minor portion of the TD algorithm.
  static boolean testOrderDebug=false; // Debug for test sequence generation portion of the TD algorithm.
  static boolean stubCreationDebug=false; // Debug for stub generation portion of the TD algorithm.
  static boolean retestCreationDebug=false; // Debug for retest class portion of the TD algorithm.
  static boolean dependenceCreationDebug=false; // Debug for retest class portion of the TD algorithm.
  static boolean frondDebug=false; // Debug for frond edge check in Yves Le Traon et al. algorithm.
  static boolean weightDebug=false; // Debug for weight check in Briand et al. algorithm.
  static boolean topologicalSortDebug=false; // Debug topological sort algorithm and related methods.
  static boolean resetEdgesDebug=false; // Debug resetting of edges from master root to individual SCC roots.
  static boolean sccDebug=false; // Debug SCC generation.
  static boolean graphDisplayDebug=false; // Debug SCC generation.
  
  
  
  // Useful constants for finding various types of node dependencies in the graph.
  final static int staticChain=1;
  final static int dynamicChain=2;
  final static int inheritanceChain=3;
  final static int frondEdge=4;
  final static int maxNodes=1000; // Maximum number nodes in a randomly generated graph (ORD)
  final static int minNodes=2; // Minimum number nodes in a randomly generated graph (ORD)
  
  // Edge labales
  
  final static String InheritanceEdgeLabel="eI";
  final static String AggregationEdgeLabel="eC";
  final static String AssociationEdgeLabel="eA";

  
  private static int totalFrond; // Number of frond dges in a graph
  static boolean TaiDaniels; // Set to true if this is Tai-Daniels algorithm
  static boolean LeTraonAlgorithm; // Set to true when LeTraon algorithm is used.
  static boolean BriandAlgorithm; // Set to true if it is Briandet al. algorithm.
  
  // Color constants used by the cycle decetion algorithm.
  
  final static int white=0;
  final static int grey=1;
  final static int black=2;
  
  // This variable keeps track of any possible program errors during execution.
  ArrayList<Integer> programError=new ArrayList<Integer>();
  
  //public static enum NodeColor {WHITE,GREY,BLACK};
  
  // Display test order from the graph.
  // Classes are tested in the increasing order of their respective major and minor levels.
  
    public static void displayTestOrder(Graph g, boolean dispGraph, ArrayList<Stub> s, ArrayList<Retest> r, 
                                 int maxMajorLevel, int [] maxMinorLevel){
        
        if(dispGraph){
            System.out.println("\nUpdated Graph:");
            displayGraph(g,true);
        }
        
       System.out.println("\nClass test order");
        
        for(int i=1; i<=maxMajorLevel; i++){
            if(maxMinorLevel[i-1]==0){
                 System.out.println("** Program error ** No nodes at level "+i);
            }
            for(int j=1; j<=maxMinorLevel[i-1]; j++){
                  ArrayList<Node> nodes=g.getNodes(i,j);
                  displayNodes(nodes, false, false);  
                    System.out.println("\t---------------------");
            }// End of minor level loop
          
        }// End of major level loop
        System.out.println("\n Total stubs created: "+s.size());
        displayStubsCreated(s);
        
        System.out.println("\nRetest nodes and edges");
        if(r!=null){
            displayRetestEdges(r);
        }
        
        
    }// End displayTestOrder()
    
    
    // Debug routine to display all nodes in a list.
    // Outgoing edges are displayed only when edgeAlso is true.
    
    
    static ArrayList<Node>succNodes;
    public static void displayNodes(ArrayList<Node> list, boolean edgesAlso, 
                                    boolean withWeight){
        
        if(list.size()==0){
            System.out.println("\tNo nodes to display.\n");
            return;
        }
        // List is not empty, hence display it.
        Iterator itr =list.iterator();
        while(itr.hasNext()){
            Node nextNode=(Node) itr.next();
            String abs="";
            String stub="";
           if(nextNode.isNodeAbstract())
               abs=" (Abstract) ";
           if(nextNode.isStub())
               stub=" (Stub)";
            
            System.out.println("\t"+nextNode.getLabel().substring(1)+abs+" Traversal index: "+
                               nextNode.getTraversalIndex()+" Level: "+
                               nextNode.getLevel(0)+"."+nextNode.getLevel(1)+stub);
            displayDependence(nextNode, edgesAlso, withWeight, Utilities.staticChain);
            displayDependence(nextNode, edgesAlso, withWeight, Utilities.inheritanceChain);
            displayDependence(nextNode, edgesAlso, withWeight, Utilities.dynamicChain);
            displayStubsUsed(nextNode);
            
            } // End of while  
        } // End of displayNodes()

    // Display dependent nodes (both static and dynamic).
    private static void displayDependence(Node n, boolean edgesAlso, boolean withWeight, int depType){
        ArrayList<Node> dependsOn=null;
        String dep="";
        switch (depType){
            case 1: {
                dependsOn=n.getStaticDependence();
                dep="Static dependency"; break;
            }
            case 2: {
                dependsOn=n.getDynamicDependence();
                dep="Dynamic dependency"; break;
            }
             case 3: {
                dependsOn=n.getInheritanceChain();
                dep="Inheritance chain"; break;
            }
            default: {System.out.println("Incorrect dependence type in displayDependence");
                break;
            }
        }// End of switch
        
         Iterator<Node> depends=dependsOn.iterator();
            if(dependsOn.size()==0){
                System.out.println("\t    This node has no "+dep); 
            }else{
                System.out.print("\t"+"    "+     dep+": ");
                String filler="";// Used for formatting
                while(depends.hasNext()){
                    Node d=depends.next();
                    System.out.print(filler+d.getLabel().substring(1));
                    filler=", ";
                }
                System.out.println();
                if(edgesAlso){
                    displayEdges(n, withWeight, null);
                }
            }
    }// End of displayDependence()
    
    // Debug routine to display all stubs in a list of stubs.[ Used in Tai-Daniels algorithm.]
    
    public static void displayStubsCreated(ArrayList<Stub> list){   
        if(list.size()==0){
            System.out.println("\tNo stubs to display.\n");
            return;
        }
        // List is not empty, hence display it.
        Iterator <Stub> itr =list.iterator();
        while(itr.hasNext()){
            Stub next=itr.next();
            Node stubNode=next.getStubNode();
            ArrayList<Node> usedByNodes=next.getUsedByNodes();
            System.out.print("\n\tStub for: "+stubNode.getLabel().substring(1)+" Level: "+
                             stubNode.getLevel());
            if(usedByNodes.size()<1){
                System.out.println("\n** This stub is not used by any node. Perhaps a program error.");          
            }else{
                System.out.print("  Used by: ");
                // Display all nodes that use this stub.
                
                Iterator <Node> itrNode=usedByNodes.iterator();
                Node nextUsedNode=new Node();
                while(itrNode.hasNext()){
                    nextUsedNode=itrNode.next();
                    System.out.print(nextUsedNode.getLabel().substring(1)+" ");
                }// End of while; display nodes that use this stub
                System.out.println();
                
                // Now display the edges that require the stub to be used.
                itrNode=usedByNodes.iterator();
                while(itrNode.hasNext()){
                    nextUsedNode=itrNode.next();
                    displayEdges(nextUsedNode, false, stubNode);
                }// End of while; displasy edges that require stub use.
            }
        } // End of while  over stubs
    } // End of displayStubsCreated()
    
    public static String displayStubs(Graph g){
        ArrayList<Node> nodes=g.getNodes();
        Iterator<Node> itr=nodes.iterator();
        String stubNodes="";
        while(itr.hasNext()){
            Node n=itr.next();
            if(n.isStub()){
                stubNodes=stubNodes+n.getLabel().substring(1)+" ";
            }
        }// End of node iteration.
        
        return stubNodes;
    }// End displayStubs()
    
    private static  void displayStubsUsed(Node n){
        ArrayList<Node> stubsUsed=n.getStubsUsed();
        System.out.print("\t"+"    "+  "Stub(s) used for node(s): ");
        if(stubsUsed.size()==0){
            System.out.println("None.");
            return;
        }
        // Else display all nodes whose stubds are used by n.
        for(int i=0; i<stubsUsed.size(); i++){
            System.out.print(stubsUsed.get(i).getLabel().substring(1)+" ");
        }// End of node loop
        System.out.println();
        
    }// End of displayStubsUsed()
    
      public static void displayRetestEdges(ArrayList<Retest> list){   
        if(list.size()==0){
            System.out.println("\tNo nodes/edges to display.\n");
            return;
        }
        // List is not empty, hence display it.
         Iterator <Retest> itr =list.iterator();
        while(itr.hasNext()){
            Retest next=itr.next();
            Node nodeToBeTested=next.getNodeUnderTest();
            ArrayList<Edge> edgesToBeTested=next.getEdgesToBeTested();
            System.out.print("\t"+"When testing  "+nodeToBeTested.getLabel().substring(1)+" Level: "+
                               nodeToBeTested.getLevel());
            System.out.println("  re-test the following edges: ");
            Utilities.displayEdges(edgesToBeTested, false, null);

        } // End of while for Retest 
    } // End of displayRetestEdges()

    public static void displayMajorLevels(Graph g, int maxLevel){
        System.out.println("\nMajor levels \n");
        for (int i=1; i<=maxLevel; i++){
            System.out.println("Nodes at major level "+i+ ":");
            Iterator itr=g.getNodes().iterator();
            while (itr.hasNext()){
                Node next=(Node)itr.next();
                int nodeLevel=next.getLevel(0);
                if(nodeLevel==i){
                    System.out.print(next.getLabel().substring(1)+ " ");
                }
            } // End of while
            System.out.println();
        } //End of for
  } // End of displayMajorLevels().  
    
    public static void displayNodeList(ArrayList<Node>n,  Node stubNode, boolean edgesAlso, boolean displayLevels) {
        if(n.size()==0){
            System.out.println("Node list is empty.");
            return;
        }
        Iterator itr=n.iterator();
        while(itr.hasNext()){
            Node next=(Node)itr.next();
            int nodeLevelMajor=next.getLevel(0);
            int nodeLevelMinor=next.getLevel(1);
            System.out.print(next.getLabel().substring(1));
            if(displayLevels)
                 System.out.print(" "+" Major:Minor level: " +nodeLevelMajor+":"+nodeLevelMinor);
                
            System.out.println();
            
           if(edgesAlso)
               displayEdges(next, false, null);
        } // End of while over nodes
        
    } // End of displayNodeList()
    
    public static String listNodeSequence(ArrayList<Node> s){
        Iterator<Node> itr =s.iterator();
        String seq="";
        while(itr.hasNext()){
            seq=seq+itr.next().getLabel().substring(1)+" ";
        }
        return seq;
    }
    // Print a graph: all nodes and the corresponding edges.
    
    public static void displayGraph(Graph g, boolean frondDisplay){ 
        String frondInfo="", stub="";
        Iterator itr=g.getNodes().iterator();
        int totalEdges=0;
        int nodeCount=0;
        int stubCount=0;
        totalFrond=0;
        String stubList=" (";
        while(itr.hasNext()){ 
            Node n=(Node) itr.next();
            stub="";
            
            if(n.isStub()){
                stub=" (Stub)";
                stubList=stubList+" "+n.getLabel().substring(1); 
            }
            nodeCount++;
            if(n.isStub())
                stubCount++;
           frondInfo="";
            if(LeTraonAlgorithm)
              frondInfo=" Frond Count:"+n.getFrondWeight();
            
            System.out.print("Node: "+ n.getLabel().substring(1)+" Traversal index: "+n.getTraversalIndex()+
            " Belongs to Scc: "+n.getSccIndex());
            if(Utilities.TaiDaniels){
                System.out.println(" "+n.getLevel(0)+" "+n.getLevel(1));
            }else{ 
                if(frondDisplay){
                    System.out.print(frondInfo);
                }
                if(n.isMasterRoot()){
                    System.out.print(" (Master root)") ;
                }else{
                    if(n.isRoot()){
                        System.out.print(" (Root)") ;
                    }
            }
            }// End of else (i.e., not TD algorithm.)
            System.out.println(stub);
            
            System.out.println(" Incoming/outgoing edges:"+n.getIncomingEdgeCount()+
                               "/"+n.getOutgoingEdgeCount());
            totalEdges=totalEdges+n.getOutgoingEdgeCount();
            displayEdges(n, false, null);
            System.out.println("---------");
        }// End of iteration over graph nodes.
        stubList=stubList+" )";
        
        System.out.print("\nNodes: "+nodeCount+". Edges: "+totalEdges+ ". "+frondInfo);
        System.out.println(" Stubs: "+stubCount);
        if(stubCount>0)
          System.out.println("Stubs: "+stubList);
    } // End displayGraph()
    
    
    // Print graph but by level numbers.
    
     public static void printGraphByLevelNumbers(Graph g, int maxMajorLevel, int[] maxMinorLevel){ 
         
         
         int nodeCount=0;
         int edgeCount=0;
         for(int i=1; i<=maxMajorLevel; i++){
             System.out.print("\nNodes at major level "+i);
             System.out.println(". Max minor level: "+maxMinorLevel[i-1]);
             System.out.print("\t");
             
             Iterator <Node> itr=g.getNodes(i).iterator();     
             while(itr.hasNext()){ 
                 Node n=(Node) itr.next();
                 nodeCount++;
                 edgeCount=edgeCount+n.getOutgoingEdgeCount();
                 System.out.print(n.getLabel().substring(1)+" "+
                                  n.getLevel(0)+"."+n.getLevel(1)+"  ");
             }// End of iteration over nodes at level i
             
         }// End of iteration over the major levels.
        
         System.out.println("\nTotal Nodes: "+nodeCount+" Total Edges: "+edgeCount);
     } // End printGraphByLevelNumbers()
    
    // Display all edges emanating from a node. ff stubNode is not null
     // then only the edges ending at stubNode are displayed.
     
    public static void displayEdges(Node n, boolean weightDisplay, Node stubNode){
        String label=n.getLabel();
        ArrayList<Edge> edges=n.getOutgoingEdges();
        Iterator itr =edges.iterator();
        while(itr.hasNext()){
            Edge nextEdge=(Edge)itr.next();
            Node destination=nextEdge.getDestination();

            if(!(stubNode==null)){
                if(destination.equals(stubNode)){ 
                   displayEdge(nextEdge, weightDisplay);
                }
            }else{
                 displayEdge(nextEdge, weightDisplay);
            }
        }// End of while over edges.
    } // End of displayEdges()
        
 
    
    public static void displayEdges(ArrayList<Edge> e , boolean weightDisplay, Node stubNode){
        if(e.size()==0){
            System.out.println("No edges to display.");
            return;
        }
        Iterator itr =e.iterator();
        while(itr.hasNext()){
            Edge nextEdge=(Edge)itr.next();
            Node source=nextEdge.getSource();
            Node destination=nextEdge.getDestination();
            String labelSource=source.getLabel();
            String labelDestination=destination.getLabel();
            String labelEdge=nextEdge.getLabel();
           
            System.out.print("\t"+labelSource.substring(1)+"-"+"("+labelEdge.substring(1) +")"+"-"+
                             labelDestination.substring(1));
            if(weightDisplay){ 
                int weight=nextEdge.getWeight();
                String sign="+";
                if(BriandAlgorithm)
                  sign="*";
                System.out.print(" Weight: "+nextEdge.getWeight()+" ("+nextEdge.getWeightIn()+
                                 sign+nextEdge.getWeightOut()+")");
            }
            if(nextEdge.isFrond()){
                totalFrond++;
                System.out.print("  (Frond)");
            }
            System.out.println();
        }// End of edge iteration loop.
    } // End of displayEdges, another version
    
    
    // Display deleted edges.
    public static void displayDeletedEdges(Graph g){
      ArrayList<Edge> deletedEdges=null;
      ArrayList<Integer> seqNum=null;
      ArrayList<Node> nodes=g.getNodes();
      if(nodes.size()==0){
        System.out.println("No deleted edges to display.");
        return;
      }
      Iterator <Node> itrNode =nodes.iterator();
      while(itrNode.hasNext()){
        Node n=itrNode.next();
        ArrayList<Edge> edges=n.getIncomingEdges();
        if(edges.size()>0){
          Iterator<Edge> itrEdge=edges.iterator();
          while(itrEdge.hasNext()){
            Edge e=itrEdge.next();
            if(e.isDeleted()){
              int seq=e.getDeleteSequence();
              if(seqNum==null){
                seqNum=new ArrayList<Integer>();
                seqNum.add(seq);
                deletedEdges=new ArrayList<Edge>();
                deletedEdges.add(e);
              }else{
                if(!seqNum.contains(seq)){
                  seqNum.add(seq);
                  deletedEdges.add(e);
                }// End of addition of a new edge
              }
            }// End of deleted edge check
          } // End of edge iteration loop. 
        }//End of edges count check
      }// End of nodes loop.
      sortEdgeList(deletedEdges, seqNum);
      
      for(int i=0; i<deletedEdges.size(); i++){
        System.out.print(seqNum.get(i)+".  ");
        System.out.println(stringifyEdge(deletedEdges.get(i), true));
      }
    }// End of displayDeletedEdges()
    
    private static void sortEdgeList(ArrayList<Edge> deletedEdges, ArrayList<Integer> seqNumber){
      
      if(deletedEdges.size()<=0)
        return;
      
      for(int i=0; i<seqNumber.size(); i++){
        for(int j=0; j<seqNumber.size()-i-1; j++){
          int prev=seqNumber.get(j);
          int next=seqNumber.get(j+1);
          if(prev>next){
            int temp=prev;
            Edge tempEdge=deletedEdges.get(j);
            seqNumber.set(j, next);
            deletedEdges.set(j, deletedEdges.get(j+1));
            seqNumber.set(j+1, temp);
            deletedEdges.set(j+1, tempEdge);
          }
        }
      }
    }// End of sortEdgeList()
    
    public static String stringifyEdge(Edge e, boolean weightDisplay){
        String edgeString=e.getSource().getLabel().substring(1)+"-"+"("+
            e.getLabel().substring(1)+")"+"-"+e.getDestination().getLabel().substring(1);    
        
        if(weightDisplay){
            edgeString=edgeString+" Weight: "+Integer.toString(e.getWeight());
        } 
        
       if(e.isFrond()){
                edgeString=edgeString+"  (Frond)";
                totalFrond++;
            }
        return edgeString;
    }// End of stringifyEdge()
    
      public static String stringifyNodes(ArrayList<Node> n){
        String nodesString="";
        Iterator<Node> itr=n.iterator();
        while(itr.hasNext()){
            Node next=itr.next();
            String label=next.getLabel().substring(1);
            String root="";
            String stub="";
            String rootStub="";
            if(next.isMasterRoot()){
                root=" (MasterRoot)";
            }else{
              if(next.isRoot()){
                root=" (Root)";
              }
            }
            if(next.isStub())
                stub="(Stub)";
            rootStub=root+stub+" ";
            nodesString=label+rootStub+nodesString;
        }
        return nodesString;
    }// End of stringifyNodes()
      
      public static String stringifyNodes(Node[] n){
          String nodesString="";
          for(int i=0; i<n.length; i++){
              if(!(n[i]==null)){
                  nodesString=n[i].getLabel().substring(1)+" "+nodesString;
                  if(n[i].isStub()){
                      nodesString=nodesString+" (Stub) ";
                  }
              }
          }
          return nodesString;
      }// End of stringifyNodes()
    
    // Print all strongly connected components.
    
      public static void displayAllScc(ArrayList<ArrayList<Node>> scc){
          
          int componentCount=scc.size();
          if(componentCount==0){
              System.out.println("No strongly connected components found.");
              return;
          }
          // Loop through the ArrayList and print all SCCs.
          System.out.println("\n"+componentCount+ "  strongly connected component(s):");
          Iterator <ArrayList<Node>> itrAll=scc.iterator(); // This iterates over all components.
          int nextComponent=1; // Index of component to be printed.
          while(itrAll.hasNext()){
            System.out.print("Component: "+nextComponent);
            ArrayList<Node> component=itrAll.next();
            if(Utilities.cycleCheckingDebug){
              if(Utilities.isThereACycle(component)){
                System.out.println(". Has a cycle.");
              }else{
                System.out.println(". Has no cycle.");
              }
            }
            displayAnScc(component);
            
            nextComponent++; 
            
          }// End of outer while for all components.
          
      } // End of displayAllScc()
     
     
     // The following method displays one SCC. It is useful in debugging.
      public static void displayAnScc(ArrayList<Node> scc){
        
        int componentSize=scc.size();
        if(componentSize==0)
            return;
        // Loop through the ArrayList and print all SCCs.
        Iterator  itr=scc.iterator(); // This iterates over all nodes in the component
        while(itr.hasNext()){    
          Node n=(Node)itr.next();
          String frondInfo="";
          if(LeTraonAlgorithm){
            frondInfo="Frond count: "+ n.getFrondWeight();
          }
          System.out.print("    "+n.getLabel().substring(1)+" Traversal index: "+
                           n.getTraversalIndex()+
                           frondInfo); 
          String rootStub="";
          if(n.isMasterRoot()){
            rootStub=" (Master Root)";
          }else{
            if(n.isRoot())
              rootStub=" (Root)";
          }
          if(n.isStub())
            rootStub=rootStub+" (Stub)";
          System.out.println(rootStub);
        }// End of inner while
        System.out.println(); // Move to the next line.
      }// End of  displayAnScc for one component
      
      //  Display  initial Sccs created during first call to tarjan() in LeTraon et al, algorithm
      public static void displayScc(ArrayList<Scc> sccs){
          Iterator<Scc> itr=sccs.iterator();
          while(itr.hasNext()){
              Scc s=itr.next();
              ArrayList<Node> nodes=s.getScc();
              Node rootInitial=s.getInitialRoot();
              Node rootCurrent=s.getCurrentRoot();
              int index=s.getIndex();
              System.out.print("Initial Scc: Index:"+ index+" Initial Root: "+ 
                                 rootInitial.getLabel().substring(1)+
                                 " Current root: "+rootCurrent.getLabel().substring(1));
              String rootStub="";
              if(rootCurrent.isMasterRoot()){
                rootStub=" (Master Root)";
              }
              else{
                if(rootCurrent.isRoot())
                  rootStub=" (Root)";
              }
              if(rootCurrent.isStub()){
                rootStub=rootStub+" (Stub)";
              }
              System.out.println(rootStub);
              displayAnScc(nodes);
          }
      }// End of  displayScc()
      
     
     // The next method displays successors of each node in the given graph.
     
     public static void displaySuccessors(Graph g, ArrayList<String> ignoreEdges){
    
         ArrayList<Node> nodes=g.getNodes();
         System.out.println("\nAll succesors.\n");
         Iterator itr=nodes.iterator();
         while(itr.hasNext()){
             Node n=(Node)itr.next();
             ArrayList<Node> succ=n.getSuccessors(ignoreEdges);
             System.out.println("Node: "+n.getLabel());
             displayNodeList(succ,null,  true, false);
         } 
     }// End of displaySuccessors()
     
     public static void displayEdge(Edge e, boolean withWeight){
         String source=e.getSource().getLabel().substring(1);
         String destination= e.getDestination().getLabel().substring(1);
         String sccDestination=e.getSccDestination().getLabel().substring(1);
         System.out.print("\tEdge: "+source+"-"+
                          "("+e.getLabel().substring(1)+")"+"-"+
                          destination);
         if(!sccDestination.equals(destination)){
             System.out.print(" (Scc destination: "+sccDestination+")");
         }
         if(withWeight){
             int weightIn=e.getWeightIn();
             int weightOut=e.getWeightOut();
             
             System.out.println("\tWeight: "+e.getWeight()+" ("+weightIn+"+"+weightOut+")");
         }
         if(e.isFrond()){
             totalFrond++;
             System.out.print(" (Frond)");
         }
         System.out.println();
     }// End of display edge.
     

     // Method to display grammatically correct counts.
     // Edge "2 edges" or "1 edge".
     
     public static String singPlu(int n){
         
        if(n>1)
            return "s";
        else 
            return "";
     }// End of singPlu
     
     // This metod find the list of class that on hwich the inut class n depends. 
     // The method understands that there may be a cycle in the graph.
     
    
     public ArrayList<Node>  dependsOn(Node n){
         
         ArrayList<Edge> edges=n.getOutgoingEdges();
         Iterator<Edge> itr=edges.iterator();
         while(itr.hasNext()){
             Edge e=itr.next();
             Node dest=e.getDestination();
             if(dest.equals(n)){
                 succNodes.add(dest);
                 return succNodes;
             }else{
                 succNodes.add(dest);
                 dependsOn(dest);
             }  
         }
         return(succNodes);
     }// End of dependsOn()
     
   
     
     // Topological sort method. [Based on the CoffmanGraham algorithm.]
     // Input graph g is assumed to be a DAG.
     
     private static ArrayList<Node> topSortOrder=new ArrayList<Node>(); 
     public static ArrayList<Node> topologicalSort(Graph g){
         
       // First get nodes with no incoming edges.
       // Make sure that deleted edges are ignored.
         ArrayList<Node>noIncomingEdges=g.getNodesNoIncomingEdgesIgnoreDeleted();
         Utilities.debug(Utilities.testOrderDebug, "Nodes with no incoming edges: "+
                         stringifyNodes(noIncomingEdges));
         
         Iterator<Node> itr=noIncomingEdges.iterator();
         
         g.resetMarks(); // Set all nodes to not visited.
         while(itr.hasNext()){
             Node n=itr.next(); 
             visit(n);
         }
         return topSortOrder;
     }// End of topSort()
     
     
     
     private static void visit(Node n){
       if (!n.getVisitStatus()){
         n.setVisitStatus(true);
         Utilities.debug(topologicalSortDebug, "Top sort: next node visited: "+
                         n.getLabel().substring(1));
         ArrayList<Edge> edges=n.getOutgoingEdges();
         Utilities.debug(topologicalSortDebug, "     # of Edges: "+edges.size());
         Iterator<Edge> itr=edges.iterator();
         while(itr.hasNext()){
           Edge e=itr.next(); 
           if(!e.isDeleted()){
             Utilities.debug(topologicalSortDebug, "     Top sort: edge used: "+
                             stringifyEdge(e, false));
             Node m;
             m= e.getSccDestination();
             Utilities.debug(topologicalSortDebug, "     Top sort: destination node visited: "+
                             m.getLabel().substring(1));
             visit(m); 
          }
         }// End of edge iteration
         topSortOrder.add(n);
         Utilities.debug(Utilities.topologicalSortDebug, "Current test order: "+
                        stringifyNodes(topSortOrder));
       }// End of if
     }// End of visit()
     
     
     // Display test sequence recursively. Used for debugging.
     public static void leTraonTestSequence(ArrayList<ArrayList<Node>> scc){
       System.out.println("Current test sequence:");
       
       Iterator<ArrayList<Node>> itr=scc.iterator();
       int serial=0;
       while(itr.hasNext()){
         ArrayList<Node> anScc=itr.next();
         serial++;
         String nodeList=stringifyNodes(anScc);
         System.out.println(serial+". "+"( "+nodeList+")");
       }
       
     }// End of laTraonTestSequence()
     
     // Find test order. This simply uses topological sort
     // on the graph. The test order is reverse of the topological sort
     
     public static ArrayList<Node> findTestOrder(Graph g){
         ArrayList<Node> testOrder=topologicalSort(g);
         
         Utilities.debug(Utilities.topologicalSortDebug, "Reverse test order: "+stringifyNodes(testOrder));
       
         return testOrder;
     }// End of findTestSequence()
     
     // Display test sequence using stub and other node dependencies.
     // The input node list contains nodes in the order they are to be tested.
     
     public static void displayTestOrder(ArrayList<Node> n){
       int specificStubs=0; // Count of specific stubs
       boolean stubAdded=false;
       Utilities.debug(Utilities.topologicalSortDebug, "Test order: "+stringifyNodes(n));
       Iterator <Node> itrNode=n.iterator();
       int testSeq=0;
       while(itrNode.hasNext()){
         String dependsOn="";
         int dependsCount=0;
         Node nextTestNode=itrNode.next();
         ArrayList<Edge> edges=nextTestNode.getOutgoingEdges();
         Iterator<Edge> itrEdge=edges.iterator();
         while(itrEdge.hasNext()){
           Edge nextEdge=itrEdge.next();
           Node d=nextEdge.getDestination();
           stubAdded=false;
           // Create a string to be displayed for dependence.
           String stub=""; 
           String andString="";
           if(dependsCount>0){
             andString=",";
           }
           if(dependsCount>0 && dependsCount==edges.size()-1){
             andString= " and ";
           } 
           
           // Check if destination is not tested and is a stub.
           if(d.isStub()&&!d.isTested()){
             stub=" Stub("+d.getLabel().substring(1)+")"; // Mark this as a stub.
             stubAdded=true;
           }
           
           // Check if destination has not ben tested and is NOT a stub.
           // If true, this means the destination should be made a stub.
           // Sometimes the algorithm does not make a node a stub because
           // its source or destination is already a stub.
           // This happens when applying Briand algorithm.
           if(!d.isTested()&&!d.isStub()){
             d.markStub();
             stub=" Stub("+d.getLabel().substring(1)+")"; 
              stubAdded=true;
           }
           
           // Check if destination has already been tested in which case no stub needed.
           if(d.isTested()){
             stub=" "+d.getLabel().substring(1);
           }
           dependsOn=dependsOn+andString+stub;
           dependsCount++;
           if(stubAdded)
             specificStubs++;
           
         }// End of edge iteration for a test node.
         nextTestNode.markTested();
         testSeq++;
         String using=" (independent)";
         if(dependsCount>0){
           using=" using "+dependsOn;
         }
         System.out.println(testSeq+" Test "+nextTestNode.getLabel().substring(1)+using); 
         
       }// End test node iteration
       System.out.println("\nSpecific stubs: "+specificStubs);
       
     }// End of displayTestOrder()
     
     // Reverse the node list so it
     // it is in the order in which nodes are to be tested.
     
     private static void reverseNodeList(ArrayList<Node> n){
         int front=0;
         int tail=n.size()-1;
         while(front<tail){
             Node f=n.get(front); // Save reference to front node.
             n.set(front, n.get(tail)); // Replace this poisition by the tail node
             n.set(tail, f); // Please frot node at tail;
             front++; // Move front one step ahead and
             tail--;   // tail one step backwards.
         }

     }// End of reverseNodeList()
     
       public static void debug(boolean debugFlag, String message){
         if(debugFlag)
             System.out.println(message);
     }// End of debug()
       
       public static void displaySccStack(Stack <ArrayList<ArrayList<Node>>> stack){
           System.out.println("\n***** Scc Stack dump");
           Iterator <ArrayList<ArrayList<Node>>> itr=stack.iterator();
           int level=0;
           while(itr.hasNext()){
               ArrayList<ArrayList<Node>> sccList=itr.next();
               System.out.println("Stack level: "+level);
               displayAllScc(sccList);
               level++;
           }
           
       }// End of displaySccStack
       
       
       // Colored DFS algorithm for detecting whether or not
       // a graph (collection of nodes/edges), has a cycle.
       
       // Reference: www.cs.berkeley.edu/~kamil/teaching/sp03/041403.pdf
       
      
       public static boolean isThereACycle(ArrayList<Node> nodes){
         Iterator<Node> itr=nodes.iterator();
         while(itr.hasNext()){
           Node n=itr.next();
           n.setColor(white);
           n.setVisitStatus(false);
         
         }// End of loop to mark nodes as not visited.
         
         itr=nodes.iterator();
         while(itr.hasNext()){
           Node n=itr.next();
           if (n.getColor()==white) {
             if (visitNode(nodes, n)){
               return true;
             }
           }
         }// End of second node iterator
         return false;
       }// End of isThereACycle()
       
       private static boolean visitNode(ArrayList<Node> nodes, Node v){
           v.setColor(grey);
           ArrayList<Edge> edges=v.getOutgoingEdges(nodes);
           Iterator<Edge> itr=edges.iterator();
           while(itr.hasNext()){
               Edge e=itr.next();
               Node u=e.getDestination();
               if (u.getColor() == grey){ 
                   return true;
               }
               else {
                   if (u.getColor() == white){
                       if (visitNode(nodes, u)) {
                           return true;
                       }
                   }
               }// End of else
           }// End of edge iterator
           v.setColor(black);
           return false;
       }// End of visitNode()
       
       
       // This is to test methods in this class.
       public static void main(String [] args){
         
       }
           
     
       
} // End of class Utilities