

/*
 * 
 * Random ORD (Graph) generation.
 * 
 * Coded by: Aditya Mathur
 * Version 1.0: January 29-31, 2012.
 * 
 * Class CreateRandomORD is used  for generating
 * an ORD (a directed graph with typed edges) in integration testing. 
 * It asks for the following inpputs:
 * 
 *     Number of nodes in the ORD (n)
 *     Required ORD density (0<=d<=1)
 *     Proportion of inheritance edges (0<=iE<=1)
 *     Proportion of aggregation edges ((0<=agE<=1))
 * 
 * The sum of the proportion of the  inheritance, 
 * aggregation and association edges is 1.
 * 
 * It then generates a graph with n nodes and e edges, where
 * e is approximately equal to n(n-1)*d. It is assumed that
 * there are no self edges in the ORD.
 * 
 * Inheritance and aggregation edges do not form a cycle whereas
 * association edges may.
 * 
 * The generated ORD is saved in a plain text file named NnEe.
 * Thus, an ORD with 15 nodes and 99 edges is saved in a file named N15E99.
 * The file is saved in folder whose path is provided as a parameter
 * to thye CreateRandopmORD constructor.
 * 
 * Format of the file containing graph nodes/edges follows.
 * 
 * Data for each node or edge occupies one line.
 * Lines are separated by CRLF (generated by \n in Java)
 * Each node contains the following information: node label
 * There is exactly one line for each nodel label.
 *
 * Node-label
 * 
 * The line for each node is followed by as many lines 
 * as there are edges going out of that node.
 * 
 * edge-label destination-node-label 
 * 
 * An & indicates the end of edges out of a node. It
 * is placed by itself on a line.  A node label by itself on a line 
 * and followed on the next line by an & indicates that this
 * node has no outgoing edges.
 * 
 * The label of a node is  prefixed by an "n". For example
 * label nB indicates that B is a node. Similarly an edge
 * label is prefixed by an "e". For
 * example, label eF indicates an edge labelled F. A node can be labelled
 * as Abstract by using the tag tA following the node name.
 * 
 * Example output:
 * nBL
 * eA nB
 * eA nY
 * &
 * B
 * eA nSS
 * eF nE
 * eF nBS
 * &
 * nX
 * tA
 * eI nY
 * &
 * nY
 * &
 * 
 * The above (hypothetical) data is for a graph with  nodes labeled BL, B, SS, E,  BS, and X.
 * Node BL has one outgoing edge, node B has three outgoing edges and
 * the remaining nodes have no outgoing edges. Node X is abstract.
 * 
 * The obvious: Each node must have a label and each edge must have a label.
 * 
 * 
 * July 7, 2011.
 * Latest update: Aug 11, 2011
 * 
 * January 28, 2012. Added randomGraph() method to generate a random ORD.
 * Aditya Mathur
 * */

import java.util.*;
import java.io.*;
import java.lang.Object;
import java.text.DateFormat;
import java.text.SimpleDateFormat;

public class CreateRandomORD{
    
    Graph g; //Graph to be created.
    private ArrayList <Node> nodeList=new ArrayList<Node>(); // For internal use.
    int nodeCount=0; // NUmber of nodes in the graph.
    int edgeCount=0; // Number of edges in the graph.
    int maxEdges; // Maximum possible edges inthe generated graph.
    double density; // DEnsity of the graph (edgeCount/maxEdges).
    int nextLine; // Used for displaying error in input file.
    boolean staticDynamic=true; // true implies dependency is static, otherwise this is false.
    boolean inheritance=true; // Indicates weather or not inheritance edges are to be generated.
    
    // Probabilities associated with edges.
    double edgeProbability; // Probability that two nodes are connected by an edge.
    // Default edge type probabilities. Must sum to 1.0
    double inheritanceProbability=0.0; 
    double aggregationProbability=0;
    double associationProbability=1.0;
    
    // Se dafault max and min nodes in the ORD.
    
    final static int maxNodes=1000; // Maximum number nodes in a randomly generated graph (ORD)
    final static int minNodes=2; // Minimum number nodes in a randomly generated graph (ORD)
    
    
    // Initialize ORD metrics for randomly generated graph.
    int nodesCreated=0;
    int edgesCreated=0;
    int inheritanceEdges=0;
    int aggregationEdges=0;
    int associationEdges=0;
    int edgesDiscardedParent=0; // Parent cannot have an edge to child.
    int edgesDiscardedCycle=0; // Inheritance and aggregation edges cannot create a cycle.
    
    boolean cycleExists; // Indicates whether or not there is a cycle in the generated ORD.
    
    
    String ordData=""; // Holds data to be writen into a file
    String ordFileName; // Name of the file in which ORD is to be written
    String fileDestination; // Path to the directory where the ORD is to be saved.
    
    // If filename is an empty string then a a graph is generated randomly.
    public void CreateRandomORD(String fileDestination, boolean staticOnly) {
        this.fileDestination=fileDestination;
        generateRandomGraph();
        
    }// End of CreateRandomORD()
    
    
    // Get number of nodes in a randomly generated graph.
    private int getNodeCount(Scanner s){
        System.out.print("How many nodes in the graph? ");
        int n=s.nextInt();
        if(n<=0){
            n=(int)(Math.random()*(double)(maxNodes-minNodes)+minNodes);
            System.out.println("Generating graph for "+n+" nodes.");
        }
        if(n>maxNodes){
            System.out.println("Generating graph for "+maxNodes+" nodes.");
            n=maxNodes;
        }
        return n;
    }// End of getNodeCount()
    
    
    // Get number of edges  in a randomly generated graph.
    private int getGraphDensity(Scanner s){
        
        maxEdges=nodeCount*(nodeCount-1); // No self edges.
        System.out.print("Graph density (0-1): ");
        density=s.nextDouble();
        if(density<0|| density>1){
            density=0.5;
            System.out.println("Sorry, density has been set to "+density);
        }
        
        // Compute edges requested in the graph.
        edgeCount=(int)(maxEdges*density);
        
        edgeProbability=density; // Probability of two nodes being connected.
        return edgeCount;
    }// End of getEdgeCount()
    
    // Generate a random edge type based on edge type probabilities.
    
    private String getEdgeType(){
        double r=Math.random();
        if(r<inheritanceProbability){ 
            return Utilities.InheritanceEdgeLabel; // Inheritence edge
        }else if(r>=inheritanceProbability && r<(aggregationProbability+inheritanceProbability)){
            return Utilities.AggregationEdgeLabel; // Inheritence edge
        }
        associationEdges++;
        return Utilities.AssociationEdgeLabel;
    }// End of getEdgeType()
    
    // Create an edge from source to destination node.
    // Make sure that all cycle and parent constraints are obeyed.
    
    private boolean createEdge(Node source, Node destination){
        
        // Is the source  parent of destination (i.e, there is an inheritance edge)?
        // Or is the source already extnding another class? If so
        // then do not add yet another inheritance edge.
        if(!source.isParent(destination)){
            double r=Math.random();// Random number between 0 and 1.
            if(r<edgeProbability){// No. Should an edge be created?
                String edgeType=getEdgeType(); // Yes.  Hence try to create an edge.
                if(edgeType.equals(Utilities.InheritanceEdgeLabel)){ // Is an inheritance edge to be created? 
                    // Then source node must not be an ancestor of destination.
                    // And source must not be inheriting from any other class.
                    if(!destination.isAncestor(source)&&
                       !(source.checkMultipleInheritance()>0)){
                        return addEdge(source,destination, edgeType);// Add an inheritance or aggregation edge.
                    }
                }else{
                    return addEdge(source,destination, edgeType); // Add a non-inheritance edge.
                }
            }// End of edge probability check.
            // This edge is not generated (random).
            return false;
        }// End of node parent check
        // THis edge is discarded due to existing parent relationship.
        edgesDiscardedParent++;
        return false;
    } // Ends of createEdge()
    
    // Add an edge between source and destination node.
    private boolean addEdge(Node source, Node destination, String edgeType){
        Edge e=new Edge(source, destination, edgeType);
        source.addOutgoingEdge(e);
        if(edgeType.equals(Utilities.AggregationEdgeLabel)||edgeType.equals(Utilities.InheritanceEdgeLabel)){
            if(Utilities.isThereACycle(nodeList)){
                // This edge not included as it would otherwise create an invalid cycle.
                source.removeOutgoingEdge(e);
                edgesDiscardedCycle++;
                return false;
            }
        }
        
        if(edgeType.equals(Utilities.InheritanceEdgeLabel)){
            destination.addParent(source); // Destination node is the parent of the source.
            inheritanceEdges++;
        }
        
        if(edgeType.equals(Utilities.AggregationEdgeLabel)){
            destination.addParent(source); // Destination node is the parent of the source.
            aggregationEdges++;
        }
        edgesCreated++;
        ordData=ordData+edgeType+" "+destination.getLabel()+"\n"; // Add node to output data.
        return true;
    }//End addEdge();
    
    
    private void setEdgeTypeProportions(Scanner s){
        System.out.print("Proportion of inheritance edges (0-1): ");
        double probInheritance;
        // Get inheritance edges proportion.
        try{
            probInheritance=s.nextDouble();
        }catch(Exception e){
            System.out.println("Invalid inpout. Proportion set to 0.1.");
            probInheritance=0.1;
        }
        // Check for valid proportion.
        if(probInheritance<0 || probInheritance>1){
            probInheritance=0.1;
            System.out.print("Sorry but the probability of generating an inheritance edge has been set to 0.1. ");
        }
        inheritanceProbability=probInheritance;
        
        // Now get the probability of generating an aggregation edge.
        double aggregationProbMax=1-probInheritance;
        System.out.print("Proportion of aggregation edges (0-"+aggregationProbMax+")");
        double probAggregation;
        try{
            probAggregation=s.nextDouble();
        }catch(Exception e){
            System.out.println("Invalid input. Proportion set to 0.3.");
            probAggregation=0.3;
        }
        if(probAggregation<0 || probAggregation>aggregationProbMax){
            probAggregation=0.1;
            System.out.print("Sorry but the probability of generating an aggregation edge has been set to 0.1. ");
        }
        aggregationProbability=probAggregation;
        associationProbability=1-(inheritanceProbability+aggregationProbability);
    }
    // The next method generates  a random ORD. It gets the number of nodes (say, n) 
    // in the graph from the user. Then it gets the number of edges (say, e) from
    // the user. It then generates an ORD with n nodes and e edges.
    
    private void generateRandomGraph(){
        double startTime=System.currentTimeMillis();
        Scanner s=new Scanner(System.in);
        
        nodeCount=getNodeCount(s); // Get number of nodes in the graph.
        edgeCount=getGraphDensity(s); // Get number of edges in the graph.
        setEdgeTypeProportions(s);// Get edge type probabilities.
        
        // Initialize node, edge and specific edge counts.
        nodesCreated=0;
        edgesCreated=0;
        inheritanceEdges=0;
        aggregationEdges=0;
        associationEdges=0;
        // Create an empty graph.
        g=new Graph(); 
        // Create and add nodes to the graph.
        // Each node is labelled as N0, N1, N2, etc.
        
        
        for(int i=0; i<nodeCount; i++){
            Node n=new Node("nN"+Integer.toString(i), 0, 0);
            nodeList.add(n);
            g.addNode(n);
        }// End of node creation loop.
        
        // Loop for all nodes randomly connecting them to
        // the remaining nodes. Parent-child nodes
        // are treated especially
        
        
        for(int i=0; i<nodeCount; i++){
            ordData=ordData+"nN"+Integer.toString(i)+"\n"; // Add node to output data.
            for(int j=0; j<nodeCount; j++){
                if(i!=j){
                    // Create edges.
                    boolean discarded=createEdge(nodeList.get(i), nodeList.get(j));
                }
            }// End edge creation loop
            ordData=ordData+"&\n"; // Add node to output data.
        }// End node traersal loop.
        
        System.out.println("\nRandom graph generation completed.");
        
        // Check graph validity.
        if(checkGraph()){
            System.out.println("ORD checks successful.");
        }else{
            System.out.println("ORD is invalid. Not saved.");
            return;
        }
        if(Utilities.isThereACycle(nodeList)){
            cycleExists=true;
            System.out.println("This graph has cycle(s)");
        }else{
            cycleExists=false;
            System.out.println("This graph has no cycle(s)");
        }
        
        int edgesGenerated=inheritanceEdges+aggregationEdges+associationEdges;
        System.out.println("Nodes: "+nodeCount);
        double actualDensity=(double)edgesGenerated/maxEdges;
        System.out.printf("Density requested: %.2f Actual density: %.2f", 
                          density,  actualDensity);
        
        double iProportion;
        if(inheritanceEdges==0||edgesGenerated==0)
            iProportion=0;
        else
            iProportion= inheritanceEdges/(double)edgesGenerated;
        
        double agProportion;
        if(aggregationEdges==0||edgesGenerated==0)
            agProportion=0;
        else
            agProportion= aggregationEdges/(double)edgesGenerated;
        
        double asProportion;
        if(aggregationEdges==0||edgesGenerated==0)
            asProportion=0;
        else
            asProportion= associationEdges/(double)edgesGenerated;
        
        // Generate and print summary.
        
        System.out.println("\nMax possible edges: "+maxEdges);
        System.out.println("Edges requested: "+edgeCount);
        System.out.println("Edges in ORD: "+edgesGenerated);
        System.out.println("Edges discarded to avoid cycle: "+edgesDiscardedCycle);
        System.out.println("Edges discarded due to parent-child relationship: "+edgesDiscardedParent);
        System.out.printf("\tInheritance edges: %d  \tProportion: %.4f ", inheritanceEdges,
                          iProportion);
        System.out.printf("\n\tAggregation edges: %d \tProportion: %.4f ", aggregationEdges,
                          agProportion);
        System.out.printf("\n\tAssociation edges: %d \tProportion: %.4f ", associationEdges,
                          asProportion);
        ordFileName="N"+Integer.toString(nodeCount)+"E"+Integer.toString(edgesGenerated)+".txt";
        
        // Calculate generation time.
        double endTime=System.currentTimeMillis();
        double timeTaken=(endTime-startTime)/(1000);
        String formattedTimetaken=String.format("%.2f", timeTaken);
        
        // Create a header for the ORD data file.
        Formatter f=new Formatter();
        // Format densities.
        String formattedActualDensity=String.format("%.2f", actualDensity);
        String formattedDensity=String.format("%.2f", density);
        String comments="% Object relation diagram (graph)";
        comments=comments+" generated in "+formattedTimetaken+" second(s) on ";
        
        // Get and format date/time.
        
        DateFormat dateFormat = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
        Date date = new Date();
        String today=dateFormat.format(date);
        // Begin creating the header information for the ORD file.
        
        
        // Get cycle information.
        comments=comments+today;
        if(cycleExists)
            comments=comments+"\n% This ORD has cycle(s).";
        else
            comments=comments+"\n% This ORD has no cycle(s).";
        
        // UPdate comments.
        
        comments=comments+"\n% Nodes: "+nodeCount;
        comments=comments+"\n% Density requested: "+formattedDensity+" Actual Density: "+formattedActualDensity;
        comments=comments+"\n% Edges: "+edgesGenerated;
        comments=comments+"\n% Inheritance edges: "+ inheritanceEdges;
        
        comments=comments+"\n% Aggregation edges: "+ aggregationEdges;
        comments=comments+"\n% Association edges: "+ associationEdges+"\n";
        
        
        System.out.printf("\nTotal time: %.4f %s\n",timeTaken," seconds.");
        
        saveORD(comments); // Save the graph in a file.
        
        if(Utilities.graphDisplayDebug)
            Utilities.displayGraph(g);
        
    }// End of generateRandomGraph()
    
    // Check for graph validity. Only
    // a simple inheritance edge check is performed.
    private boolean checkGraph(){
        boolean   graphValid=true;
        Iterator<Node> itr=nodeList.iterator();
        
        
        
        while(itr.hasNext()){
            Node next=itr.next();
            if(!next.checkInheritance()){
                graphValid=false;
                System.out.println("Node "+next.getLabel()+" has an edge coming in from its parent.");
            }
            if(next.checkMultipleInheritance()>1){
                graphValid=false;
                System.out.println("Node "+next.getLabel()+" has more than one parent.");
            }
        }// End of node iterator
        
        return graphValid;
        
    }// End checkGraph();
    
    
    // Write generated ORD into a file.
    private void saveORD(String comments){
        String filePath=fileDestination+ordFileName;
        File f=new File(filePath);
        PrintWriter p=null;
        System.out.println("\n\nSaving ORD in "+ordFileName);
        try{
            p= new PrintWriter(f);
        }catch(Exception e){
            System.out.println("ORD File could not be created. Program terminated prematurely.");
            System.exit(0);
        }
        p.printf("%s", comments);
        p.printf("%s", ordData);
        p.print("\n% End of ORD ordFileName");
        p.close();
        System.out.println("ORD saved in "+filePath);
    }
    
}// End of class CreateRandomORD;

