/*
 * Class Edge is used in the algorithm for generating
 * sequence of classes in integration testing
 * using various algorithms for class order determination. 
 * It is used in class Node to contain edges going out of a node.
 * 
 * Original: July 7, 2011.
 * Latest update: Aug 11-18, 2011
 * 
 * Added masterRootDestinaion to assist with tst order determination in YLT et al. algorithm.
 * Aditya Mathur
 * 
 * Note:
 * 
 * I had forgotten to initialize  masterRootDestination in the constructor. This
 * did not create any problem with the YLT et al. example. But
 * it caused a null pointer error during test order determination
 * when applied to the simple example in  Braind et al.'s paper.
 * */

public class Edge{
    
    Node source; // Source node of the edge (this is redundant though at times useful).
    Node destination; // Destination node of edge.
    Node sccDestination;// Point to root node of an Scc. Does not indicate dependence.
    Node sccSource;// Point to root node of an Scc. Does not indicate dependence.
    Node masterRootDestination; // This is used during test order determination.
    String label; // Edge label. 
    boolean visited; // Used in graph traversal, e.g., Depth First Search.
    boolean frond; // Used in LeTraon et al. and Briand et al. algorithms.
    boolean deleted; // Used in Le Traon algorithm
    int deleteSequence; // Sequence number when this edge was deleted (0 if never deketed).
   
    public static String deleteString="*"; // Used to mark an edge as deleted.
    
    // For the Tai-Daniels algorithm an edge label can be one of the following:
    // eA: Association. eC: Collection (Aggregation) and eI  Inheritance. The prefix e indicates
    // that the label is that of an edge. A prefix of n indicates a node.
    
    // An edge is always associated with a node, its source node.
    //  Hence only a destination node is needed to define an edge. Nevertheless,
    // to make things a bit simple, we also add a source node to each edge.
    
    // The weight of an edge e is =s+d, where s is the number of incoming edges to the
    // source node of e and d is the number of edges outgoing from
    // the destination node of e. [Tai-Daniels refer to source and destination
    // as, respectively, head and tail.]
    
    
       
      
    public Edge(){ 
        source=null;
        destination=null;
        sccDestination=null;
        label="Undefined";  
        visited=false;
        frond=false;
        deleted=false;
        deleteSequence=0;
     
        frond=false;
    }
    
    public Edge(Node s, Node d, String l){ 
        source=s;
        destination=d;
        sccDestination=null;
        masterRootDestination=d; 
        label=l;   
        visited=false;
        frond=false;
        deleted=false;
        deleteSequence=0;
      
    }
    
    public Node getSource(){ // Extract source node.
        return source;
    }
    
   
    
    public Node getDestination(){ // Extract destination node.
        return destination;
    }
    
     public Node getSccDestination(){ // Extract  destination node in another SCC.
        if(sccDestination==null)
            return destination;
        else
            return sccDestination;
    }
     
     public Node getSccSource(){ // Extract  destination node in another SCC.
        if(sccSource==null)
            return source;
        else
            return sccSource;
    }
    
      public Node getMasterRootDestination(){
          return masterRootDestination;
          
      }
  
      public int getDeleteSequence(){
        return deleteSequence;
      }
    
      public boolean isDeleted(){
          return deleted;
      }
      
      public boolean isFrond(){
          return frond;
      }
    
   

     public void markFrond(){
         frond=true;
     }
     
    // This method is used when the label
    // of an edge is changed in the Tai-Daniels algorithm
    // to indicate that it has been deleted from the graph.
    // Thus, the edge remains in the graph except that an asterisk is appended to its label.
    // For example, a label "eA" would become "eA*"
    
    public void delete(){
      if(!deleted){
        this.label=label+deleteString;
        deleted=true;
        }
    }
    public void delete(int sequence){
      if(!deleted){
        this.label=label+deleteString;
        deleted=true;
        deleteSequence=sequence;
        }
    }
    
    public void setDestination(Node n){
        destination=n;
    }
    
     public void setSccDestination(Node n){
        sccDestination=n;
    }
     
      public void setSccSource(Node n){
        sccSource=n;
    }
    
      public void setMasterRootDestination(Node n){
        masterRootDestination=n;
    }
    public String getLabel(){ // Extract label.
        return label;
    }
    
    public void markVisited(){
        visited=true;
    }
    
     public void markNotVisited(){
        visited=false;
    }
     
     public boolean getVisited(){
         return visited;
     }
}