

/*
 * Class CreateGraph is used in the algorithm for generating
 * sequence of classes (nodes in a digraph) in integration testing
 * using Tai-Daniels algorithm. It reads graph 
 * data from a file and creates the Graph class. 
 * 
 * Format of the file containing graph nodes/edges follows.
 * 
 * The file containing the graph must be a plain text file.
 * Such a file can be created using, for example, 
 * Microsoft Word and saving it in plain text format.
 * 
 * Data for each node or edge occupies one line.
 * Lines are separated by CRLF.
 * Each node contains the following information: node label
 * There is exactly one line for each nodel label.
 * 
 * Node-label
 * 
 * The line for each node is followed by as many lines 
 * as there are edges going out of that node.
 * 
 * edge-label destination-node-label 
 * 
 * An & indicates the end of edges for a node. It
 * is placed by iteself on a line. It indeicates the end
 * of all edges for a node. A node label by iteself on a line 
 * and followed on the next line by an & indicates that this node has no edges.
 * 
 * Label of a node must be prefixed by an "n". For example
 * label nB indicates that B is a node. Similarly an edge
 * label must be prefixed by an "e". For
 * example, label eF indicates an edge labelled F. A node can be labelled
 * as Abstract by using the tage tA following the node name.
 * 
 * Example input:
 * nBL
 * eA nB
 * eA nY
 * &
 * B
 * eA nSS
 * eF nE
 * eF nBS
 * &
 * nX
 * tA
 * eI nY
 * &
 * nY
 * &
 * 
 * The above (hypothetical) data is for a graph with  nodes labeled BL, B, SS, E,  BS, and X.
 * Node BL has one outgoing edge, node B has three outgoing edges and
 * the remaining nodes have no outgoing edges. Node X is abstract.
 * 
 * The obvious: Each node must have a label and each edge must have a label.
 * 
 * 
 * July 7, 2011.
 * Latest update: July 28, 2011
 * Aditya Mathur
 * */

import java.util.*;
import java.io.*;

public class CreateGraph{
    
    Graph g;
    private ArrayList <Node> nodeList=new ArrayList<Node>(); // For internal use.
    int nodeCount=0;
    int edgeCount=0;
    int nextLine; // Used for displaying error in input file.
    boolean staticDynamic=true; // true implies dependency is static, otherwise this is false.
    char nodePrefix='n';  // Prefix for node label.
    char edgePrefix='e';  // Prefix for edge  label.
    char nodeTypePrefix='t';  // Prefix for node type.
    char commentStartChar='%'; // Start of a comment in the ORD input file.
    
    
    public void createGraph(String filename, boolean staticOnly) {
        
        // Set node, edge and type prefixes.
      
        Scanner in=null;
        try{
            FileReader f=new FileReader(filename);
            in= new Scanner(f);
        }catch(IOException e){
            System.out.println("File: "+filename+" not found.");
            System.exit(0);
        }// End of try-catch for4 file open 
        g=new Graph(); // Create an empty graph.
        // Read node and edge data.
        nextLine=0; 
        while(in.hasNext()){
          String sourceNodeLabel=in.next(); // Get label of a possibly new node. 
          if(sourceNodeLabel.charAt(0)!=commentStartChar){// Ignore comments.
            nextLine++; // Increment line number of error display 
            checkLabel(sourceNodeLabel, nodePrefix);
            if(!sourceNodeLabel.equals("&")){
                Node sourceNode=isNodeInGraph(sourceNodeLabel);// Check if node already in graph.
                if(sourceNode==null){// If not then 
                    sourceNode=new Node(sourceNodeLabel, 0, 0); // create new node with label and levels (=0).
                    g.addNode(sourceNode); // Add it to the graph if it does not exist.
                    nodeList.add(sourceNode); // Also add it to the local list of nodes.
                    if(Utilities.graphInputDebug)
                      System.out.println("Node: "+sourceNodeLabel.substring(1));
                    nodeCount++;
                }
                getEdges(sourceNode, in ); // Get possible abstract indicator, outgoing edges, update sourceNode.  
            }
          }else{
              String skip=in.nextLine(); // Discard comment line.
              //   System.out.println("Skipped: "+skip); // For debugging
            }// End of comment check
          } // End of while for reading inout from  a file.  
        computeStaticDependence();
        computeInheritanceChains();
        if(!staticOnly)
            computeDynamicDependence();
    }// End of createGraph()
    
    /*
     * getEdges() reads edge label and destination node label.
     * It then creates an edge and adds it to node n.
     * */
    
    private void getEdges(Node sourceNode, Scanner in){
        while (in.hasNext()){
            String edgeLabel=in.next(); // Get edge label or abstract type; it might be an & implying no more edges.
            // Check if this is the end of data related o sourceNode. 
            if(edgeLabel.equals("&")){
                return; // No more outgoing edges for this node (source).
            } 
            
            if(edgeLabel.charAt(0)!=commentStartChar){// Check for start of a comment.
              
            // If not then  an abstract node indicator.
            if(edgeLabel.charAt(0)==nodeTypePrefix){
                sourceNode.setAbstract();
            }else{ // It is not a type identifier, hence check for edge.
                if(edgeLabel.charAt(0)==edgePrefix){
                    String destinationNodeLabel=in.next(); // There is an edge. Get label of destination node.
                    nextLine++; // Update line number.
                    checkLabel(destinationNodeLabel, nodePrefix);
                    
                    // Create new node from the destination label only if
                    // such a node does not exist, else use 
                    // the existing node as the destination.
                    
                    Node destinationNode=createNode(destinationNodeLabel);
                    
                    Edge e=new Edge(sourceNode,destinationNode, edgeLabel);
                    if(Utilities.graphInputDebug){
                      System.out.println("Node-Edge-Node: "+sourceNode.getLabel().substring(1)+"-"+
                                         edgeLabel.substring(1)+"-"+destinationNode.getLabel().substring(1));
                    }
                    sourceNode.addOutgoingEdge(e); 
                    edgeCount++; 
                }else{ // This is an invalid entry.
                  System.out.println("At line: "+nextLine+ " invalid coding of edge, node, or node type: "+
                                     edgeLabel+" Prefix expected: e or t");
                }
            }
            }else{
              String skip=in.nextLine();
              //   System.out.println("Skipped: "+skip); // For debugging
            }
        }// End of edge loop
    } // End of getEdges()    

  /*
   * createNode() creates a node with label
   * l if such a node does not exist.
   */
  
  private Node createNode(String l){
    Node n=isNodeInGraph(l); // Check if there is a node with label l.
    if(n!=null){
      return n; // Yes, it does, hence no need to create it again!
    }
    
    // Node with label l does not exist. Hence create a new node.
    
    Node newNode=new Node(l, 0, 0);
    g.addNode(newNode);
    nodeCount++;
    nodeList.add(newNode);
    if(Utilities.graphInputDebug){
        System.out.println("Node: "+l.substring(1));
    }
    return newNode;
  } // End createNode()
  
  
  // Check if a node with label l is already in the graph constructed so far.
  private Node isNodeInGraph(String l){
    Iterator itr=nodeList.iterator();
    while(itr.hasNext()){
      Node next=(Node)itr.next();
      String nodeLabel=next.getLabel();
      if(l.equals(nodeLabel)){
        return next; // Node in graph, return it.
      }
    }// End of while
    return null; // Node not in graph, return null.
  }// End of isNodeInGraph()
  
  public Graph getGraph(){ // Return the graph created.
    return g;
  }
  
  // THis method checks if the label of a node or an
  // edge is correctly prefixed (n for node an e for edge).
  
  private void checkLabel(String label, char prefix){
      if(label.equals("&"))
             return;
         
      if(!(label.charAt(0)==prefix))
          System.out.println("At line: "+nextLine+ " invalid coding of edge, node, or node type: "+
                             label+" Prefix expected: "+prefix);
      
  }// End of checkLabel()
  
  
  
  
   // Find static dependence for all nodes in the graph.
    private void computeStaticDependence(){
        ArrayList<Node> dependsOn;
        int chainType;
        ArrayList<Node> nodes=g.getNodes();
        Iterator<Node> itr=nodes.iterator();
        while(itr.hasNext()){
            Node n=itr.next();
            g.resetNodeAndEdgeMarks(nodes); // Reset all node and edge marks to not visited.
            dependsOn=new ArrayList<Node>(); // Reset dependsOn.
            chainType=Utilities.staticChain; // We need to findstatic dependencies.
            DFS(dependsOn, n, chainType); // Find them baby!
            n.setStaticDependence(dependsOn);
        }
        
    } // End of updateStaticDependence()
    
    private void  computeDynamicDependence(){
        
        ArrayList<Node> nodes=g.getNodes(); // Get all nodes in the graph.
        ArrayList<Node> destNodesAdded; // Contains list of nodes to which an edge has already been added.
        g.resetNodeAndEdgeMarks(nodes); // Set all nodes to not visited.
        Iterator<Node> itrNode=nodes.iterator(); // Set up node iterator.
        while(itrNode.hasNext()){
            Node chainRootNode=itrNode.next(); // Get the next root node of an inheritence chain.
                ArrayList<Node> inheritanceChain=chainRootNode.getInheritanceChain();
                Iterator <Node> itrChain=inheritanceChain.iterator(); // Set up chain iterator.
                
                while(itrChain.hasNext()){
                    Node chainNode=itrChain.next();// Get next node on the chain.
                     destNodesAdded=new ArrayList<Node>();
                    
                    ArrayList<Edge> chainNodeEdges=chainNode.getIncomingEdges(); // Get incoming edges.
                    Iterator<Edge>  itrEdges=chainNodeEdges.iterator();// Set up edge iterator.
                    while(itrEdges.hasNext()){
                        Edge chainNodeEdge=itrEdges.next(); // Get next incoming edge.
                        if(!chainNodeEdge.getLabel().equals("eI")){// Process only if this is not inheritence. 
                            Node source=chainNodeEdge.getSource(); // Get source of this edge.
                            if(!destNodesAdded.contains(chainRootNode)){
                                source.updateDynamicDependence(chainRootNode); // Add chainRootNode as a dependent node
                                Edge newEdge=new Edge(source, chainRootNode, "eA");
                                source.addOutgoingEdge(newEdge);
                                destNodesAdded.add(chainRootNode); // Mark node an avoid duplicate link.
                            }// End of chainRootNode added test
                        }     
                    }// End of edge iteration  
                }// End of chain node iteration.
       
        }// End of iterator chain.
    }// End of  computeDynamicDependence();
    
    
     // Find inheritance chains for all nodes in the graph.
     private void computeInheritanceChains(){
        ArrayList<Node> dependsOn;
        int chainType;
        ArrayList<Node> nodes=g.getNodes();
        Iterator<Node> itr=nodes.iterator();
        while(itr.hasNext()){
            Node n=itr.next();
            g.resetNodeAndEdgeMarks(nodes); // Again reset all node and edge marks to not visited.
            dependsOn=new ArrayList<Node>(); // Reset dependsOn.
            chainType=Utilities.inheritanceChain; // Now we need to determine inheritance chain.
            DFS(dependsOn, n, chainType); // Find them baby!
            n.setInheritanceChain(dependsOn);
        }
        
    } // End of computeInheritanceChains()
     
       // The following method does a depth first search 
    // of a graph starting at a given node.  
    // Static (chain) dependence considers all edge types.
    // Inheritence chains consider only the "I" type edges that correspond
    // to the inheritance relation.
    
    
   private void DFS(ArrayList<Node> dependsOn, Node n, int chainType){
        n.markVisited();
        ArrayList<Edge> edges=n.getOutgoingEdges();
        
        int numEdges=edges.size();
        for (int i=0; i<numEdges; i++){
            Edge e=edges.get(i);
            if(!ignoreEdge(e, chainType)){
                if(!e.getVisited()){
                    Node nextNode=edges.get(i).getDestination();
                    if(nextNode!=null){
                        if(!nextNode.getMark()){
                            edges.get(i).markVisited();
                            Utilities.debug(Utilities.dependenceCreationDebug, "--> "+nextNode.getLabel().substring(1));
                            dependsOn.add(nextNode);
                            DFS(dependsOn, nextNode, chainType);
                        }else{
                             Utilities.debug(Utilities.dependenceCreationDebug, " Back Edge: "+n.getLabel().substring(1)+"-"+
                                               edges.get(i).getLabel().substring(1)+"-"+
                                               edges.get(i).getDestination().getLabel().substring(1));
                        }
                    }// End of node check
                } // End of edge visited check
            }// End of chainType check.
        } // End of for  
    }// End of DFS() 
   
   private boolean ignoreEdge(Edge e, int chainType){
       if(chainType==Utilities.staticChain)
           return false; // Do not ignore any edge.
       if(chainType==Utilities.inheritanceChain && e.getLabel().equals("eI"))
           return false;
       else 
           return true;
   }// End of ignoreEdge()
   
   
   // Get node count.
   public int getNodeCount(){
       return nodeCount;
   }
   
   // Get edge count.
   public int getEdgeCount(){
       return edgeCount;
   }
}// End of class CreateGraph