

/*
 * This class implements the algorithm developed by
 * Yves Le Traon, Thierry Jron, Jean-Marc Jzquel, and Pierre Morel
 * for generating a reasonable test order 
 * for class integration testing. The classes are input
 * to the program in the form of a graph where the nodes represent the classes
 * to be tested and the edges represent the relationships among the classes.
 * 
 * Unlike in Tai-Daniels algorithm, there are no edge labels. Cycles may or may not exist.
 * And it does not matter what relationship(s) create a cycle.
 *
 * 
 * Reference:
 * 
 * Efficient Object-Oriented Integration and Regression Testing,
 * Yves Le Traon, Thierry Jron, Jean-Marc Jzquel, and Pierre Morel
 * IEEE TRANSACTIONS ON RELIABILITY, VOL. 49, NO. 1, MARCH 2000
 * pp 12-25.
 * 
 * This class coding:
 * Original: July 7, 2011
 * Latest update: Aug 12, 2011. Jan 10, 2012
 * 
 * Major update: Jan 11, 2012: I had an incorrect understanding of the algorithm.
 * I was using the unmodified algorithm of Treon et al., rather than the modified
 * one that uses weights to decide on the new root of the SCC right from the first 
 * set of SCCs. 
 * 
 * The stubs created depend heavily on the sequence in which the nodes are vsited in Tarjan's algorithm
 * to find the strongly connected components.
 * 
 * Notes:
 * 
 * 1. The first complete version of this tool passed tests on July 27, 2011.
 * 2. The second version began on July 28, 2011. It includes abstract classes, static as well
 * as dynamic dependencies, and addition information in test order display.
 * 3. This version implements the Le Traon et al algorithm that
 *  is quite different from that of Tai-Daniels.
 * 
 * Bugs found and fixed:
 * 
 * Aug 5, 2011, and many many more.
 * 
 * Error: The root of an SCC with only two nodes was not being marked inside tarjan()
 * as a frondRoot node. This is because tarjan() was marking a node as a root only for 
 * SCCs that have cycles, i.e., that have more than 1 node. But aan SCC with only two
 * nodes goes into torjan(), during a recursive call, AFTER the frond edge has been removed.
 * Thus tarjan() find no cycles in this Scc as the cycle is removed by the removal
 * of the frond edge.
 * 
 * Correction: An SCC with only two nodes is processed right after its frond edge is detected and removed.
 * The correction is inside setFrond() method.
 * 
 * Aug 12, 2011: Another error found when tried the implementatio n on the Tai/Daniel's example. In this
 * case one of the SCC's before the recursive call has more than 1 node but has no
 * cycles. It was being considered as one with cycles. But when tarjan() is applied again on this scc,
 * it splits into 4 singleton scc's (each with exactly one node). Because there is no cycle in any of these four nodes, 
 * no stub was crerated. Thus, for node SS, there was no stib created. This bug was fizxed by updating
 * the logic for finding cycle free sccs.
 * 
 * 
 * Author: Aditya Mathur
 * 
 */


/*
 * Abrreviations used:
 * 
 * SCC: Strongly connected component. scc: same but used as the name of an object.
 * 
 * Note: The first correct implementation of YLTet al. algoruthm was available at 3:21pm on August 11.
 * It gave the correct answer to the simple example in YLT et al. paper.
 */


import java.util.*;
import java.io.*;

public class LeTraonAlgorithm {
  
  // Object Relation Diagram as a graph. Attributes of nodes in this
  // graph are modified as the Tai-Daniels algorithm progresses.
  Graph g; 
  
  int maxMajorLevel=0; // Maximum major level number assigned in the graph.
  int [] maxMinorLevel; // Maximum minor level number assigned in the graph to each level.
  int majorMinor; // =0 implies major, =1 implies minor. Used in deciding whether to assign major or minor level.
  String deletedEdgeLabel="eA"+Edge.deleteString; // Label of an association edge that has been removed from g.
  ArrayList<ArrayList<Node>> testSequence; // Sequence in which classes are to be tested.
  ArrayList<Stub> stubs=new ArrayList<Stub>(); // Stubs needed.
  ArrayList<Retest> retestNodes; // Nodes to be retested.
  ArrayList<Node> testOrder=new ArrayList<Node> (); // Class tst order is recorded in this list.
  
  CreateGraph cg; // This object creates a graph from ndoes and edges input from a file.
  
  //Next flag indicates whether or not dynamicstatic dependencies are to be considered.
  // Set it to false to include dynamic dependencies.
  boolean staticOnly=true; 
  
  boolean cycleExists=true;
  public static boolean recursiveCall; // Used in findCycles() to start assigning stubs only after the first recursive call.
  public static int counter=0; // To assign to nodes when finding an during recursive call to tarjan().
  ArrayList<ArrayList<Node>> scc; // Holds SCC's found after the first call to Tarjan's algorithm.
  ArrayList<ArrayList<Node>> originalSccs;
  ArrayList<Scc> sccInitial;
  
  
  Stack <ArrayList<ArrayList<Node>>> sccStack;
  
  static ArrayList<String> ordFiles; // List of ORD file names.
  
  static String ordFolder="../TestGraphs/SecondEditionExperiments"; // This is where the ORDs are located.
  // static String testOrderStatistics="../TestGraphs/testOrderStatistics";
  static String testOrderStatistics="../TestGraphs/testOrderStatisticsFullyConnectedORDs";
  
  static String testData; // Stub count etc is appended to this file.
  
  
  // The constructor creates a graph g from an input text file.
  
  public LeTraonAlgorithm(String fileName)throws IOException{
    // TD algorithm Step 1: Read a graph from a file.
    System.out.println("Input graph from file: "+fileName);
    cg=new CreateGraph();
    cg.createGraph(fileName, staticOnly);
    
    System.out.println("--Graph successfully input.");
    // cg.createGraph("sccTest1NoEdges.txt");
    // cg.createGraph("sccTest2SymmetricTwoComponents.txt");
    
    //cg.createGraph("sccTest3ThreeComponents.txt");
    // cg.createGraph("sccTest4OneNodeDegenerate.txt");
    //cg.createGraph("sccTest5TwoNodesDegenerate.txt");
    g=cg.g; // Save the graph locally.
    
  }// End of LeTraonAlgorithm constructor
  
  
  
  // The following recursive method detects 
  // (a) strongly connected components in graph g,
  // (b) frond edges as defined by LeTraon et al, and (c) removes frond edges.
  
  public void findCycles(ArrayList<ArrayList<Node>> scc){  
    
    Utilities.debug( Utilities.frondDebug, "\n# of SCCs to be processed: "+scc.size());
    
    // Check if there is any SCC in the list (scc) with cycles.
    cycleExists=doesACycleExist(scc);
    if(!cycleExists){
      if(Utilities.sccDebug){
        Utilities.displayAllScc(scc);
      }
      processSingleton(scc); // There might be a stub in one of the SCCs.
      return;
    }
    
    // Yes, there is at least one non-trivial SCC. 
    // Break cycles in each non-trivial SCC.
    Iterator <ArrayList<Node>> itr=scc.iterator();
    int nextSccToBeProcessed=0;
    
    while(itr.hasNext()){
      ArrayList<Node> nextScc=itr.next();// Extract next SCC from the list (scc).
      nextSccToBeProcessed++; // Update its counter (for debugging)
      ArrayList<ArrayList<Node>> newSccList=new  ArrayList<ArrayList<Node>>();
      
      if(Utilities.frondDebug){
        newSccList.add(nextScc); // This is to ensure parameter compliance.
        System.out.println("\nNext SCC to be processed:"+nextSccToBeProcessed);
      } // End of debug check.
      
      if(nextScc.size()>1){// It can have cycles only if the nextScc has more than one node. 
        Graph newGraph=new Graph(nextScc);// Create a graph out of it (again, for parameter type compliance).
        
        // Create an object to compute SCCs in newGraph.
        // Apply Tarjan' algorithm to newGraph and get SCCs (frond edges are marked).
        
        SccTarjan sccAlg=new SccTarjan(newGraph); // This is a recursive call to
        scc=sccAlg.tarjan(); //  tarjan().
        
        
        if(Utilities.sccDebug){
          System.out.println("\nSCCs after next call to Tarjan's algorithm:"+"\n--------");
          Utilities.displayAllScc(scc);
          System.out.println("\n--------");
        }
        
        
        // Process all SCCs. 
        processSccs(scc); 
        if(Utilities.testOrderDebug){
          Utilities.leTraonTestSequence(scc);
        }
        findCycles(scc);// Recurse on the SCCs created.
      }// End of singleton scc check
      Utilities.debug( Utilities.frondDebug, "\n  SCC "+nextSccToBeProcessed + " processed:" );
    }// End of node ieration loop.   
  }// End of findCycles
  
  
  // The next method checks if a component of a graph  has any cycles in it.
  // The component (scc)is input as a list of nodes.
  
  private boolean doesACycleExist(ArrayList<ArrayList<Node>> scc){
    
    // A cycle exists if there is any element in scc with more than 1 node
    
    Iterator<ArrayList<Node>> itrScc=scc.iterator();
    while (itrScc.hasNext()){
      ArrayList<Node> anScc=itrScc.next();
      int anSccSize=anScc.size();
      if(anSccSize>1){
        Utilities.debug(Utilities.cycleCheckingDebug, 
                        "\n\tThere are "+scc.size()+" SCCs."+ 
                        " \n\tThere may be an SCC with a cycle as at least one SCC has more than one node. ");
        return true; 
      }
    }
    Utilities.debug(Utilities.cycleCheckingDebug, "\n     There is no cycle.");
    return false;
  }// End of cycles()
  
  // Process a list of SCCs at least one of which contains exactly one node.
  // One of these must be a root and is marked as a stub. If all SCCs have exactly one 
  // node then one of these must be a root and should be marked as a stub.
  private void processSingleton(ArrayList<ArrayList<Node>> scc){
    Utilities.debug(Utilities.frondDebug, "\nProcess SCCs with no cycles.");
    Node singleton;
    Node makeThisStub=null;
    Iterator<ArrayList<Node>> itr=scc.iterator();
    boolean allSizeOne=true; // Assume that all have exactly one node.
    while(itr.hasNext()){
      ArrayList<Node> anScc=itr.next();
      singleton=anScc.get(0);
      if(singleton.isRoot()){
        makeThisStub=singleton;
      }
      if(anScc.size()>1){
        allSizeOne=false;
        break; // Get out of this loop.
      }// End size check.
    }// End scc iteration loop.
    if(allSizeOne){
      if(makeThisStub!=null){
        makeThisStub.markStub();
      }else{
        System.out.println("** Program error (processSingleton/LeTraonAlgorithm). All SCCs singleton but no root.");
      }
    }
  }// End of process singleton.
  
  // Within each scc identify the node to be selected based on frond-based weights.
  
  private void processSccs(ArrayList<ArrayList<Node>> scc){
    
    Iterator< ArrayList<Node>> itr=scc.iterator(); // Set node iterator
    Node selected=null;
    int index=0;
    while(itr.hasNext()){
      ArrayList<Node> anScc=itr.next();
      
      if(anScc.size()>1){   
        //Update node weights, select a node, and delete its incoming edges.
        if(Utilities.frondDebug){
          System.out.println("Next (sub) Scc to be processed: ");
          Utilities.displayAnScc(anScc);
        }
        Node root=updateAnSccFrondWeights(anScc, index); 
        deleteIncomingEdges(root, anScc); // Remove  edges coming into the selected root node.
        index++;
      }// End of SCC iterator loop.
    }// End of SCC size check.
  }// End of processSccs()
  
  
  
  private void updateAllSccFrondWeights(ArrayList<ArrayList<Node>> scc){
    Iterator <ArrayList<Node>> itr=scc.iterator();
    Utilities.debug(Utilities.frondDebug, "Process initial SCCs.");
    
    int index=1;
    while(itr.hasNext()){
      ArrayList<Node> anScc=itr.next(); // Get next SCC
      Utilities.debug(Utilities.frondDebug, "\nUpdate weights in Scc "+ index);
      if(anScc.size()==1){ // Ignore if it is trivial.
        Utilities.debug(Utilities.frondDebug, "\tScc is trivial and hence ignored.");
      }else{ // Process only non-trivial Sccs.
        
        
        Node root=updateAnSccFrondWeights(anScc, index); // Update weights and select a root.
        deleteIncomingEdges(root, anScc); // Remove  edges coming into the selected root node.
        
        // For an SCC with only 2 nodes,
        // we have broken the cycle and hence the root must be selected as a stub.
        if(anScc.size()==2){
          root.markStub();    
          System.out.println("Node selected and marked stub: "+root.getLabel().substring(1));
        } 
      }// End of an SCC processing.
      index++;
    }// End of scc loop
  }// End of updateAllSccFrondWeights
  
  
  
  // Mark frond edges in component scc. Then find the weight 
  // of each node (=sum of incoming and outgoing frond edges).
  // Then select node with the highest weight. Finally delete all
  // edges coming in to the selected node
  
  private Node updateAnSccFrondWeights(ArrayList<Node> scc, int index){
    
    //  If the size of this SCC is 1 and the only node
    // in it is a root (from a previous SCC), then this must be
    // marked as a stub because there will be no more cycles in this scc.
    // Note this was a bug in the Aug 10 version of the implementation that
    // was discovered when the Tai-Daniels's example was input.
    if(scc.size()==1){
      Node singleton=scc.get(0);
      if(singleton.isRoot()){
        singleton.markStub();
      }else
        return null;
    }
    
    // Update weight of each node. The weight is
    // the sum of all imcoming and outgoing frond edges.
    // An edge going to its ancestor within the SCC being processed is frond.
    // [January 14, 2012]
    Iterator<Node> itrNode=scc.iterator(); 
    Node root=null; // This is the current root of the scc. 
    
    while(itrNode.hasNext()){
      // Node root=getRoot(scc);
      Node nextNode=itrNode.next(); // Get next node in this SCC      
      if(nextNode.isRoot()){
        root=nextNode;
      }
      updateFrondWeight(nextNode, scc); // Update its weight.
      
    }// End of node iterator within an SCC. All weigths computed.
    
    if(Utilities.frondDebug){ 
      Utilities.displayAnScc(scc);
    }
    
    Node newRoot=selectVertex(scc); // Find in scc the vertex with maximum weight.
    Utilities.debug(Utilities.frondDebug, "Node "+newRoot.getLabel().substring(1)+
                    " selected as root/stub."+" Frond weight: "+newRoot.getFrondWeight());
    newRoot.markStub(); // make it as a stub, and 
    newRoot.markRoot();
    resetSccRoot(newRoot); // Reset current root of the original Scc.
    
    if(!newRoot.equals(root)){
      root.unMarkRoot();// Unmark the current root of the SCC as it is no more a root.
      newRoot.markRoot();// Mark as the new root.
      resetSccRoot(newRoot); // Reset current root of the original Scc.
      Utilities.debug(Utilities.frondDebug, "Node "+root.getLabel().substring(1)+ " removed as a root.");
    }
    return newRoot;
    
  } // End of updateFrondWeights()
  
  // Update frond count for all nodes in component sccc that have 
  // traversal index higher than that of node n.
  
  public void updateFrondWeight(Node n, ArrayList<Node> scc){  
    // Get all incoming edges that (a) have their source within  scc
    // and (b) n is the ancestor of source.
    Iterator<Edge> itr=n.getIncomingEdges(scc).iterator();
    int traversalIndex=n.getTraversalIndex();
    boolean frondFound=false;
    if(Utilities.frondDebug){
      System.out.println("Frond edges to node "+n.getLabel().substring(1)+"\n------");
    }
    n.setFrondWeight(0); // Initialize frond wight.
    while(itr.hasNext()){
      Edge e=itr.next();      
      Node source=e.getSource();
      
      int sourceNodeIndex=source.getTraversalIndex();
      if(sourceNodeIndex>traversalIndex){
        e.markFrond();
        n.incFrondWeight();
        source.incFrondWeight();
        frondFound=true;
        if(Utilities.frondDebug){ 
          Utilities.displayEdge(e,false);
        }   
      }
      
    }// End of edge iterator.
    if(Utilities.frondDebug){
      if(!frondFound){
        System.out.println("\tNone.");
      }
      System.out.println("\n------");
    }
    
  }// End of updateFrondWeight()
  
  private Node selectVertex(ArrayList<Node> scc){
    
    ArrayList<Node> rList=new ArrayList<Node>();
    rList=Utilities.reverse(scc);
    Iterator<Node> itr=rList.iterator(); // Set node iterator
    int maxWeight=0; // Initialize max weight.
    Node selected=null;
    while(itr.hasNext()){
      Node inScc=itr.next();// Get next node from the scc
      int w=inScc.getFrondWeight();
      if(w>maxWeight){
        maxWeight=w;
        selected=inScc;
      }
      
    }// End of node selection loop.
    return selected;
    
  }// End of selectMaxWeightFrondVertex
  
  // Delete all edges incoming into the selected node n in an SCC.
  // Doing so will remove some cycles but not necessarilly all.
  
  private void  deleteIncomingEdges(Node n, ArrayList<Node> scc){
    
    if(n!=null){
      ArrayList<Edge> inComingEdges=n.getIncomingEdges(scc);
      for(int i=0; i<inComingEdges.size(); i++){ 
        Edge next=inComingEdges.get(i); // Extract the next edge.
        //      if(recursiveCall)
        next.delete();
        Utilities.debug(Utilities.cycleBreakingDebug, "    Edge deleted: "+
                        Utilities.stringifyEdge(next, false));
      }// End of edge iteration
    }// End of null node check.
  }// End of deleteIncomingEdges()
  
  
  private Node getRoot(ArrayList<Node> scc){
    Node root;
    if(!recursiveCall){
      root=getRootInInitialScc(scc);
    }else{
      root=selectVertex(scc);     
    }
    return root;
  }// End of getRoot()
  
  private Node getRootInInitialScc(ArrayList<Node> scc){
    Iterator<Node> itr=scc.iterator();
    while(itr.hasNext()){
      Node n=itr.next();
      if(n.isRoot())
        return n;
    }// End of node iteration.
    return null; // Root not found; program error.  
  }// End of getRootsInInitialScc()
  
  private void unMarkRoots(Graph g){
    ArrayList<Node> nodes=g.getNodes();
    Iterator <Node> itr=nodes.iterator();
    while(itr.hasNext()){
      Node n=itr.next();
      if(!n.isStub())
        n.unMarkRoot();
    }
  }// End of  unMarkRoots()
  
  
  // Reset the current root node in the initial Scc.
  private void resetSccRoot(Node n){
    int sccIndex=n.getSccIndex(); // Get the index of the initial Scc to which this node belongs
    sccInitial.get(sccIndex-1).setCurrentRoot(n); // Reset the current root of the Scc.
    
  }// End of resetSccRoot()
  
  
  // Edges coming from the root of the graph into
  // the intial Sccs must now point to the current root of the Scc.
  
  private void resetEdgesIntoInitialSccs(){
    // Find all roots of trees in the forest (the master roots). It is likely that
    // when using this algorithm there will be only one
    // master root.
    ArrayList<Node>noIncomingEdges=g.getNodesNoIncomingEdges(); 
    Iterator<Node> itrNode=noIncomingEdges.iterator(); 
    while(itrNode.hasNext()){
      Node root=itrNode.next();
      ArrayList<Edge> edges=root.getOutgoingEdges();
      Iterator<Edge> itrEdge=edges.iterator();
      while(itrEdge.hasNext()){
        Edge nextEdge=itrEdge.next(); // Get next outgoing edge from the root of the tree.
        Node dest=nextEdge.getDestination(); // Get its destination node.
        int hostIndex=dest.getSccIndex();// Get the host Scc of the destination node.
        Node currentRoot=sccInitial.get(hostIndex-1).getCurrentRoot(); // Get current root of the host Scc
        // Check if the root is pointing to the current Scc root. If not then reset the destination node
        // of the tree root.
        if(!dest.equals(currentRoot)){ 
          nextEdge.setMasterRootDestination(currentRoot);
          Utilities.debug(Utilities.topologicalSortDebug, "Destination of "+root.getLabel().substring(1)+
                          " reset to "+ currentRoot.getLabel().substring(1)+ " in Scc "+ hostIndex+ 
                          "\nEdge: "+Utilities.stringifyEdge(nextEdge, false));
          
        }      
      }// End of edge iteration for a root of the tree.
      
    }// End of iteration of roots of the forest.
    
  }// End resetEdgesIntoInitialSccs()
  
  
  
  
  // Program to test the implementation of TD algorithm.
  
  public static void main(String[] args) throws IOException{
    
    
    
    Utilities.TaiDaniels=false; // This is Yves Le Traon et al. and not Tai-Daniels algorithm.
    
    
    // LeTraonAlgorithm l=new LeTraonAlgorithm("../TestGraphs/LabichExample2000.txt");
    //LeTraonAlgorithm l=new LeTraonAlgorithm("../TestGraphs/BriandExample2003.txt");
    
    // LeTraonAlgorithm l=new LeTraonAlgorithm("../TestGraphs/BriandExample2003V3.txt");
    // LeTraonAlgorithm l=new LeTraonAlgorithm("../TestGraphs/BriandExample2003V2.txt");
    // LeTraonAlgorithm l=new LeTraonAlgorithm("../TestGraphs/BriandExample2003V3.txt");
    // LeTraonAlgorithm l=new LeTraonAlgorithm("../TestGraphs/BriandExample2003V4.txt");
    // LeTraonAlgorithm l=new LeTraonAlgorithm("../TestGraphs/BriandExample2003V5.txt");
    // LeTraonAlgorithm l=new LeTraonAlgorithm("../TestGraphs/BriandExample2003V6.txt");
    
    // LeTraonAlgorithm l=new LeTraonAlgorithm("../TestGraphs/fullyConnected4Nodes.txt");
    // LeTraonAlgorithm l=new LeTraonAlgorithm("../TestGraphs/fullyConnected6Nodes.txt");
    
    // LeTraonAlgorithm l=new LeTraonAlgorithm("../TestGraphs/MinimalRetestExample.txt");
    
    //LeTraonAlgorithm l=new LeTraonAlgorithm("../TestGraphs/MinimalRetestExample2.txt");
    // LeTraonAlgorithm l=new LeTraonAlgorithm("../TestGraphs/MinimalRetestExample3.txt");
    // LeTraonAlgorithm l=new LeTraonAlgorithm("../TestGraphs/MinimalRetestExample4.txt");
    // LeTraonAlgorithm l=new LeTraonAlgorithm("../TestGraphs/MinimalRetestExample5.txt");
    // LeTraonAlgorithm l=new LeTraonAlgorithm("../TestGraphs/MinimalRetestExample6.txt");
    // LeTraonAlgorithm l=new LeTraonAlgorithm("../TestGraphs/MinimalRetestExample7.txt");
    // LeTraonAlgorithm l=new LeTraonAlgorithm("../TestGraphs/BriandExample2003OneEdgeRemoved.txt");
    // LeTraonAlgorithm l=new LeTraonAlgorithm("../TestGraphs/BriandExample2003TwoEdgesRemoved.txt");
    // LeTraonAlgorithm l=new LeTraonAlgorithm("../TestGraphs/BriandExample2003ThreeEdgesRemoved.txt");
    // LeTraonAlgorithm l=new LeTraonAlgorithm("../TestGraphs/BriandExample2003OneEdgeOneNodeRemoved.txt");
    // LeTraonAlgorithm l=new LeTraonAlgorithm("../TestGraphs/BriandExample2003OneEdgeTwoNodesRemoved.txt");
    
    // Files for experiments for the book. second edition. 
    // Set 1
    
    // LeTraonAlgorithm l=new LeTraonAlgorithm("../TestGraphs/TaiDanielsGraph.txt");
    // LeTraonAlgorithm l=new LeTraonAlgorithm("../TestGraphs/LeTraonExampleV1.txt");
    // LeTraonAlgorithm l=new LeTraonAlgorithm("../TestGraphs/BriandExample2003.txt");
    
    // Set 2
    
    //  LeTraonAlgorithm l=new LeTraonAlgorithm("../TestGraphs/SecondEditionExperiments/LeTraonExampleV2.txt");
    // LeTraonAlgorithm l=new  LeTraonAlgorithm("../TestGraphs/SecondEditionExperiments/N15E42.txt");
    // LeTraonAlgorithm l=new  LeTraonAlgorithm("../TestGraphs/SecondEditionExperiments/N15E78.txt");
    // LeTraonAlgorithm l=new  LeTraonAlgorithm("../TestGraphs/GraphN15E99.txt");
    LeTraonAlgorithm l=new  LeTraonAlgorithm("../TestGraphs/SecondEditionExperiments/N15E134.txt");
    
    // Set 3
    
    // LeTraonAlgorithm l=new  LeTraonAlgorithm("../TestGraphs/SecondEditionExperiments/N100E1452.txt");
    // LeTraonAlgorithm l=new  LeTraonAlgorithm("../TestGraphs/SecondEditionExperiments/N100E3044.txt");
    // LeTraonAlgorithm l=new  LeTraonAlgorithm("../TestGraphs/SecondEditionExperiments/N1004581.txt");
    // LeTraonAlgorithm l=new  LeTraonAlgorithm("../TestGraphs/SecondEditionExperiments/N100E5925.txt");
    
    
    // TJJM algorithm. Find cycles and break them.
    // First call to Tarjan's algorithm. [Step 1 of Yves Le Traon et al. algorithm.]
    // Set up loop to apply the TD test order generation algorithm.
    
    boolean lastCall=false; // Last ORD processed?
    
    System.out.println("\n-------------Begin Test Order Determination using TJJM algorithm. V1.0, July 29, 2013. ");
    
    System.out.println("\nFind strongly connected components.");
    
    l.recursiveCall=false;
    SccTarjan sccAlg=new SccTarjan(l.g); // Create object to compute SCCs in graph g.
    ArrayList<ArrayList<Node>> scc=sccAlg.tarjan(); // Apply Tarjan' algorithm to get SCCs.
    
    // Save the initial SCC for future reference.
    l.sccInitial=sccAlg.getInitialScc();
    
    
    System.out.println("\nSCCs after first call to Tarjan's algorithm:"+"\n--------");
    if(Utilities.frondDebug){
      Utilities.displayScc(l.sccInitial);
    }
    System.out.println("\n--------");
    
    l.updateAllSccFrondWeights(scc);
    
    // Utilities.displayAllScc(scc);
    
    
    // Now apply Tarjan's algorithm recursively on each noin-trivial Scc. 
    // [Step 2 of YTL et al. algorithm.]
    
    System.out.println(" ------  Start: Recursive call to find additional Sccs.");
    l.recursiveCall=true;
    l.findCycles(scc);
    System.out.println("-----  End: Recursive call.\n");
    
    if(Utilities.frondDebug){
      Utilities.displayScc(l.sccInitial);
    }
    
    
    l.unMarkRoots(l.g);
    
    
    if(Utilities.sccDebug){
      Utilities.displayScc(l.sccInitial);
    }
    //  l.resetEdgesIntoInitialSccs();
    System.out.println("\nUpdated Graph:");
    if(Utilities.frondDebug){
      Utilities.displayGraph(l.g, true);
    }
    
    // YLT algorithm final step to generate test sequence for classes.
    
    l.testOrder=Utilities.findTestOrder(l.g);
    
    Utilities.displayTestOrder(l.g, l.testOrder);
    double endTime=System.currentTimeMillis();
    
    System.out.println("\nEnd of TJJM algorithm.");
    
    
  }// End of main()
}// End of class LeTraonAlgorithm.

