
/*
 * Class Node used in the algorithm for generating
 * sequence of classes in integration testing
 * using Tai-Daniels algorithm. It is used to create nodes in
 * a Graph.
 * 
 * frond edges and their destination nodes are needed when applying
 * the Le Traon et al algorithm.
 * 
 * July 7, 2011. 
 * Updates: July 28, 2011. August 11, 2011
 * Aditya Mathur
 * */

import java.util.*;
public class Node{
    
    ArrayList<Edge> incomingEdges=new ArrayList<Edge>(); // Contains all incoming edges to this node.
    ArrayList<Edge> outgoingEdges=new ArrayList<Edge>(); // Contains all outgoing edges from this node.
    ArrayList<Node> staticDependence=new ArrayList<Node>(); // Transtive closure of static "depends" relation.
    ArrayList<Node> dynamicDependence=new ArrayList<Node>(); // Transtive closure of dynamic "depends" relation.
    ArrayList<Node> inheritanceChain=new ArrayList<Node>(); // Transtive closure of dynamic "depends" relation.
    ArrayList<Node> stubsUsed=new ArrayList<Node>(); // List of nodes whose stubs are needed to test this node.
    String label; // Node label (prefixed with an n, e.g. nXY, but only XY displayed). 
    
    int levelMajor; // Major level assigned by Tai-Daniels algorithm
    int levelMinor; // Minor level assigned by Tai-Daniels algorithm
    int incomingEdgeCount=0; // Number of incoming edges
    int incomingEdgeNotDeletedCount=0; // Number of incoming edges that have not been deleted by the algorithm
    int outgoingEdgeCount=0; // Number of outgoing edges
    
    boolean visited; // Indicates whether or not this node is visited during a graph traversal.
    
    // Next variable used in Le Traon et al. algorithm to denote the sequence number
    // in which this node was traversed while finding SCCs during a recursive call.
    int traversalIndex=0; 
    
    // The next index keeps track if the initial Scc to which the node belongs.
    int sccIndex=0; // 
    
    boolean stub;// True means a stub for this class exists, false: does not exist.
    boolean tested; // True means this node (i.e. the corresponding class) has been tested.
    boolean isAbstract=false; // Tue if this node represents an abstract class, false otherwise.
    int frondWeight;// This is used in LaTraon algorithm to count the number of incoming/outgoing frond edges
    boolean rootNode; // Set to true if this the root of a strongly connected component.
    boolean frondRoot; // Set to true if this node might have an incoming frond edge.
    boolean masterRoot;// True implies that this node is the root of a tree in the forest.
    
    // Attributes for use in Tarjan's Strongly Connected Component finding algorithm
    int index; 
    int lowLink;
    
    public Node(){
        visited=false;
        label=" Undefined";   
        levelMajor=0;
        levelMinor=0;
        incomingEdgeCount=0;
        incomingEdgeNotDeletedCount=0;
        outgoingEdgeCount=0;
        frondWeight=0;
        rootNode=false;
        frondRoot=false;
        masterRoot=false;
        stub=false; // No stub created for this node (also a class).
        index=-1; // // Implies that this index is undefined as in Tarjan's algorithm.
        
    }
    
    public Node(String l, int levMajor, int levMinor){ // Create an empty node.
        visited=false;
        label=l;   
        levelMajor=levMajor;
        levelMinor=levMinor;
        incomingEdgeCount=0;
        incomingEdgeNotDeletedCount=0;
        outgoingEdgeCount=0;
        frondWeight=0;
        rootNode=false;
        frondRoot=false;
        masterRoot=false;
        stub=false; // NO stub created for this node (also a class).
        index=-1; // Implies that this index is undefined as in Tarjan's algorithm.
    }
    
    public void addOutgoingEdge(Edge e){ // Add an outgoing edge to this node.
        outgoingEdges.add(e);   
        outgoingEdgeCount++;
        Node dest=e.getDestination();
        dest.addIncomingEdge(e); 
    }// End of addEdge()
    
    public void addIncomingEdge(Edge e){ // Add an incoming edge to this node.
        incomingEdges.add(e);   
        incomingEdgeCount++; 
        incomingEdgeNotDeletedCount++; // Unless edges are deleted this is the same as incomingEdgeCount.
    }
    
    // The next method is invoked when an edge is removed using one of 
    // the test order determination algorithms.
    public void edgeDeleted(){
        incomingEdgeNotDeletedCount--;
    }
    
    public void removeOutgoingEdge(Edge e){ // Remove an edge from the node.
        outgoingEdges.remove(e);  
        outgoingEdgeCount--;
        Node dest=e.getDestination();
        staticDependence.remove(dest);
    }
    
    
    public ArrayList<Edge>  getOutgoingEdges(){ // Return all outgoing edges    
        return(outgoingEdges);
    }
    
    public ArrayList<Edge>  getIncomingEdges(){ // Return all incoming edges   
        return(incomingEdges);
    }
    
     // Return all incoming edges  that are emanating from within the component scc.
     public ArrayList<Edge>  getIncomingEdges(ArrayList<Node> scc){  
         ArrayList<Edge> returnEdgeList=new ArrayList<Edge>();
         Iterator<Edge> itr=incomingEdges.iterator();
         
         
         while(itr.hasNext()){
             Edge e=itr.next();
             Node n=e.getSource();
             if(scc.contains(n)){ // Source of this edge must be in scc.
                 returnEdgeList.add(e);
                 }
         }// End of edge iteration loop.
          return(returnEdgeList); // This might be an empty list if no edge exists that satisfies the condition
    }// End of getIncomingEdges()
     
     // Return all incoming edges  that are emanating from within the input component
     // and whose source is an ancestor of the given node n
     public int getCountIncomingEdgesFromAncestors(ArrayList<Node> scc){ // Return all incoming edges  
      
         Iterator<Edge> itr=incomingEdges.iterator();
         int count=0;
         while(itr.hasNext()){
             Edge e=itr.next();
             Node n=e.getSource();
             int trIndex=n.getTraversalIndex();
             if(scc.contains(n) && trIndex>this.traversalIndex){ // Source of this edge must be in scc.
                 count++;
             }
         }// End of edge iteration loop.
        return(count); // This might be an empty list if no edge exists that satisfies the condition
    }// End of getIncomingEdges()
    
   
    
    
    public void addToStubsUsed(Node n){
        stubsUsed.add(n);
    }
    
    public ArrayList<Node> getStubsUsed(){
        return stubsUsed;
    }
    
    public int getTraversalIndex(){
        return traversalIndex;
    }
    
       public int getSccIndex(){
        return sccIndex;
    }
    
    public void setTraversalIndex(int index){
        traversalIndex=index;
    }
    
      public void setSccIndex(int index){
        sccIndex=index;
    }
    
    public void markStub(){
        stub=true;
    }
    
    public boolean isStub(){
        return stub;
    }
    
      public void markTested(){
        tested=true;
    }
    
     public boolean isTested(){
        return tested;
    }
     
       public void markMasterRoot(){
        masterRoot=true;
    }
    
     public boolean isMasterRoot(){
        return masterRoot;
    }
    
    // Return only edges that have destination nodes at majorLevel=level
    
    public ArrayList<Edge>  getOutgoingEdges(int level){ 
        ArrayList<Edge> edgesAtLevel=new ArrayList<Edge> ();
        Iterator<Edge> itr=outgoingEdges.iterator();
        while(itr.hasNext()){
            Edge e=itr.next();
            if(e.getDestination().getLevel(0)==level){
                edgesAtLevel.add(e);
            } 
        }// End of edge loop
        return(edgesAtLevel);
    }//End of getEdges
    
 public ArrayList<Edge>  getOutgoingEdgesNotDeleted(){ 
        ArrayList<Edge> edgesNotDeleted=new ArrayList<Edge> ();
        Iterator<Edge> itr=outgoingEdges.iterator();
        while(itr.hasNext()){
            Edge e=itr.next();
            if(!e.isDeleted()){
                edgesNotDeleted.add(e);
            } 
        }// End of edge loop
        return(edgesNotDeleted);
    }//End of getEdges
 
    // Return all non-frond edges 
    // A frond edge is one that goes from a node to
    // a root of a component.
    
    public ArrayList<Edge>  getFrondlessOutgoingEdges(){  
        ArrayList<Edge> frondLessEdges=new ArrayList<Edge> ();
        Iterator<Edge> itr=outgoingEdges.iterator();
        while(itr.hasNext()){
            Edge e=itr.next();
            Node dest=e.getDestination();
            if(!e.isFrond()){
                frondLessEdges.add(e);
            } // End of frond check
        }// End of edge loop
        return(frondLessEdges);
    }//End of getEdges
    
    public String getLabel(){ // Extract node label.
        return label;
    }
    
    public int getLevel(int majorMinor){ // Extract node major/minor level set by the Tai-Daniel's algorithm.
        if(majorMinor==0)
            return levelMajor;
        else
            return levelMinor;
    }// End of getLevel()
    
    public String getLevel(){ // Extract node major/minor level set by the Tai-Daniel's algorithm.
        return Integer.toString(levelMajor)+"."+Integer.toString(levelMinor);
    }// End of getLevel()
    
    
    public boolean isRoot(){
        return rootNode;
    }
    
    public boolean isFrondRoot(){
        return frondRoot;
    }
    
    
    public int getOutgoingEdgeCount(){
        if(outgoingEdgeCount!=outgoingEdges.size()){
            System.out.println("**Program error** Outgoing edge counts do not match for node: "+
                               label.substring(1));
        }
        return outgoingEdges.size();
    }
    
    public int getIncomingEdgeCount(){
        if(incomingEdgeCount!=incomingEdges.size()){
            System.out.println("**Program error** Incoming edge counts do not match for node: "+
                               label.substring(1));
        }
        return incomingEdges.size();
    }
    
     public int getIncomingEdgeNotDeletedCount(){
        return incomingEdgeNotDeletedCount;
    }
    public int getIndex(){ 
        return index;
    }
    public int getLowLink(){
        return lowLink;
    }
    
    public boolean getMark(){
        return visited;
    }
    
    public int getFrondWeight(){
        return frondWeight;
    }
    
     public void setFrondWeight(int w){
         frondWeight=w;
    }
    
    public void markRoot(){
        rootNode=true;
    }
    
        
    public void unMarkRoot(){
        rootNode=false;
    }
    
    
    public void setOutgoingEdges(ArrayList<Edge> e){
        outgoingEdges=e;
    }
    
    public void setIncomingEdges(ArrayList<Edge> e){
        incomingEdges=e;
    }
    
    public void setLevel(int l, int majorMinor){ // Set node major level 
        if(majorMinor==0)
            levelMajor=l;
        else
            levelMinor=l;
    }//End of setLevel()
    
    public void setIndex(int index){
        this.index=index;
    }
    
    public void setLowLink(int link){
        lowLink=link;
    }
    
    public void incFrondWeight(){
        frondWeight++;
    }
    
    public void markVisited(){ // mark=true implies visited, false implies not visited.
        visited=true;
    }
    
    public void markNotVisited(){ // mark=true implies visited, false implies not visited.
        visited=false;
        Iterator<Edge> itr =outgoingEdges.iterator();
        while(itr.hasNext()){
            Edge e=itr.next();
            e.markNotVisited(); 
        }
    }
    
    public void markFrondRoot(){
        frondRoot=true;
    }
    
    public void setAbstract(){
        isAbstract=true;
    }
    
    public boolean isNodeAbstract(){ 
        return isAbstract;
    }
    
    
    
    // Find the set of (immediate) successor nodes of a given node.
    // Given that there might be a loop in the graph due to association edges,
    // we omit destination nodes of edges labeled eA (Association).
    
    // ignoreEdges contains labels of edges that are to be ignored when finding 
    // successors of a node. It is used to implement the Tai-Daniels algorithm.
    // Of course, it could empty when all edges in a graph are to be considered.
    
    public ArrayList<Node> getSuccessors(ArrayList<String> ignoreEdges){
        ArrayList <Node> successorNodes=new ArrayList<Node>();
        
        Iterator itr =outgoingEdges.iterator();
        while(itr.hasNext()){
            Edge nextEdge=(Edge)itr.next();
            Node destination=nextEdge.getDestination();
            if(ignoreEdges.indexOf(nextEdge.getLabel())==-1){
                successorNodes.add(destination);
            }
        }
        return successorNodes;
        
    } // End of successors()
    
    public ArrayList<Node> getSuccessors(){
        
        return getSuccessors(new ArrayList<String>());
        
    } // End of successors()
    
    
    // The next method is similar to getSuccessor() except that
    // it returns a list of indices (integers) to nodes that
    // are immediate successors of this node. The indices point
    /// directly into the graph.
    
    public ArrayList<Integer> getSuccessorsIndex(){
        ArrayList <Integer> indices=new ArrayList<Integer>();
        
        Iterator itr =outgoingEdges.iterator();
        int index=0;
        while(itr.hasNext()){
            Edge nextEdge=(Edge)itr.next();
            Node destination=nextEdge.getDestination();
            if(!nextEdge.getLabel().equals("eA")){
                indices.add(index);
            }
        }
        return indices;
        
    } // End of successors()
    
    // This method adds a a list of statically dependent nodes 
    // to that of this node.
    
    public void setStaticDependence( ArrayList<Node> list){
        staticDependence=list;
        
    }// End of updateStaticDependence()
    
    public ArrayList<Node> getStaticDependence(){
        
        return staticDependence;
        
    }// End getStaticDependence()
    
    public void updateDynamicDependence(Node n){
        dynamicDependence.add(n);
    } // End of updateDynamicDependence()
    
    public ArrayList<Node> getDynamicDependence(){ 
        return dynamicDependence;
    }// End getDynamicDependence()
    
    public void setDynamicDependence(ArrayList<Node> list){ 
        dynamicDependence=list;
    }// End of etDynamicDependence
    
    public ArrayList<Node> getInheritanceChain(){ 
        return inheritanceChain;
    }// End getDynamicDependence()
    
    public void setInheritanceChain(ArrayList<Node> chain){ 
        inheritanceChain=chain;
    }// End of etDynamicDependence
  
    
    public void updateInitialSccFrondWeight(ArrayList<Node> scc){
        Iterator<Edge> itr=getIncomingEdges().iterator();
        boolean frondFound=false;
        while(itr.hasNext()){
            Edge e=itr.next(); 
            Node source=e.getSource();
            int sourceIndex=source.getTraversalIndex();
            if(sourceIndex>traversalIndex && scc.contains(source)){
              if(Utilities.frondDebug){
                Utilities.displayEdge(e, false);
              }
                e.markFrond();
                incFrondWeight();
                source.incFrondWeight();
                frondFound=true;
            }
        }// End of edge iterator.
        if(!frondFound){
            System.out.println("** Program error (updateFrondWeight()/Node)."+
                               "\n No incoming edge from a descendent or Scc contains any of the source nodes.");
            System.exit(0); 
        }
        
    }// End of updateInitialSccFrondWeight()
    
    public void updateFrondWeight(ArrayList<Node> scc){
        Iterator<Edge> itr=getIncomingEdges().iterator();
        boolean frondFound=false;
        while(itr.hasNext()){
            Edge e=itr.next(); 
            Node source=e.getSource();
            int sourceIndex=source.getTraversalIndex();
            if(scc.contains(source)){
               if(Utilities.frondDebug){
                Utilities.displayEdge(e, false);
               }
                e.markFrond();
                incFrondWeight();
                source.incFrondWeight();
                frondFound=true;
            }
        }// End of edge iterator.
        if(!frondFound){
            System.out.println("** Program error (updateFrondWeight()/Node)."+
                               "\n No incoming edge from a descendent or Scc does not contain any of the source nodes.");
            System.exit(0); 
        }
    }
    
} // End of Node class
