
/*
 * Tarjan's algorithm for finding Strongly Connected Componets (SCC) in a
 * directed graph.
 * 
 * Coded by: Aditya Mathur
 * Algorithm source: http://en.wikipedia.org/wiki/Tarjan%27s_strongly_connected_components_algorithm
 * Original coding: July 13, 2011
 * Latest update: July 20, 2011 [Inadvertently introduced a bug and then removed it!!]
 * 
 * January 10, 2012 : Frond edges were not correctly recognized in the
 * previous version. This version corrects the problem.
 *
 * Another update: August 11, 2011: Added indexing of the initial SCCs. 
 * 
 * When applying Tarjan's algorithm recursively on sub-SCCs,
 * it is important to start at the root of the sub-SCC. The sub-SCC arises
 * when applying Le Traon et al. algorithm. This start node is 
 * not needed if only one iteration of Tarjan's algorithm is applied. Thus,
 * one could start at any node and the algorithm will give the correct SCCs.
 * It is only in the recursive application of the algorithm that the start node
 * is needed. 
 * If the start node is not accounted for then one might get orphan nodes in the trivial SCCs.
 * These are nodes not reachable from any node in the (original) input graph.
 * 
 * 
 * Each component is an ArrayList<Node>. Thus, the method tarjan() returns
 * an ArrayList<ArrayList<Node>> object that
 * contains one or more SCCs.
 * 
 * tarjan() ignores all edges whose label does not match
 * "eA". 
 */

import java.util.*;
public class SccTarjan{
    
    private int index;
    private Stack <Node> s=new Stack<Node>(); // For use in Tarjan's algorithm
    private Graph g;
    private ArrayList<ArrayList<Node>> scc=new ArrayList<ArrayList<Node>>(); // This is where all SCCs in g are saved.
    ArrayList<Node> nodes;
    ArrayList<Scc> sccInitial; // To save the Sccs created.
    Scc sccTemp; // An SCC
    int sccIndex; // Indexes the SCCs created in the first call to tarjan() in LeTraon etal. algorithm.
    
    // Constructor.
    public SccTarjan(Graph g){
        this.g=g; // Save graph for future use.
        if(!LeTraonAlgorithm.recursiveCall){
            sccInitial=new ArrayList<Scc>();
            sccIndex=1; // Index of the next initial Scc.
        }     
    }// End of constructor.
    
    
    // Find all SCCs. tarjan() is prelude to the algorithm
    // to find SCCs.
    public   ArrayList<ArrayList<Node>> tarjan(){
        index=0;  
         nodes=g.getNodes(); // Get all nodes in g.
        int numNodes=nodes.size(); // Number of nodes in g.
        
        if(numNodes==0){
            Utilities.debug(Utilities.tarjanDebug, "--Graph is empty.");
            return scc; // Empty scc returned.
        }
        // Loop over all nodes, one at a time.
        Iterator itr=nodes.iterator();
        int nodeIndex=0; // Index into nodes.
        for(int i=0; i<nodes.size(); i++){  
            Node n=nodes.get(i);
            if(n.getIndex()==-1){ // Process this node if not  visited.
                Utilities.debug(Utilities.tarjanDebug, "\n     Now processing node:"+
                                ": " +n.getLabel().substring(1));
               if(!LeTraonAlgorithm.recursiveCall)
                   n.setTraversalIndex(LeTraonAlgorithm.counter++); // Set the node's traversal index.
                  
                
               strongConnectTarjan(nodes.get(i)); // Get the ith SCC with root as n.
                
                Utilities.debug(Utilities.tarjanDebug,"     Node: "+n.getLabel().substring(1)+" processed.");
            }// End of if
        }// End of for over nodes.
        return scc; // scc is updated inside strongConnect() after each component is found.
        
    } // End of tarjan() starter portion of the algorithm.
    

    // Tarjan's strongConnect() method follows.
    
    private void strongConnectTarjan(Node v){
        
        v.setIndex(index);
        v.setLowLink(index);
        index++;
        s.push(v);
        ArrayList<Edge> edges;
        if(Utilities.TaiDaniels) // Check if this is Tai-Daniels algorithm
            edges =v.getOutgoingEdges(); // Get all edges in the component defined by n.
        else
            //  edges =v.getFrondlessOutgoingEdges(); // Frond edges are not returned.
            edges =v.getOutgoingEdges(); // Frond edges are not returned.
        
        // Consider successors of node v.
        
        Iterator itrEdge=edges.iterator();
        
        while(itrEdge.hasNext()){
            Edge e=(Edge)itrEdge.next();
            //    if(!e.isFrond()){ // Consider only edges that have not been not deleted.
            if(!e.isDeleted()){ // Consider only edges that have not been not deleted.
                Node w=e.getDestination(); // Get the successor in g.
                
                // Make sure the destination node is in the graph.
                // Note: This may be a subgraph of a graph and hence
                // the check is needed because an edge of a subgraph (an SCC)
                // may take us out of the subgraph.
                if(inGraph(w)){
                    if(w.getIndex()==-1){ // Check if not visited.
                        if(!LeTraonAlgorithm.recursiveCall){
                            w.setTraversalIndex(LeTraonAlgorithm.counter++); // Set the node's traversal index.
                        }
                        strongConnectTarjan(e.getDestination());
                        v.setLowLink(min(v.getLowLink(), w.getLowLink()));
                    } else{
                        if(s.search(w)!=-1)
                            v.setLowLink(min(v.getLowLink(), w.getIndex()));
                    }// End of else part  
                }// End of node in graph check.
            }  // End of edge deleted heck
        }// End of while to iterate through successors of n.
        
        // If n is the root node then pop the stack and generate an SCC.
        
        if(v.getLowLink()==v.getIndex()){
            ArrayList<Node> anScc=new ArrayList<Node>(); // Initialize an Scc
            Node w;
            do{
                w=s.pop();
                anScc.add(w); // Add node to SCC.
                if(!LeTraonAlgorithm.recursiveCall){
                    w.setSccIndex(sccIndex); // Set the index of its Scc.
                }
            }while(!v.getLabel().equals(w.getLabel()));
            
            // Mark v as the root of anScc that has a cycle but only if it is non-trivial (>1 node).
            if(anScc.size()>1){
                v.markRoot();
            } 
            scc.add(anScc); // Add this SCC to the list of SCCs. 
            if(!LeTraonAlgorithm.recursiveCall){
                sccTemp=new Scc(anScc, v, sccIndex);
                sccIndex++;
                sccInitial.add(sccTemp);
            }// End of if block to save the Scc just created.
            
        }// End of if
    }// End of strongConnect()
    
    
    // Check if the input node exists in the graph (or subgraph) g.
    
    private boolean inGraph(Node w){
        ArrayList<Node> nodes =g.getNodes();
        Iterator itr=nodes.iterator();
        while(itr.hasNext()){
            Node n=(Node)itr.next();
            String labelN=n.getLabel();
            String labelW=w.getLabel();
            if(labelW.equals(labelN)){
                return true;
            }     
        }
        return false;
    }  // End of inGraph()

    private int min(int a, int b){
        if(a<b)
            return a;
        else
            return b;
    }// End of min()
    
    // DEbugging method
    
    public void displayIndices(Graph g){
        ArrayList<Node> nodes=g.getNodes();
        int count=nodes.size();
        for(int i=0; i<count; i++){
            System.out.print(" "+nodes.get(i).getLabel().substring(1)+" "+nodes.get(i).getIndex()); 
        }
        System.out.println();
    }
    
    // Reorder to ensure that the root node is the first element of the SCC.
    private void reorderScc(ArrayList<Node> nodes){
        boolean rootFound=false;
        for(int i=0; i<nodes.size(); i++){
            Node n=nodes.get(i);
            if(n.isRoot()){// Check if it is the root node.
                rootFound=true;
                Node temp=nodes.get(0); // Yes it is, hence bring it to the start of the SCC.
                nodes.set(0, n);
                nodes.set(i, temp);
              
                break; // exit from loop (not from the if as AT&T did in the early 90's)
            }                 
        }// End of iteration over SCC nodes.
        if(!rootFound){
            Utilities.debug(Utilities.tarjanDebug,
                            "** Program error (reorderScc (SccTarjan)). Root not found in an Scc.");
            System.exit(0);
        }
        
    }// End of reorderNodes()
    
    public ArrayList<Scc> getInitialScc(){
        return sccInitial;
    }
    
} // End of class SccTarjan