

/*
 * This class implements the algorithm developed by
 * Tai and Daniels for generating a reasonable test order 
 * for class integration testing. The classes are input
 * to the program in the form of a graph where the nodes represent the classes
 * to be tested and the edges represent the relationships among the classes.
 * 
 * Three types of edges are considered: 
 * 
 * Association (labelled A)
 * Aggregation (labelled C)
 * Inheritence (labelled I)
 * 
 * It is assumed that in the input graph only the edges labelled A can create cycles in the graph. 
 * This assumpiton is not correct as in a Java program, for example,
 * it is possible to have the Ag edges also create cycles. 
 * For example, a Class object can hold multiple Sdtudent objects and 
 * a Student object can hold multiple Class objects thus creating a cycle due
 * to aggregation.
 * 
 * Reference:
 * 
 * Test Order for Inter-Class Integration Testing of Object-Oriented Software,
 * KC Tai and Fonda J Daniels, Proceedings of the 21st International Computer 
 * Software and Applications Conference, COMPSAC '97, pp 602-607, 1997.
 * 
 * This class coding 
 * Original: July 7, 2011
 * Latest update: Aug 2, 2011. 
 * More update: January 5, 2012 (Deletion status of edges added)
 * More update: Feb 1, 2012: Added automated ORD file input from a folder
 * and appending test order summary to a file.
 * 
 * Notes:
 * 
 * 1. The first complete version of this tool passed tests on July 27, 2011.
 * 2. The second version began on July 28, 2011. It includes abstract classes, static as well
 * as dynamic dependencies, and addition information in test order display.
 * 
 * Author: Aditya Mathur
 * 
 */


/*
 * Note: There is one minor error and one omission in Tai and Daniels' paper. 
 * Error: If Fig 1 is correct then Figure 2 is incorrect; i.e., CS should be at level 5.
 * If Fig 2 is correct then the edge from C to CS in Figure 1 
 * must be labelled Ag.
 * 
 * Omission: On page 605, right column, para above section 5:
 * "When S is being tested, the edge from I1 to S is retested. 
 * Also, when B is being tested, the edge from BL to B is retested." In addition, we add:
 * 
 * "When W is being tested, the association edge from E to W is retested. 
 * Also, when C is being tested, the edge fromd CS to C is  retested."
 */

/* An interesting "bug" (perhaps a bug)  discovered on Jan 12, 2012:
 * 
 * The program works just fine for a fully connected graph with 2 nodes or with 3 nodes. 
 * But it does not work for fully connected graph with nodes greater than 3. 
 * I tried it for fully connected graphs with 4, 5, and 6 nodes. The number of stubs remains
 * at 2!! It should be (n01) where n is the number nodes. 
 * 
 * The program incorrectly assigns minor level numbers. In fact it assigns minor level 1 to all 
 * nodes from 1, 2,...(n-2) and then assigns minor levels 2 and 3 to nodes (n-1) and n.
 * The order in which nodes are input has no impact on the number of stubs generated.
 * 
 * Is this really a bug? Well, according to tai-Daniels 
 * 
 * "In step (3.3), two or more classes in ML(G,i) that are onnected by association edges deleted 
 * in step (3.2) may be assigned the same minor level number. 
 * What is the test order for classes with the same level number? 
 * Since the number of classes with the same level number is usually small, 
 * a practical solution is to perform inter-class integration of these classes at the same time."
 * 
 * Thus, getting no stubs for classes 1 through (n-2) is OK as per Tai-dainels' algorithm.
 * However, this may not be a good solution in some cases especially when the algorithm
 * suggests testing togerher a "large" number of classes.
 * 
 * Hence a fix I used for this "bug" is the following:
 * (a) Offer user a chhoice of whether to use the TD algorithm strictly or to 
 * generate additional stubs as in (b). If strict application is requested then do not use (b).
 * 
 * (b) Let the successors of class X be Y and Z. If Y and Z are not connected by an association edge
 * then use the strict version of the TD algorithm. Otherwise, if Z is a descendent of Y then assign 1
 * higher minor level number to Y than to Z. Thus, if we have X-->Y-->Z then the minor level numbers are
 * X:3, Y:2, Z:3.
 * 
 * Correction. January ?? 2012.

/*
 * Abrreviations used:
 * 
 * SCC or Scc or scc: Strongly connected component
 * ML: Major Level
 */


import java.util.*;
import java.io.*;

public class TaiDanielsAlgorithms {
    
    // Object Relation Diagram as a graph. Attributes of nodes in this
    // graph are modified as the Tai-Daniels algorithm progresses.
    Graph g; 
    
    final int processMajor=0; // Indiates that major level is to be processed. 
    int maxMajorLevel=0; // Maximum major level number assigned in the graph.
    int [] maxMinorLevel; // Maximum minor level number assigned in the graph to each level.
    int majorMinor; // =0 implies major, =1 implies minor. Used in deciding whether to assign major or minor level.
    String deletedEdgeLabel="eA"+Edge.deleteString; // Label of an association edge that has been removed from g.
    String deletionCause; // Used for debugging. Set to "s" for stubs or "w" for weights".
    ArrayList<ArrayList<Node>> testSequence; // Sequence in which classes are to be tested.
    ArrayList<Stub> stubs=new ArrayList<Stub>(); // Stubs needed.
    ArrayList<Retest> retestNodes; // Nodes to be retested.
    ArrayList<Edge> deletedEdges=new ArrayList<Edge>();
   
    static ArrayList<String> ordFiles; // List of ORD file names.
    
    static String ordFolder="../TestGraphs/SecondEditionExperiments"; // This is where the ORDs are located.
    // static String testOrderStatistics="../TestGraphs/testOrderStatistics";
    static String testOrderStatistics="../TestGraphs/testOrderStatisticsFullyConnectedORDs";
    static String testData; // Stub count etc is appended to this file.
    
    //Next flag indicates whether or not dynamicstatic dependencies are to be considered.
    // Set it to false to include dynamic dependencies.
    boolean staticOnly=true; 
    
    
    // The constructor creates a graph g from an input text file.
    
    public TaiDanielsAlgorithms(String fileName)throws IOException{
      // TD algorithm Step 1: Read a graph (from a file with name in fileName).
    
      System.out.println("Input graph from file: "+fileName);
      CreateGraph cg=new CreateGraph();
      cg.createGraph(fileName, staticOnly);
      
      System.out.println("--Graph successfully input."+ " Nodes: "+ cg.getNodeCount()+ "  Edges: "+
                         cg.getEdgeCount());
      // cg.createGraph("sccTest1NoEdges.txt");
      // cg.createGraph("sccTest2SymmetricTwoComponents.txt");
      
      //cg.createGraph("sccTest3ThreeComponents.txt");
      // cg.createGraph("sccTest4OneNodeDegenerate.txt");
      //cg.createGraph("sccTest5TwoNodesDegenerate.txt");
      g=cg.g; // Save the graph locally.
      
    }// End of TaidanielsAlgorithm constructor
    
    // Assign level numbers to each node in the acyclic directed graph (ADG).
    // Parameter majorMinor indicates whether the level to
    // be assigned is major (=0) or minor (=1). ignoreEdges contains
    // labels of edges that are to be ignored when finding 
    // successors of a node.
    
    public void levelADG(Graph g, int majorMinor, ArrayList<String> ignoreEdges, int majorLevel){
        
        // Set up iterator to iterate over all nodes in the input graph.
        Iterator<Node> itr=g.getNodes().iterator();  
        // Mark all nodes as not-visited.
        Node n; // Temporary node object.
        while(itr.hasNext()){
            n=(Node)itr.next();
            n.markNotVisited(); // Set node to not visited.
        }
        
        // As the previous iterator  has iterated over all nodes,
        // create another iterator over nodes in the graph.
        
        itr=g.getNodes().iterator();  
        // Do a level-sort for each node that is not visited.
        while(itr.hasNext()){
            n=itr.next();
            if (!n.getMark()) {// Assign level number only if node not yet visited
                levelSort(g, n, majorMinor, ignoreEdges, majorLevel);  
            }
        }
    } // End of levelADG()
    
    // Sort and set level numbers of nodes. 
    // This is a step in the Tai-Daniel's method.
    
    private void levelSort(Graph g, Node n, int majorMinor, ArrayList<String> ignoreEdges, int majorLevel){
        ArrayList<Node> successor=n.getSuccessors(ignoreEdges);
        Utilities.debug(Utilities.levelAssignmentDebug, "\nAssign "+majorMinorString(majorMinor) +
                        " level to node "+n.getLabel().substring(1));
        
        // Graph g might be a subgraph of another graph.
        // Hence the list of successors might contain nodes that
        // are outside of g. Remove all such nodes.
        successor=removeNodesNotInGraph(g, successor);
        
        if(Utilities.levelAssignmentDebug){
            System.out.println("Successor node(s) of "+n.getLabel().substring(1)); 
            Utilities.displayNodes(successor, false, false); 
        }
        
        n.markVisited(); // Mark n to visited.
        if(successor.isEmpty()){
            n.setLevel(1, majorMinor);
            if(majorMinor==processMajor){
              if(maxMajorLevel==0)
                maxMajorLevel=1;
            }else{
              setMaxMinorLevel(1, majorLevel);
            }
            Utilities.debug(Utilities.levelAssignmentDebug, "Assigned  "+majorMinorString(majorMinor) +
                            " level 1 to node "+n.getLabel().substring(1));
        }else{ // Successor not empty
            Iterator itr=successor.iterator();
            while(itr.hasNext()){
                Node next=(Node)itr.next();
                if(!next.getMark())
                    levelSort(g, next, majorMinor, ignoreEdges, majorLevel);
            }// End of loop to assign levels to successors. 
            
           // A bug in my implementation was that the loop continued before 
            // all successors were assigned minor level. This error was not detected
            // when tested on  Tai-Daniels example as it did not have more than one successor of
            // anode at the same level. The bug was revealed [January 12, 2012] when I tried 
           //  to apply the algorithm on a fully conected graph with more than 3-nodes.
            
                int level=findMaxLevel(successor, majorMinor)+1;
                n.setLevel(level, majorMinor);
                Utilities.debug(Utilities.levelAssignmentDebug, "Assigned  "+majorMinorString(majorMinor) +
                                " level "+ level+" to node "+n.getLabel().substring(1)+
                                ". Max major level: "+maxMajorLevel);
                if(maxMajorLevel<level&&majorMinor==0){
                    maxMajorLevel=level; // Update highest major level number.
                }else{
                    if(majorMinor==1){
                        setMaxMinorLevel(level, majorLevel); // Update maximum minor level at this majorLevel.
                    }   
                }
        } // End of else (non-empty successor set.
    } // End of levelSort()
    
    // Return the maximum level number 
    // among successors of a node.  
    
    private int findMaxLevel(ArrayList<Node> s, int majorMinor){
        Iterator itr=s.iterator();
        int max=-1; // Initialize maximum level number found so far.
        int nextLevel; // Temporary variable.
        while(itr.hasNext()){
            Node next=(Node)itr.next();
            nextLevel=next.getLevel(majorMinor); 
            if(max<nextLevel)
                max=nextLevel;
        }// End of while
        return max;
    } // End of maxMajorLevel()
    
    // Update max minor level at this major level.
    private void setMaxMinorLevel(int level, int majorLevel){
        if(maxMinorLevel[majorLevel-1]<level) 
            maxMinorLevel[majorLevel-1]=level;
    } // End of setMaxMinorLevel()
    
    // The following method assigns minor level numbers to each class.
    
    public void assignMinorLevelNumbers(Graph g, int maxMajorLevel, ArrayList<String> ignoreEdges){
        
        // Let ML(g,i) denote a subgraph of Graph g that contains
        // only classes at major level i (1<=i<=maxMajorLevel) and association (A)
        // edges between these classes.
        
        // Find and process ML(G, i).
        
        for(int i=1; i<=maxMajorLevel; i++){
            Utilities.debug(Utilities.minorLevelProcessingDebug, "\n    Process subgraph at Major Level "+i);
            ArrayList<Node> ML=g.getNodes(i);  // Place all nodes at major level i into ML.
            if(Utilities.minorLevelProcessingDebug)
                Utilities.displayNodes(ML, false, false);   // Display only nodes, not edges.
            processML(ML, i, ignoreEdges); // Process this subgraph (ML) as per Tai-Daniels algorithm. 
            Utilities.debug(Utilities.minorLevelProcessingDebug, "\n-----Subgraph at Major Level "+i+" processed.");
            if(Utilities.stubCreationDebug){
              System.out.println("Stubs added so far:");
              Utilities.displayStubsCreated(stubs);
            }
        }
        
    }
    // The following method assigns minor level numbers to a set of nodes 
    // that constitute a subgraph of g.
    // After having assigned the minor level numbers to nodes in ML
    // it updates g to reflect the newly assigned minor level numbers.
    
    private void processML(ArrayList<Node> ML, int majorLevel, ArrayList<String> ignoreEdges){
        
        // Check if ML has size 1 (i.e., contains only one node, hence no cycles).
        
        if(ML.size()==1){ 
            // TD algorithm Step 3.1: Assign minor level numbers to nodes.   
            Node n=ML.get(0);// Extract the sole node at this major level (ML).
            n.setLevel(1,1); // Set the minor level of this node to 1.
            setMaxMinorLevel(1, majorLevel); // Update max minor level at this major level.
            Utilities.debug(Utilities.levelAssignmentDebug, "Assigned  "+majorMinorString(majorMinor) +
                            " level 1 to node "+n.getLabel().substring(1));
            Graph gML=new Graph(ML); // Create a graph out of ML.
            updateGraph(g, gML); // Update graph g with gML.
            if(Utilities.minorLevelProcessingDebug)
                Utilities.displayNodes(ML, false, false); // Display ML (for debugging).
            
            Utilities.debug(Utilities.minorLevelProcessingDebug, "\n----Processing of ML complete. No cycles.");
           
            return;
            
        }// End of Step 3.1
        
        // No, it contains more than 1 node.
        // Hence, check if there is a cycle in ML.
        // This is done using Tarjan's algorithm.
        
        ArrayList<ArrayList<Node>> scc=findScc(ML); // Find SCCs in ML.
        
        if(Utilities.minorLevelProcessingDebug)
            Utilities.displayAllScc(scc); // Display SCCs (for debugging).
        
        if(!cycles(ML, scc)){ // If ML has no  cycles then...
            // apply TD algorithm Step 3.2, i.e., assign minor level numbers to nodes.
            Graph gML=new Graph(ML);
            levelADG(gML, 1,ignoreEdges, majorLevel); // assign minor level numbers, and 
            setMaxMinorLevel(1, majorLevel); // Update max minor level at this major level.
            updateGraph(g, gML); // update graph.
            Utilities.debug(Utilities.minorLevelProcessingDebug, "\n----Processing of ML complete. No cycles.");
            return;
            
        } // End of Step 3.2
        
        // ML has cycles. So process it differently.
        // TD algorithm Step 3.3: Break cycles and then 
        // assign minor level numbers to nodes.
        
        processMLWithCycles(scc, majorLevel); // This will break cycles and mark "deleted" edges.
        int cycleCount=countComponentsWithCycles(scc);// Used only for printing.
        Graph gML=new Graph(ML);
        levelADG(gML, 1, ignoreEdges, majorLevel); // assign minor level numbers, and 
        updateGraph(g, gML); // update graph.
        Utilities.debug(Utilities.minorLevelProcessingDebug, "\n----Processing of ML complete. "+
                        cycleCount+" cycle"+Utilities.singPlu(cycleCount)+" found");
        return;
        
    } // End of process ML
    
    // The next method checks if a component of a graph  has any cycles in it.
    // The component is input as a list of nodes.
    
    private boolean cycles(ArrayList<Node> n, ArrayList<ArrayList<Node>> scc){
        
        // If the number of SCC is less than the number of nodes in n then
        // there must be a cycle.
        if(scc.size()<n.size()){
            Utilities.debug(Utilities.cycleCheckingDebug, "\n     There is at least one cycle. ");
            return true; 
        }else{
            Utilities.debug(Utilities.cycleCheckingDebug, "\n     There is no cycle.");
            return false;
        }
        
    }// End of cycles()
    
    // Find and return SCCs in input list of nodes which may be a DAG.
    private ArrayList<ArrayList<Node>> findScc(ArrayList<Node> c){
        Graph newGraph=new Graph(c); // Create a graph from the component.
        newGraph.resetMarks(); // Make all nodes not visited.
        if(Utilities.cycleBreakingDebug)
            Utilities.displayNodeList(c, null, true, false);
        
        SccTarjan sccAlg=new SccTarjan(newGraph); // Create object to compute SCCs in graph newGraph.
        ArrayList<ArrayList<Node>> scc=sccAlg.tarjan(); // Find strongly connected components.
        return scc;
    }// End findScc()
    
    // The following method applies Tai-Daniels algorithm to process a subgraph (ML) of g
    // and assign minor level numbers to its nodes. The subgraph 
    // contains SCCs with at least one SCC containing one or more
    // cycles. Cycles are broken by removing edges and then 
    // minor level numbers are assigned.
    
    private void processMLWithCycles(ArrayList<ArrayList<Node>> scc, int majorLevel){
        
        // Create an iterator to iterate over all SCCs in ML
        Iterator<ArrayList<Node>> itr=scc.iterator(); 
        ArrayList<Node> component;
        ArrayList<Edge> deletedEdgeList;
        
        // Iterate over each SCC in ML. Some may have cycles and some may not.
        int totalComponents=scc.size();
        
        for(int nextComponent=1; nextComponent<=totalComponents; nextComponent++){
            component=itr.next(); // Get the next SCC in major level (ML).
            
            if(Utilities.cycleBreakingDebug){
                System.out.println("\nProcessing component: "+nextComponent);
                Utilities.displayAScc(component);
            }
            
            if(component.size()>1){ // Check if this component has a cycle.
                Utilities.debug(Utilities.cycleBreakingDebug, "\n Break cycles in component: "+nextComponent);
                
                // Now first try to break cycles by removing edges (Step 3.2(a) in TD algorithm).
                // If cycles still remain then apply the weights method to break any
                // remaining cycles (Step 3.2(b) of TD algorithm).
                
                Utilities.debug(Utilities.cycleBreakingDebug, "  Break cycles to minimize stubs.");
                deletionCause="S"; // Set deletion criteria to "Stub" for debugging
                breakCyclesByStubs(component, nextComponent, majorLevel);// Try removing edges to minimize stubs needed.
                
                // Cycles may still exist. So apply Step 3.2(b) of the TD algorithm.
                ArrayList<ArrayList<Node>> sccThisComponent=findScc(component);
                if(!cycles(component, sccThisComponent)){
                    Utilities.debug(Utilities.cycleBreakingDebug, "No more cycles in this component.");
                    return;
                }
                Utilities.debug(Utilities.cycleBreakingDebug,"  \nBreak any remaining cycles based on weights.");
                deletionCause="W"; // Set deletion criteria to weights for debugging
                breakCyclesByWeight(component); // Remove edges by weight
                Utilities.debug(Utilities.cycleBreakingDebug, "Component "+nextComponent+ " modified.\n");
            }else{
                Utilities.debug(Utilities.cycleBreakingDebug, "This component has no cycles.");

            }// End of component size check
        } // End of component iteration
        
    }// End of processMLWithCycles()
    
    
    // The following method processes a component with cycles.
    // It takes an SCC (c), with cycles as input.
    // It then removes edges to break cycles in the SCC to minimize the
    // number of stubs needed. The edges ar not "truely" deleted/removed.
    // Instead, the label of a "deleted" edge is appended with a special symbol.
    // It returns a new component that might not have any cycle.
    // This is Step 3.2(a) of the TD algorithm.
    
    private void breakCyclesByStubs(ArrayList<Node> c, int cID, int majorLevel ){
        ArrayList<Edge> removedEdgeList=new ArrayList<Edge>();
        
        // stubs is the set of nodes in component c that
        // have incoming association edges (A) from classes at
        // major level less than majorLevel.
        ArrayList<Node> stubs=new ArrayList<Node>();
        
        // Find nodes at level less than majorLevel that have
        // association edges coming into a node k in c. Each such k
        // that has at least one incoming association edge is added to stubs.
        
        stubs=getStubs(majorLevel, c);
        
        if(Utilities.cycleBreakingDebug){
            System.out.println("   Stubs for component "+cID+ " at major level "+ majorLevel);
            if(stubs.size()==0){
                System.out.println("     There is no stub at this major level.");
            }else{  
                Utilities.displayNodes(stubs, false, false);
            }
        }// End of debugging code
        
        if(stubs.size()>0){ // Edges can be deleted only if there are nodes in the stub.
            ArrayList<Edge> edgesToStubs=new ArrayList<Edge>();
            // Find association edges from nodes in component c that are 
            // coming into nodes in stubs.
            Utilities.debug(Utilities.cycleBreakingDebug,"    Find edges to nodes in the stub.");
            edgesToStubs=findEdgesToStubs(stubs, c);
            if(edgesToStubs.size()==0){
                Utilities.debug(Utilities.cycleBreakingDebug,"     No edges to display.");
            }else{
                Utilities.displayEdges(edgesToStubs, true);
                removedEdgeList=removeEdges(edgesToStubs, c); // Delete edges from c to break cycles.
            }
        }else{
            Utilities.debug(Utilities.cycleBreakingDebug, "  No edges deleted.");
        }
        
    }// End of breakCyclesByStubs()
    
    
    // The next method breaks any remaining cycles using
    // edge weights. Edges with the highest weight are
    // deleted in that order until no cycles remain.
    
    // Inputs: c is the component with possible cycles.
    // Again, edges ar not "truely" deleted. Instead, the label of
    // a "deleted" edge is appended with a special symbol.
    
    private void breakCyclesByWeight(ArrayList<Node> c){
        
        ArrayList<Edge> sortedEdgeList=new ArrayList<Edge>();
        // Get all edges in component c and set their respective 
        // weights (=#incoming edges+#outgoing edges)
        ArrayList<Edge> allEdges=getAllEdges(c); 
        
        sortedEdgeList=sortEdgesByWeight(allEdges); // Sort the edges by weight.
        
        if(Utilities.cycleBreakingDebug){
            System.out.println("\n   Sorted edge list: ");
            Utilities.displayEdges(sortedEdgeList, true);
        }
        
        int indexOfNextEdgeToBeDeleted=0;
        int edgesRemaining=sortedEdgeList.size();
        boolean cycleExists=true; // We checked this prior to entering this method.
        int edgesDeletedCount=0; // Track the number of edges deleted.
        // Continue edge deletion until cycles exist and there is an edge to delete.
        while(cycleExists && edgesRemaining>0){ 
            // Select next edge to be removed.
            Edge e=sortedEdgeList.get(indexOfNextEdgeToBeDeleted);
            if(Utilities.cycleBreakingDebug){
                System.out.println("     Next edge to be removed:" );
                Utilities.displayEdge(e, true);
            }
            removeEdgeFromComponent(c,e);
            edgesDeletedCount++;
            //Utilities.updateEdgeInGraph(g, e);
            edgesRemaining--;
            // Check if a cycle still exists. 
            indexOfNextEdgeToBeDeleted++;
            ArrayList<ArrayList<Node>> scc=findScc(c);
            if(Utilities.cycleBreakingDebug)
                Utilities.displayAllScc(scc);
            
            cycleExists=cycles(c, scc); 
            
        }//End while
        Utilities.debug(Utilities.cycleBreakingDebug, "     All cycles removed. "+ edgesDeletedCount+
                        " edge"+Utilities.singPlu(edgesDeletedCount)+" removed.");
        
    }// End of  breakCyclesByWeight()
    
    
    // Count the number of SCCs with cycles.
    private int countComponentsWithCycles(ArrayList<ArrayList<Node>> scc){
        if(scc.size()==0) // If scc is empty then no comppnents with cycles.
            return 0;
        
        Iterator<ArrayList<Node>>  itr=scc.iterator();
        int count=0;
        while(itr.hasNext()){
            ArrayList<Node> component=itr.next();
            if(component.size()>1)
                count++;
        }// End of while
        return count;
        
    }// End of countComponentsWithCycles()
    
    // Sort edges in descending order of their respective weights. 
    private ArrayList<Edge> sortEdgesByWeight(ArrayList<Edge> e){
        int listSize=e.size();
        for (int i=0; i<listSize; i++){
            for (int j=0; j<listSize-1; j++){
                Edge e1=e.get(j);
                Edge e2=e.get(j+1);
                int w1=e1.getWeight();
                int w2=e2.getWeight();
                if(w1<w2){
                    Edge tempEdge=e1;
                    e.set(j, e2);
                    e.set(j+1, tempEdge);
                } 
            }
        }
        return e;
    }//End sortEdgesByWeight()
    
    
    // Get all outgoing edges from
    // nodes in a list of nodes. Edges returned must be "inside"
    // component n. It is assumed that n is a strongly connected component.
    // There might be edges going out of this SC, and these
    // are to be ignored. Also ignored are edges that are coming in from
    // outside of this component.
    
    private ArrayList<Edge> getAllEdges(ArrayList<Node> n){
        ArrayList<Edge> allEdges=new ArrayList<Edge> ();
        Iterator <Node> itr=n.iterator();
        while(itr.hasNext()){
            Node next=itr.next();
            ArrayList<Edge> e=next.getOutgoingEdges();
            for(int i=0; i<e.size(); i++){
                Edge nextEdge=e.get(i);
                Node source=nextEdge.getSource();
                Node destination=nextEdge.getDestination();
                // Include this edg, and its weight, only if its source and 
                // destination are inside the component (n).
                if(n.contains(source) && n.contains(destination)){
                    nextEdge.setEdgeWeight(computeEdgeWeight(nextEdge, n));
                    allEdges.add(nextEdge);  
                    if(Utilities.cycleBreakingDebug)
                        Utilities.displayEdge(nextEdge, true);
                }
                
            }// End of for, edge iterator.
            
        }// End of while; node iterator.
        return allEdges;
    }// End of getAllEdges()
    
    private EdgeWeight computeEdgeWeight(Edge e, ArrayList<Node> c){
      Node source=e.getSource(); // Get source node.
      Node destination=e.getDestination(); // Get destination node.
      ArrayList<Edge> edges=source.getIncomingEdges(); // Get all incoming edges
      
      // For each edge whose source node is in c, add 1 to weight
      int weightIn=0;
     
      Iterator <Edge> itr=edges.iterator();
      while(itr.hasNext()){ // Loop over all incoming edges.
          Edge nextEdge=itr.next(); // Get next edge.
          Node s=nextEdge.getSource(); // Get its source.
              if(c.contains(s))
              weightIn++;
      }
      // Compute wight due to outgoing edges of the destination node.
       int weightOut=0;
      edges=destination.getOutgoingEdges(); // Get all outgoing edges
      itr=edges.iterator();
      while(itr.hasNext()){ // Loop over all incoming edges.
          Edge nextEdge=itr.next(); // Get next edge.
          Node d=nextEdge.getDestination(); // Get its destination.
          if(c.contains(d))
              weightOut++;
      } 
      return (new EdgeWeight(weightIn+weightOut, weightIn, weightOut));  
  }// End of computeEdgeWeights()
    
    // The next method "removes" an edge from a node.
    
    private void removeEdgeFromComponent(ArrayList<Node> c, Edge e){
        
        for(int i=0; i<c.size(); i++){
          Node n=c.get(i); // Extract the next node from component c.
          ArrayList<Edge> edges=n.getOutgoingEdges();
          for (int j=0; j<edges.size(); j++){
              Edge nextEdge=edges.get(j);
              if(nextEdge.equals(e)){
                  e.delete(deletionCause);
                  deletedEdges.add(e);
                  Utilities.debug((Utilities.cycleBreakingDebug||Utilities.minorLevelProcessingDebug), 
                                  "    Edge after deletion: "+
                                  Utilities.stringifyEdge(e, false));
              }
          }// End of for; edge iteration in a node.
        }// End of for; node iteration in c.
          
    }// End of removeEdgeFromComponent()
    
    // Find nodes at all levels less than majorLevel that have
    // association edges coming into a node k in c. Each such k
    // with at least one incoming association edge is added to stubs.
    
    private ArrayList<Node> getStubs(int majorLevel, ArrayList<Node> c){
        
        ArrayList<Node> stubs=new ArrayList<Node>();
        if(majorLevel==1){
            return stubs; // There are no nodes at majorLevel<1, hence stubs is empty.
        }else{ 
            for(int i=1; i<majorLevel; i++){ // Iterate over major levels.
                ArrayList<Node> nodes=g.getNodes(i); // Get all nodes at level i.
                if(Utilities.stubCreationDebug){
                  System.out.println("Nodes at major level "+i);
                  Utilities.displayNodes(nodes, false, false);
                }
                Iterator<Node> itrNode=nodes.iterator();
                while(itrNode.hasNext()){ // Iterate over nodes at a level
                    Node nextNode=itrNode.next();
                    ArrayList<Edge>edges=nextNode.getOutgoingEdges();
                    Iterator<Edge> itrEdge=edges.iterator();
                    while(itrEdge.hasNext()){ // Iterate over edges coming out of a node
                        Edge e=itrEdge.next();
                        if(e.getLabel().equals("eA")){ // Only the association edges are considered.
                            String destinationLabel=e.getDestination().getLabel();
                            Node n=checkNode(destinationLabel, c); // Find if any node in c matches destination node.
                            if(n!=null){
                                stubs.add(n);
                                nextNode.addToStubsUsed(nextNode);
                                Utilities.debug(Utilities.stubCreationDebug, "Node "+nextNode.getLabel().substring(1)+
                                                   " uses stub for "+destinationLabel.substring(1));
                            }// End of null check if
                        }// End of edge check if 
                    }// End of inner while
                }// End of outer while 
            } // End of for 
        }// End of else that checks majorLevel
        return stubs;
    }// End of getStubs()
    
    
    // The following method finds and returns all edges that 
    // are incident on nodes in stubs and originate at
    // nodes in component c. This is part of TD algorithm Step 3.2(a).
    
    private ArrayList<Edge> findEdgesToStubs(ArrayList<Node> stubs, ArrayList<Node> c){
        ArrayList<Edge> edges=new ArrayList<Edge>();
        if(stubs.size()>0&& c.size()>0){
            Iterator<Node> itrNode=c.iterator();
            while(itrNode.hasNext()){
                Node nextNode=itrNode.next();
                ArrayList<Edge>nodeEdges=nextNode.getOutgoingEdges();
                Iterator<Edge> itrEdge=nodeEdges.iterator();
                while(itrEdge.hasNext()){
                    Edge e=itrEdge.next();
                    String destNodeLabel=e.getDestination().getLabel();
                    Node n=checkNode(destNodeLabel, stubs);
                    if(n!=null){
                        edges.add(e); 
                    }  
                }// End of inner  while
            } // End of outer while  
        }// End of if
        return edges;
        
    }// End of findEdgesToStubs()
    
    // The next method encodes steps 3.2(a) of the TD algorithm.
    // It breaks cycles in a given SCC (denoted by c).
    // This is done by removing, one at a time, edges from SCC.
    // It is possible that all cycles are removed by deletion of one or more edges.
    // It is also possible that some cycles remain even after all edges have been deleted.
    // If ycles remain after all edges have been deleted then Step 3.2(b)
    // is applied at a later stage in a different method.
    
    // This method does not actually remove edges. Instead,
    // it relabels all such edges by appending an asterick (*) to the edge label.
    
    // To determine whether or not there remains a cycle after
    // an edge has been deleted, it makes use of Tarjan's algorithm
    // after each deletion. Boy! This is expensive! There must be an easier way.
    
    private ArrayList<Edge> removeEdges(ArrayList<Edge>edgesToStubs, ArrayList<Node>c){
        ArrayList<Edge> deletedListOfEdges=new ArrayList<Edge> ();
        
        
        if(Utilities.cycleBreakingDebug){
            if(edgesToStubs.size()>0){
                System.out.println("     Attempting edge deletion to minimize stubs.");
            }
        }
        
        // First try breaking cycles by deleting edges.
        
        Iterator<Edge> itrEdges=edgesToStubs.iterator(); // Create an iterator for edges to the stubs.
        
        while(itrEdges.hasNext()){ // Iterate over all edges.
            Edge nextEdgeTobeDeleted=itrEdges.next(); // Try to delete this one.
            
            if(Utilities.cycleBreakingDebug)
                System.out.print("     Edge to be deleted: ");
            
            String labelTobeDeleted=nextEdgeTobeDeleted.getLabel(); // Label of the edge to be deleted.
            Iterator <Node>itrNode=c.iterator(); // To iterate over all nodes in component c (an SCC).
            
            while(itrNode.hasNext()){
                Node nextNode=itrNode.next(); // Get the next node.
                ArrayList<Edge> edges=nextNode.getOutgoingEdges(); // Get all its outgoing edges.
                Iterator<Edge> itrD=edges.iterator(); // Iterate over the edges.
                int nextEdge=0;
                while(itrD.hasNext()){
                    Edge maybeDeleted=itrD.next(); // Get edge to be deleted.
                    String label=maybeDeleted.getLabel(); // Get its label.
                    if(label.equals("eA")&&label.equals(labelTobeDeleted)){ // Do the labels match?
                        maybeDeleted.delete(deletionCause);
                        Utilities.debug(Utilities.cycleBreakingDebug,"Edge deleted: "+
                                        nextNode.getLabel().substring(1)+"-"+"eA*"+
                                        maybeDeleted.getDestination().getLabel().substring(1));  
                        
                    } 
                }// End of innermost while
                
                
                Graph g=new Graph(c); // Create a graph from the input component nodes.
                SccTarjan sccAlg=new SccTarjan(g); // Create an object to compute SCCs.
                ArrayList<ArrayList<Node>> componentScc=sccAlg.tarjan(); // Find strongly connected components.
                Utilities.displayAllScc(componentScc); // Display SCCs for this component(for debugging).
            } // End of outer while
        }// End of outermost while
        
        return deletedListOfEdges; // This may be empty.
        
    }// End of removeEdges()
    
    // The next method trims the node list
    // successor by removing from it all nodes that are not in 
    // graph g.
    
    private ArrayList<Node> removeNodesNotInGraph(Graph g, ArrayList<Node> successor){
        ArrayList<Node> component=g.getNodes(); // Get all nodes in the subgraph.
        Iterator <Node> itrOverSuccsessor=successor.iterator();
        ArrayList<Node> updatedSuccessor=new ArrayList<Node>();
        Node n; // Temporary variable.
        
        // Loop over all nodes in succsessor and
        // remove nodes that are NOT in g.
        while (itrOverSuccsessor.hasNext()){
            n=itrOverSuccsessor.next();
            if(checkNode(n.getLabel(), component)==null){
                Utilities.debug(Utilities. minorLevelProcessingDebug, 
                                "Node: "+n.getLabel().substring(1)+" removed from ML");
            }else{
                updatedSuccessor.add(n);
            }// End of if 
        }// End of successor iterator loop  
        return updatedSuccessor;
    } // End of removeNodesNotInGraph()
    
    // Search for a node in a list of nodes.
    // If the node is found then it itself is returned
    // else a null value is returned.
    
    private Node checkNode(String label, ArrayList<Node> component){
        Iterator<Node> itr=component.iterator();
        while(itr.hasNext()){
            Node n=itr.next();
            if(n.getLabel().equals(label)){
                return n;
            }// End of if
        }// End of while search loop
        return null; // No node matches label.
    }// End of checkNode()
    
    private Graph updateGraph(Graph toUpdate, Graph fromUpdate){
        ArrayList<Node> nodesToUpdate=toUpdate.getNodes();
        ArrayList<Node> nodesFromUpdate=fromUpdate.getNodes();
        ArrayList<Node> temp=new ArrayList<Node>();
        
        
        int toUpdateSize=nodesToUpdate.size();
        int fromUpdateSize=nodesFromUpdate.size();
        
        
        for(int i=0; i<fromUpdateSize; i++){
            Node nextFromNode=fromUpdate.getNode(i);
            String fromLabel=nextFromNode.getLabel();
            boolean updated=false;
            for(int j=0; j<toUpdateSize; j++){ 
                Node n=toUpdate.getNode(j);
                if(n.getLabel().equals(fromLabel)){
                    n.setLevel(nextFromNode.getLevel(1), 1);
                    toUpdate.nodes.set(j, n);
                    updated=true;
                    break; // Exit loop as node has been updated.
                }
                
            }// End of inner while. One node in original graph updated.
            if(!updated){ // Updated must be false at this point, else program error.
                System.out.println("\nNode: "+nextFromNode.getLabel().substring(1)+" not found in original graph\n");
                System.exit(0);   
            }
        } // End of outer while. One from node considered.
        
        return toUpdate;
    }
    
  
    
    // Convert majorMinor to a string for aid in printing.
    private String majorMinorString(int majorMinor){
        
        if(majorMinor==0)
            return "major";
        else
            return "minor";
    }// End majorMinorString
    
    
    // The next method generates a node (class) test seqeunce from the modified graph. 
    // The inout graph does not contain any cycles. It has 
    // major and minor level numbers assigned to each node.
    
    public void generateTestSequence(Graph g){
        // testSequence contains a list of sets of
        // nodes (classes) in the order in which they are
        // recommended for testing
        
        testSequence=new ArrayList<ArrayList<Node>>();
        
        // The stubs object contains a list of nodes (classes) for which a stub is needed.
        // retest object contains classes to be tested and the corresponding edgees to be retested.
        
        stubs=new ArrayList<Stub>();
        retestNodes=new ArrayList<Retest>();
        
        for(int i=0; i<maxMajorLevel; i++){ 
            for(int j=0; j<maxMinorLevel[i]; j++){
                Utilities.debug(Utilities.testSequenceProcessingDebug, "\nProcessing level: "+(i+1)+"."+(j+1));
                // Step 1.
                // Begin by finding classes that are at the same level (i.j) and are connected
                // by association edges. These are to be integrated together.
                ArrayList<Node> integrateTogether=findClassesConnectedByAEdges(i+1, j+1);
                Utilities.debug(Utilities.testSequenceProcessingDebug, 
                                "    Nodes connected by A edges at level "+(i+1)+"."+(j+1));
                testSequence.add(integrateTogether);
                if(integrateTogether.size()>0 && Utilities.testSequenceProcessingDebug)
                    Utilities.displayNodes(integrateTogether, false, false);
                else
                    Utilities.debug(Utilities.testSequenceProcessingDebug,
                                    "    No nodes connected by A edge at level "+(i+1)+"."+(j+1));
                
                // Now do the following for each node n at level i.j.
                
                ArrayList<Node> nodesAtThisLevel=g.getNodes(i+1, j+1); // Get all nodes at level i.j
                if(Utilities.stubCreationDebug){
                    System.out.println("Nodes to be checked for stub creation and retesting.");
                    Utilities.displayNodes(nodesAtThisLevel, true, false);
                }
                
                Iterator<Node> itr=nodesAtThisLevel.iterator();// Set to iterate over all nodes at level i.j.
                while(itr.hasNext()){
                    Node n=itr.next(); // Get next node.
                    // Step 2:
                    // Process node n that has outgoing edges to nodes at
                    // higher levels (major or minor).  All such  destination nodes of n
                    // need to be stubbed.
                    
                    processNodesAtHigherLevels(n, i+1, j+1);
                    
                    // Step 3:
                    // Process each node at level i.j that has incoming edges to nodes from
                    // lower levels. The destination nodes need to be retested for 
                    // the edges incoming from lower levels. 
                    
                    processNodesAtLowerLevels(n, i+1, j+1);    
                }// End of loop to process nodes at level i.j.
            }// End of minor level (j) loop.    
        }// End of major level (i) loop.
        
    }// End of generateTestSequence()
    
    // Find classes at level i.j that are connected
    // by Association (A) edges. These are to be integrated together.
    
    private ArrayList<Node> findClassesConnectedByAEdges(int majorLevel, int minorLevel){
        ArrayList<Node> integrateTogether=new ArrayList<Node>();  // Output nodes list.
        ArrayList<Node> nodes=g.getNodes(majorLevel); // Get all nodes at this majorLevel.
        Iterator <Node> itrN=nodes.iterator();
        while(itrN.hasNext()){
            Node n=itrN.next();
            Utilities.debug(Utilities.testSequenceProcessingDebug, "    Processing node: "+n.getLabel().substring(1));
            if(!n.getMark()){
                int sourceNodeMinorLevel=n.getLevel(1); // Get its minor level
                ArrayList<Edge> edges=n.getOutgoingEdges(majorLevel); // Get all outgoing edges to nodes at this majorLevel.
                if(Utilities.testSequenceProcessingDebug)
                    Utilities.displayEdges(edges, false);
                
                Iterator<Edge> itrE=edges.iterator();
                while(itrE.hasNext()){
                    Edge e=itrE.next();
                    Node destNode=e.getDestination();
                    int destNodeMinorLevel=destNode.getLevel(1); // Get its minor level.
                    if(sourceNodeMinorLevel==destNodeMinorLevel&& ( 
                                                                   e.getLabel().equals(deletedEdgeLabel) || e.getLabel().equals("eA"))){
                        integrateTogether.add(n); // Add it to the  nodes list.
                        integrateTogether.add(e.getDestination());// and this one too.
                        n.markVisited();// Mark it  as added to nodes list,
                        destNode.markVisited(); // and mark this one too.
                        Utilities.debug(Utilities.testSequenceProcessingDebug, "    Nodes "+n.getLabel()+" and"+
                                        destNode.getLabel()+" to be interated together.");
                    }
                }// End of edge loop for a node
            }// End of check for visited node.
        }// End of loop over all nodes at level (i.j).
        return integrateTogether; // Return the list of nodes to be integrated together.
    }// End of findClassesConnectedByAEdges()
    
    // For each class at level (i.j) find connected nodes
    // that are at a higher level (major or minor). These need to be stubbed.
    
    private void  processNodesAtHigherLevels(Node n,  int majorLevel, int minorLevel){
        
        boolean stubAdded=false; // Stub not added yet!
        ArrayList<Node> newList=new ArrayList<Node>();
        Utilities.debug(Utilities.stubCreationDebug, "Next node to be tested: "+n.getLabel().substring(1)); 
        newList.add(n);
        testSequence.add(newList);
        
        ArrayList<Edge>edges=n.getOutgoingEdges();
        if(edges.size()==0){
            Utilities.debug(Utilities.stubCreationDebug, 
                            "    No stub needed");
        }else{
            Iterator<Edge>  itrE=edges.iterator(); // Create an iterator to iterate over all edges of node n.
            Stub s=null;// This might become a stub.
            boolean nodeUsesThisStub=false;
            int index; // Used to index in the stubs list.
            while(itrE.hasNext()){
                Edge e=itrE.next(); // Get next edge.
                Node dest=e.getDestination(); //Get its destination node, and
                int destMajorLevel=dest.getLevel(0);// its major and 
                int destMinorLevel=dest.getLevel(1); // minor levels.
                if((destMajorLevel>majorLevel) || (destMajorLevel==majorLevel && destMinorLevel>minorLevel)){
                    stubAdded=false; // Stub not added yet!
                   
                    index=stubExistsForThisNode(dest);
                    if(index==-1){
                        stubAdded=true; // Stub for this node does not exit, so
                        s=new Stub(dest);// create a stub for the destination node and  
                        s.addUsedByNode(n); // // Add n to the list of nodes that use the stub.
                        stubs.add(s); // add it to the list of all stubs. 
                        n.addToStubsUsed(dest);// Add the stub node and a stub used by node n.
                        Utilities.debug(Utilities.stubCreationDebug, 
                                        "    Stub needed for class "+dest.getLabel().substring(1)+ " when testing class  "+ 
                                        n.getLabel().substring(1));
                    }else {// Stub for this node (dest) already exists
                        Utilities.debug(Utilities.stubCreationDebug, 
                                        "    Needed stub for class "+ dest.getLabel().substring(1)+" exists." );
                        Stub existingStub=stubs.get(index); // Find which stub it is and
                        existingStub.addUsedByNode(n); // add node n to the list of nodes that use the stub for dest node.
                        n.addToStubsUsed(dest);// Add the stub node ad a stub used by node n.
                    }// End of else part (stub already exists)
                }// End of level check.
            }// End of edge loop 
            if(!stubAdded)
                Utilities.debug(Utilities.stubCreationDebug, 
                                "    No (new) stub needed.");
        }// End of if to check edge count.
    }// End   processNodesAtHigherLevels()
    
    // For each class at level (i.j) find connected nodes
    // that are at a lower level. These need to be retested
    
    private void  processNodesAtLowerLevels(Node n, int majorLevel, int minorLevel){
        
        // Get the list of nodes, if any, that use the stub created
        // for node n and have their level numbers less than (majorLevel.minorLevel).
        // List returnd by getUseStub() is in increasing order of level numbers.
        
        Utilities.debug(Utilities.retestCreationDebug, "\n   Next node to be checked for re-testing: "+
                        n.getLabel().substring(1));
        ArrayList<Node> nodes=new ArrayList<Node>();
        nodes=getUseStub(n, majorLevel, minorLevel); 
        
        if(nodes==null)
            return;
        int m=nodes.size();
        int index; // Used to search in retestNodes list.
        for(int i=0; i<m; i++){
            Node next=nodes.get(i);
            ArrayList<Edge> edges =next.getOutgoingEdges();
            Iterator<Edge> itr=edges.iterator();
            while(itr.hasNext()){
                Edge e=itr.next();
                Utilities.debug(Utilities.retestCreationDebug,"     Next edge checked for re-testing: "+
                                Utilities.stringifyEdge(e, true));
                if(e.getDestination().equals(n)){
                    index=nodeInRetestList(n); // Find if node n has already been added.
                    if(index==-1){
                        Retest r=new Retest(n); // It has not been added, so add it and
                        r.addEdgeToBeRetested(e); // the edge that causes the stu (node n) to be used, to
                        retestNodes.add(r); // the list of nodes to be retested.
                        Utilities.debug(Utilities.retestCreationDebug, "     Edge "+Utilities.stringifyEdge(e, true)+ 
                                        " to be retested.");
                    }else{
                        Retest nodeInRetest=retestNodes.get(index);
                        nodeInRetest.addEdgeToBeRetested(e);
                    }
                }else{
                    Utilities.debug(Utilities.retestCreationDebug, "     Edge not included for retesting ");
                }// End of node equality check     
            }// End of edge iteration.
        }// End of iteration over nodes that use n as stub
    }// End   processNodesAtLowerLevels()
    
    
    // Search for node n in the list of stubs already
    // created. Return the index of the stub if it already exists, 
    // otherwise return -1.
    private int stubExistsForThisNode(Node n){
        if(stubs.size()==0){
            return -1;
        }
        
        Iterator <Stub> itr=stubs.iterator();
        int index=0;
        while(itr.hasNext()){
            Stub s=itr.next();
            Node stubNode=s.getStubNode();
            if(n.equals(stubNode)){
                return index;
            }
            index++; // Move to the next element in stubs.
        }// End of iteration over stubs
        // Stub does not exist.
        return -1;     
    }//End of stubExistsForThisNode()
    
    // Find and return all nodes that are at a level less than the
    // input major-minor level of node n and have outgoing edges to n,
    // i.e., they use the stub of node n.
    
    // Note that the list of stubs has at most one entry for each stub.
    // And this entry contains all the nodes that use n.
    
    private ArrayList<Node> getUseStub(Node n, int majorLevel, int minorLevel){
        Iterator<Stub> itrStub=stubs.iterator();
       
        while(itrStub.hasNext()){
            Stub s=itrStub.next();
            Node nextNode=s.getStubNode();
            if(nextNode.equals(n)){ // Check if this stub is for node n.
               return (s.getUsedByNodes()); // Yes it is, hence get all nodes that use stub of n and return.
            }
        }// End of loop over stubs created so far.
        return null; // No nodes use the stub for  node n.

    }// End of getUseStub();
    
    
    // Return the index of node n in the retest list.
    // A -1 is returned if n is not found in the retest list.
    private int nodeInRetestList(Node n){
        Iterator<Retest> itr=retestNodes.iterator();
        int index=0;
        while(itr.hasNext()){
            Retest r=itr.next();
            Node next=r.getNodeUnderTest();
            if(next.equals(n)){
                return index;
            }
            index++;   
        }// End of loop over retestNode.
        return -1; // Node not found in retest list.
    }// End of nodeInRetestList()
    
    
    // Get ORD file names and any test data generated from previous test orders.
    private  static void getORDFileNames(){
   
      // Get names of all ORD files in the ordFolder.
      
      File f=new File(ordFolder); // Contains all ORDs on which this algorithm is to be applied.
      String [] fileName=f.list(); // Get list of all ORD files in the folder.
      ordFiles=new ArrayList<String>();
      
      // Filter out all useful (ORD) files.
      for (int i=0; i<fileName.length; i++){
        String nextFile=fileName[i];
        if(nextFile.endsWith(".txt")){// Select only files whose name ends with .txt.
          ordFiles.add(nextFile);
        }
      }// End of file filter loop.
      System.out.println(ordFiles.size()+" ORD files found. ");
      
      // Get existing data in the testOrderStatistics file.
      File stubFile=new File(testOrderStatistics);
      Scanner s=null;
      boolean fileOpened=false; // Did the input file open successfully?
      try{
        s=new Scanner (stubFile);
        fileOpened=true;
      }catch(Exception e){
        System.out.println("Past test summary file does not exist. New file will be opened.");
      }
    
      // Read all data from the inout file and 
      // save it in memory in testData as a string.
      testData=""; // Initialize the area where past data is to be saved and new will be added.
      if(fileOpened){
         // This will hold all the past data obtained from the file.
        String lastUpdate=s.nextLine();
        System.out.println(lastUpdate);
         while(s.hasNext()){
          String nextLine=s.nextLine();
          testData=testData+"\n"+nextLine; 
        }// End of test data input loop.
        
        System.out.println("Existing test order summary  saved in memory.");
        s.close(); // Close this scanner.
      }
      
    }// End of getORDFileNames()
    
   // Get ORD file names and any test data generated from previous test orders.
    private  static void saveTestOrderSummary(int stubs, int specificStubs, 
                                               double timeTaken, boolean lastCall){
   
      File outFile=new File(testOrderStatistics);
      PrintWriter p=null;
      boolean fileOpened=false; // Did the output file open successfully?
      try{
        p=new PrintWriter(outFile);
        fileOpened=true; // Yes, now it did!
      }catch(Exception e){
        System.out.println(testOrderStatistics+" file could not be opened. Summary not saved.");
      }
      
      if(fileOpened){
        p.println("Last update: "+Utilities.todayInfo());
        testData=testData+"\tStubs:"+ stubs+"\tSpecific stubs: "+specificStubs;
        String secMin=" second(s)";
        if(timeTaken>60){
          secMin=" minute(s)";
            timeTaken=timeTaken/60;
        }
        String formattedTimeTaken=String.format("%.3f", timeTaken);
        testData=testData+"\tTime: "+formattedTimeTaken+secMin;
        if(lastCall){
          testData=testData+"\n\n ------End of a run of the Tai-Daniels Algorithm-------\n";
        }
        p.print(testData);
        p.close();
        System.out.println("Test order summary appended to file "+testOrderStatistics);
      }// End of file update
      
    }// Endof  saveTestOrderSummary
  
    
    // Program to test the implementation of TD algorithm.
    
    public static void main(String[] args) throws IOException{
       
      // TaiDanielsAlgorithms t=new TaiDanielsAlgorithms("../TestGraphs/TaiDanielsGraph.txt");
      // TaiDanielsAlgorithms t=new TaiDanielsAlgorithms("../TestGraphs/LabichExample2000.txt");
      // TaiDanielsAlgorithms t=new TaiDanielsAlgorithms("../TestGraphs/BriandExample2003.txt");
      // TaiDanielsAlgorithms t=new TaiDanielsAlgorithms("../TestGraphs/fullyConnected4Nodes.txt");
      // TaiDanielsAlgorithms t=new TaiDanielsAlgorithms("../TestGraphs/fullyConnected2Nodes.txt");
      // TaiDanielsAlgorithms t=new TaiDanielsAlgorithms("../TestGraphs/fullyConnected3Nodes.txt");
      // TaiDanielsAlgorithms t=new TaiDanielsAlgorithms("../TestGraphs/fullyConnected4Nodes.txt");
      // TaiDanielsAlgorithms t=new TaiDanielsAlgorithms("../TestGraphs/fullyConnected5Nodes.txt");
      // TaiDanielsAlgorithms t=new TaiDanielsAlgorithms("../TestGraphs/LeTraonExample.txt");
      // TaiDanielsAlgorithms t=new TaiDanielsAlgorithms("../TestGraphs/LeTraonExampleV1.txt");
      
      
      // TaiDanielsAlgorithms t=new TaiDanielsAlgorithms("../TestGraphs/fullyConnected6Nodes.txt");
      // TaiDanielsAlgorithms t=new TaiDanielsAlgorithms("../TestGraphs/TaiDanielsAdditionalStubTest1.txt");
      // TaiDanielsAlgorithms t=new TaiDanielsAlgorithms("../TestGraphs/MinimalRetestExample.txt");
      // TaiDanielsAlgorithms t=new TaiDanielsAlgorithms("../TestGraphs/MinimalRetestExample2.txt");
      // TaiDanielsAlgorithms t=new TaiDanielsAlgorithms("../TestGraphs/MinimalRetestExample3.txt");
      
      // TaiDanielsAlgorithms t=new TaiDanielsAlgorithms("../TestGraphs/DAG4Nodes.txt");
      
      // TaiDanielsAlgorithms t=new TaiDanielsAlgorithms("../TestGraphs/BriandExample2003OneEdgeRemoved.txt");
      // TaiDanielsAlgorithms t=new TaiDanielsAlgorithms("../TestGraphs/BriandExample2003TwoEdgesRemoved.txt");
      // TaiDanielsAlgorithms t=new TaiDanielsAlgorithms("../TestGraphs/BriandExample2003ThreeEdgesRemoved.txt");
      // TaiDanielsAlgorithms t=new  TaiDanielsAlgorithms("../TestGraphs/BriandExample2003OneEdgeOneNodeRemoved.txt");
      // TaiDanielsAlgorithms t=new TaiDanielsAlgorithms("../TestGraphs/BriandExample2003OneEdgeTwoNodesRemoved.txt");
      
      // Set 1
      
      // TaiDanielsAlgorithms t=new TaiDanielsAlgorithms("../TestGraphs/TaiDanielsGraph.txt");
      // TaiDanielsAlgorithms t=new TaiDanielsAlgorithms("../TestGraphs/LeTraonExampleV1.txt");
      // TaiDanielsAlgorithms t=new TaiDanielsAlgorithms("../TestGraphs/BriandExample2003.txt");
      
      // Set 2
      
      //  TaiDanielsAlgorithms t=new TaiDanielsAlgorithms("../TestGraphs/SecondEditionExperiments/LeTraonExampleV2.txt");
      // TaiDanielsAlgorithms t=new  TaiDanielsAlgorithms("../TestGraphs/SecondEditionExperiments/N15E42.txt");
      // TaiDanielsAlgorithms t=new  TaiDanielsAlgorithms("../TestGraphs/SecondEditionExperiments/N15E78.txt");
      // TaiDanielsAlgorithms t=new  TaiDanielsAlgorithms("../TestGraphs/GraphN15E99.txt");
      TaiDanielsAlgorithms t=new  TaiDanielsAlgorithms("../TestGraphs/SecondEditionExperiments/N15E134.txt");
      
      // Set 3
      
      // TaiDanielsAlgorithms t=new  TaiDanielsAlgorithms("../TestGraphs/SecondEditionExperiments/N100E1452.txt");
      // TaiDanielsAlgorithms t=new  TaiDanielsAlgorithms("../TestGraphs/SecondEditionExperiments/N100E3044.txt");
      // TaiDanielsAlgorithms t=new  TaiDanielsAlgorithms("../TestGraphs/SecondEditionExperiments/N1004581.txt");
      // TaiDanielsAlgorithms t=new  TaiDanielsAlgorithms("../TestGraphs/SecondEditionExperiments/N100E5925.txt");
          
 
      
    
      
      // Set up loop to apply the TD test order generation algorithm.
      boolean lastCall=false; // Last ORD processed?
     
        System.out.println("\n-------------Begin Test Order Determination using the TD algorithm. V1.0. July 29, 2013 ");
        
        if(Utilities.graphInputDebug){
          Utilities.displayGraph(t.g);
        }
        
        // TD algorithm Step 2: Assign major level numbers to nodes.
        System.out.println("\nAssign major level numbers.");
        ArrayList<String> ignoreEdges=new ArrayList<String>();
        ignoreEdges.add("eA"); // Ignore association edges as they may create cycles.
        t.levelADG(t.g, 0, ignoreEdges,0); // Assign major level numbers.
        System.out.println("\n-----Major level numbers assigned. Max major level: "+t.maxMajorLevel);
        
        if(Utilities.graphInputDebug){
          System.out.println("\nMajor levels and nodes.");
          Utilities.displayMajorLevels(t.g, t.maxMajorLevel);
        }
        
        // TD algorithm Step 3: Assign minor level numbers to nodes.
        System.out.println("\nAssign minor level numbers.");
        
        // Only the association edges are to be considered in this step.
        // Hence C, I, and  edges removed are  ignored.
        ignoreEdges=new ArrayList<String>();
        ignoreEdges.add("eI");
        ignoreEdges.add("eC");
        ignoreEdges.add("eA*");
        
        t.maxMinorLevel=new int [t.maxMajorLevel];
        t.assignMinorLevelNumbers( t.g, t.maxMajorLevel, ignoreEdges);
        System.out.println("\n-----Minor level numbers assigned.");
        // System.out.println("\nUpdated Graph:");
        if(Utilities.stubCreationDebug){
          Utilities.printGraphByLevelNumbers(t.g,t.maxMajorLevel, t.maxMinorLevel);
        }
        
        // TD algorithm final step to generate test sequence for classes.
        
        System.out.println("\nGenerate class testing sequence.");
        t.g.resetMarks();
        t.generateTestSequence(t.g);
        System.out.println("-----Class testing sequence generated.");
        // System.out.println("\nStubs needed for the following classes: ");
        // Utilities.displayStubs(t.stubs);
       // System.out.println("\nEdges to be re-tested:");
        
        System.out.println("\nGenerate stubs and retest requirements.");
        // Utilities.displayRetestEdges(t.retestNodes);
        System.out.println("-----Stubs and retest requirements generated.");
        
        if(Utilities.stubCreationDebug){
          System.out.println("Modified graph");
          Utilities.displayGraph(t.g);
        }
        
        Utilities.displayTestOrder(t.g, false, t.stubs, t.retestNodes, 
                                   t.maxMajorLevel, t.maxMinorLevel);
        
        System.out.println("\nEnd of TD algorithm. ");
        
  
      
    }// End of main()
}// End of class TaiDanielsAlgorithms.

