/*
 * This class contains utiliy methods.
 * Most of these are print methods and useful during debugging the 
 * Tai-Daniels (TD) algorithm as
 * well as other agorithms that it uses.
 * 
 * 
 * Author: Aditya Mathur
 * Original: July 7, 2011

 * Latest update: July 24-28, 2011
 * January 5, 2012 (Deletion status of edges is now displayed)
 * January 13, 2012: New stub generation added to displayNodes().
 * 
 */

import java.util.*;
import java.io.*;
import java.lang.Object;
import java.text.DateFormat;
import java.text.SimpleDateFormat;


public  class Utilities{
    
    // Debugging switches. Set to true to make debugging  active
    // false to make it inactive.
    
    static boolean graphInputDebug=false; // Debug graph input method
    static boolean levelAssignmentDebug=false; // Debug for assigning major/minor levels in TD algorithm.
    static boolean tarjanDebug=false;// Debug Tarjan's SCC algorithm
    static boolean edgeDeletionDebug=false; // Debug edge deletion algorithm
    static boolean cycleBreakingDebug=false; // Debug the cycle breaking portion of the TD algorithm.
    static boolean cycleCheckingDebug=false; // Debug the cycle checking portion of the TD algorithm.
    static boolean majorLevelProcessingDebug=false; // Debug for major level portion of the TD algorithm.
    static boolean minorLevelProcessingDebug=false; // Debug for minor portion of the TD algorithm.
    static boolean testSequenceProcessingDebug=false; // Debug for test sequence generation portion of the TD algorithm.
    static boolean stubCreationDebug=false; // Debug for stub generation portion of the TD algorithm.
    static boolean retestCreationDebug=false; // Debug for retest class portion of the TD algorithm.
    static boolean dependenceCreationDebug=false; // Debug for retest class portion of the TD algorithm.
    static boolean newStubCreationDebug=false; // Debug for retest class portion of the TD algorithm.
    static boolean miscDebug=false;
    static boolean testOrderDebug=false;
    
    // When the next boolean is true use TD algorithm as the authors have described it.
    // Otherwise assign minor level numbers based on association edges deleted.
    // This variable is used in the levelSort() method. Change made on Jan 12, 2012.
    
    static boolean strictTaiDaineisl=false; 
    static boolean createTestOrder=false; // Set to true when test order is to be created.
    
    // Useful constants for finding various types of node dependencies in the graph.
    final static int staticChain=1;
    final static int dynamicChain=2;
    final static int inheritanceChain=3;
    
    private static int specificStubs=0; // Specific stub is a incarnation of a stub for a specific class.
    
    // Declare a list of stubs generated while scanning nodes at various levels.
    // These stubs are the ones that are needed but not generated by the Tai-daniels algorithm.
    // This is used in the displayNodes method.
    // Change made: Jan 13, 2012.
    
    private static ArrayList<Node> newStubs=new ArrayList<Node> ();
    
    // This variable keeps track of any possible program errors during execution.
    ArrayList<Integer> programError=new ArrayList<Integer>();
    
    // Display test order from the graph.
    // Classes are tested in the increasing order of their respective major and minor levels.
    
    public static void displayTestOrder(Graph g, boolean displayGraph, ArrayList<Stub> s, ArrayList<Retest> r, 
                                        int maxMajorLevel, int [] maxMinorLevel){

        specificStubs=0;
        if(displayGraph){
            System.out.println("\nUpdated Graph:");
            displayGraph(g);
        }
        
        createTestOrder=true;
        Utilities.debug(testOrderDebug, "\nClass test order");
        // Process nodes at major levels.
        
      
        for(int i=1; i<=maxMajorLevel; i++){
          if(maxMinorLevel[i-1]==0){
            System.out.println("** Program error ** No nodes at major level "+i);
          }
          
          // For each major level i, process nodes at minor levels.
          for(int j=1; j<=maxMinorLevel[i-1]; j++){
            ArrayList<Node> nodes=g.getNodes(i,j);
              displayNodes(nodes, false, false);  
            
          }// End of minor level loop
          
        }// End of major level loop
        
        int newStubsCreated=newStubs.size();// Get number of additional stubs created. 
        String sOrNot=singPlu(newStubsCreated); // Get string for correct grammatical display.
        
        if(newStubsCreated==0&& stubCreationDebug){
          System.out.println("No additional stubs created.");
        }else{
          System.out.println(" Additional stub"+sOrNot+ " created for node"+sOrNot+stringifyNodes(newStubs));
          
        }
        
        sOrNot=singPlu(s.size()+newStubsCreated); // Get string for correct grammatical display.
        
        if(stubCreationDebug){
          displayStubsCreated(s);
        }
        
        // Display summary.
        System.out.println("\n\tTotal stub"+sOrNot+ " created: "+(s.size()+newStubsCreated));
        System.out.println("\tSpecific stubs created: "+(specificStubs+newStubsCreated));
        System.out.println("\nRetest nodes and edges");
        if(testSequenceProcessingDebug){
          displayRetestEdges(r);
        }

    }// End displayTestOrder()
   
    
    
     
     public static int getSpecificStubCount(){
      return specificStubs;
    }
     
    // Debug routine to display all nodes in a list.
    // Outgoing edges are displayed only when edgeAlso is true.
    
    
    static ArrayList<Node>succNodes;
    public static void displayNodes(ArrayList<Node> list, boolean edgesAlso, 
                                    boolean withWeight){
        
        if(list.size()==0){
            System.out.println("\tNo nodes to display.\n");
            return;
        }
        // List is not empty, hence display it.
        Node prevNode=null; // Previous node on this list.
        
        Iterator itr =list.iterator();
        while(itr.hasNext()){
            Node nextNode=(Node) itr.next();
            // Check if prevNode depends on nextNode through an association edge
            // if tst order is being generated.
            if(createTestOrder){
                createStub(prevNode, nextNode);
                prevNode=nextNode; // Update previous node processed.
            }
            
            String abs="";
            if(nextNode.isNodeAbstract())
                abs=" (Abstract) ";
            
            if(miscDebug){
              System.out.println("\tClass "+nextNode.getLabel().substring(1)+abs+" Major.Minor Levels: "+
                                 nextNode.getLevel(0)+"."+nextNode.getLevel(1));
              displayDependence(nextNode, edgesAlso, withWeight, Utilities.staticChain);
              displayDependence(nextNode, edgesAlso, withWeight, Utilities.inheritanceChain);
              displayDependence(nextNode, edgesAlso, withWeight, Utilities.dynamicChain);
            }
              displayStubsUsed(nextNode);
          
            
        } // End of while  
        
         if(miscDebug){
           System.out.println("\t---------------------");
         }
        
    } // End of displayNodes()
    
    // THis method checks if the prevNode is linked to the nextNode by an association
    // edge. In case no stub has been created for nxtNode then a new stub is created.
    
    private static void createStub(Node prevNode, Node nextNode){
        if(prevNode!=null){
            debug(newStubCreationDebug, "Check nodes "+prevNode.getLabel().substring(1)+" and "+
                  nextNode.getLabel().substring(1)+ " for association");
            if(relatedByAssociation(prevNode,nextNode)){
                if(!nextNode.isStub()){
                    newStubs.add(nextNode);
                    if(newStubCreationDebug){
                        System.out.println("\nNew stub created for node: "+nextNode.getLabel().substring(1));
                    }
                }
            }else{
                if(prevNode!=null){
                    debug(newStubCreationDebug, "\nNo association edge from "+prevNode.getLabel().substring(1)+
                          " to "+ nextNode.getLabel().substring(1)+".");
                }
            }
        }
    }// End of createStub()
    
    // Display dependent nodes (both static and dynamic).
    private static void displayDependence(Node n, boolean edgesAlso, boolean withWeight, int depType){
        ArrayList<Node> dependsOn=null;
        String dep="";
        switch (depType){
            case 1: {
                dependsOn=n.getStaticDependence();
                dep="Static dependency"; break;
            }
            case 2: {
                dependsOn=n.getDynamicDependence();
                dep="Dynamic dependency"; break;
            }
             case 3: {
                dependsOn=n.getInheritanceChain();
                dep="Inheritance chain"; break;
            }
            default: {System.out.println("Incorrect dependence type in displayDependence");
                break;
            }
        }// End of switch
        
         Iterator<Node> depends=dependsOn.iterator();
            if(dependsOn.size()==0){
                System.out.println("\t    This node has no "+dep); 
            }else{
                System.out.print("\t"+"    "+     dep+": ");
                String filler="";// Used for formatting
                while(depends.hasNext()){
                    Node d=depends.next();
                    System.out.print(filler+d.getLabel().substring(1));
                    filler=", ";
                }
                System.out.println();
                if(edgesAlso){
                    displayEdges(n, withWeight, null);
                }
            }
    }// End of displayDependence()
    
    // Debug routine to display all stubs in a list of stubs.
    
    public static void displayStubsCreated(ArrayList<Stub> list){   
        if(list.size()==0){
            System.out.println("\tNo stubs to display.\n");
            return;
        }
        // List is not empty, hence display it.
        Iterator <Stub> itr =list.iterator();
        while(itr.hasNext()){
            Stub next=itr.next();
            Node stubNode=next.getStubNode();
            ArrayList<Node> usedByNodes=next.getUsedByNodes();
            System.out.print("\n\tStub for: "+stubNode.getLabel().substring(1)+" Level: "+
                             stubNode.getLevel());
            if(usedByNodes.size()<1){
                System.out.println("\n** This stub is not used by any node. Perhaps a program error.");          
            }else{
                System.out.print("  Used by: ");
                // Display all nodes that use this stub.
                
                Iterator <Node> itrNode=usedByNodes.iterator();
                Node nextUsedNode=new Node();
                while(itrNode.hasNext()){
                    nextUsedNode=itrNode.next();
                    System.out.print(nextUsedNode.getLabel().substring(1)+" ");
                }// End of while; display nodes that use this stub
                System.out.println();
                
                // Now display the edges that require the stub to be used.
                itrNode=usedByNodes.iterator();
                while(itrNode.hasNext()){
                    nextUsedNode=itrNode.next();
                    displayEdges(nextUsedNode, false, stubNode);
                    System.out.println();
                }// End of while; displasy edges that require stub use.
            }
        } // End of while  over stubs
    } // End of displayStubsCreated()
    
    private static  void displayStubsUsed(Node n){
      ArrayList<Node> stubsUsed=n.getStubsUsed();
      
      
      if(stubsUsed.size()==0){
        Utilities.debug(Utilities.miscDebug, "\t"+"    "+"No stubs needed when testing this class.");
        return;
      }
      // Else display all nodes whose stubds are used by n.
      String sOrNot=singPlu(stubsUsed.size());
      if(stubCreationDebug){
        System.out.print("\t"+"    "+  "Stub"+sOrNot+" used: ");
      }
      specificStubs++;
      
      for(int i=0; i<stubsUsed.size(); i++){
        if(stubCreationDebug){
          System.out.print(stubsUsed.get(i).getLabel().substring(1)+" ");
        }
      }// End of node loop
      if(stubCreationDebug){
        System.out.println();
      }
      
    }// End of displayStubsUsed()
    
      public static void displayRetestEdges(ArrayList<Retest> list){   
        if(list.size()==0){
            System.out.println("\tNo nodes/edges to display.\n");
            return;
        }
        // List is not empty, hence display it.
         Iterator <Retest> itr =list.iterator();
         while(itr.hasNext()){
             Retest next=itr.next();
             Node nodeToBeTested=next.getNodeUnderTest();
             ArrayList<Edge> edgesToBeTested=next.getEdgesToBeTested();
             System.out.print("\t"+"When testing  "+nodeToBeTested.getLabel().substring(1)+" Level: "+
                              nodeToBeTested.getLevel());
             if(edgesToBeTested.size()>0){
                 System.out.println("  re-test the following edges: ");
                 Utilities.displayEdges(edgesToBeTested, false);
             }

        } // End of while for Retest 
    } // End of displayRetestEdges()

    public static void displayMajorLevels(Graph g, int maxLevel){
        System.out.println("\nMajor levels \n");
        for (int i=1; i<=maxLevel; i++){
            System.out.println("Nodes at major level "+i+ ":");
            Iterator itr=g.getNodes().iterator();
            while (itr.hasNext()){
                Node next=(Node)itr.next();
                int nodeLevel=next.getLevel(0);
                if(nodeLevel==i){
                    System.out.print(next.getLabel().substring(1)+ " ");
                }
            } // End of while
            System.out.println();
        } //End of for
  } // End of displayMajorLevels().  
    
    public static void displayNodeList(ArrayList<Node>n, Node stubNode, boolean edgesAlso, boolean displayLevels) {
        if(n.size()==0){
            System.out.println("Node list is empty.");
            return;
        }
        Iterator itr=n.iterator();
        while(itr.hasNext()){
          Node next=(Node)itr.next();
          System.out.print("Node: "+next.getLabel().substring(1));
          if(displayLevels){
            int nodeLevelMajor=next.getLevel(0);
            int nodeLevelMinor=next.getLevel(1);
            System.out.print(" "+" Major:Minor level: " +nodeLevelMajor+":"+nodeLevelMinor);  
          }
          System.out.println(); 
          if(edgesAlso)
               displayEdges(next, false, stubNode);
          System.out.println();
        } // End of while over nodes
        
    } // End of displayNodeList()
    
      
    // Print a graph: all nodes and the corresponding edges.
    
    public static void displayGraph(Graph g){ 
      
        Iterator itr=g.getNodes().iterator();
        int totalEdges=0;
        int nodeCount=0;
        System.out.println("Graph. Edge nomenclature: A: Association. I: Inheritance. C: Aggregate (Composite)");
        while(itr.hasNext()){ 
            Node n=(Node) itr.next();
            nodeCount++;
            System.out.println("Node: "+ n.getLabel().substring(1)+" "+" Major level: "+ n.getLevel(0)+" "+
                               " Minor level: "+n.getLevel(1));
            System.out.println("Number of incoming/outgoing edges:"+n.getIncomingEdgeCount()+
                               "/"+n.getOutgoingEdgeCount());
            totalEdges=totalEdges+n.getOutgoingEdgeCount();
            displayEdges(n, false, null);
            System.out.println("\n---------");
        }// End of iteration over graph nodes.
        System.out.println("\nNodes: "+nodeCount+" Edges: "+totalEdges);
        System.out.println("\nGraph printed.");
        System.out.println("\n---------");
       
    } // End displayGraph()
    
    
    // Print graph but by level numbers.
    
     public static void printGraphByLevelNumbers(Graph g, int maxMajorLevel, int[] maxMinorLevel){ 
         
         
         int nodeCount=0;
         int edgeCount=0;
         for(int i=1; i<=maxMajorLevel; i++){
             System.out.print("\nNodes at major level "+i);
             System.out.println(". Max minor level: "+maxMinorLevel[i-1]);
             System.out.print("\t");
             
             Iterator <Node> itr=g.getNodes(i).iterator();     
             while(itr.hasNext()){ 
                 Node n=(Node) itr.next();
                 nodeCount++;
                 edgeCount=edgeCount+n.getOutgoingEdgeCount();
                 System.out.print(n.getLabel().substring(1)+" "+
                                  n.getLevel(0)+"."+n.getLevel(1)+"  ");
             }// End of iteration over nodes at level i
             
         }// End of iteration over the major levels.
        
         System.out.println("\nTotal Nodes: "+nodeCount+" Total Edges: "+edgeCount);
     } // End printGraphByLevelNumbers()
    
    // Display all edges emanating from a node. ff stubNode is not null
     // then only the edges ending at stubNode are displayed.
     
    public static void displayEdges(Node n, boolean weightDisplay, Node stubNode){
        String label=n.getLabel();
        ArrayList<Edge> edges=n.getOutgoingEdges();
        Iterator itr =edges.iterator();
        while(itr.hasNext()){
            Edge nextEdge=(Edge)itr.next();
            Node destination=nextEdge.getDestination();
            if(!(stubNode==null)){
                if(destination.equals(stubNode)){ 
                   displayEdge(nextEdge, weightDisplay);
                }
            }else{
                 displayEdge(nextEdge, weightDisplay);
            }
        }// End of while over edges.
    } // End of displayEdges()
        
 
    
    public static void displayEdges(ArrayList<Edge> e , boolean weightDisplay){
        if(e.size()==0){
            System.out.println("No edges to display.");
            return;
        }
        
        Iterator itr =e.iterator();
        while(itr.hasNext()){
            Edge nextEdge=(Edge)itr.next();
            Node source=nextEdge.getSource();
            Node destination=nextEdge.getDestination();
            String labelSource=source.getLabel();
            String labelDestination=destination.getLabel();
            String labelEdge=nextEdge.getLabel();
            int weight=nextEdge.getWeight();
            System.out.print("\t"+labelSource.substring(1)+"-"+labelEdge.substring(1) +"-"+
                             labelDestination.substring(1));
            if(weightDisplay){ 
                System.out.print("\tWeight: "+nextEdge.getWeight()+" ("+nextEdge.getWeightIn()+
                                   "+"+nextEdge.getWeightOut()+")");
            }
            String status=nextEdge.getDeletionStatus();
             if(!status.equals("N"))
                 System.out.println("\tDeletion status: "+nextEdge.getDeletionStatus()); 
             else
                 System.out.println();
        }// End of edge iterator loop.
    } // End of displayEdges, another version
    
    public static String stringifyEdge(Edge e, boolean weightDisplay){
        String edgeString=e.getSource().getLabel().substring(1)+"-"+
            e.getLabel()+"-"+e.getDestination().getLabel().substring(1);    
        
        if(weightDisplay){
            edgeString=edgeString+Integer.toString(e.getWeight());
        } 
        return edgeString;
    }// End of stringifyEdge()
    
    public static String stringifyNodes(ArrayList<Node> list){
      Iterator<Node> n=list.iterator();
      String nodeList="";
      while(n.hasNext()){
        Node nextNode=n.next();
        nodeList=nodeList+" "+nextNode.getLabel().substring(1);
      }
      return nodeList;
    }// End of stringifyNodes()
    
    // Print all strongly connected components.
    
    public static void displayAllScc(ArrayList<ArrayList<Node>> scc){
        
        int componentCount=scc.size();
        if(componentCount==0){
            System.out.println("No strongly connected components found.");
            return;
        }
        
        // Loop through the ArrayList and print all SCCs.
        System.out.println("\n"+componentCount+ " Strongly connected components found\n");
        
        
        Iterator <ArrayList<Node>> itrAll=scc.iterator(); // This iterates over all components.
        int nextComponent=1; // Index of component to be printed.
        while(itrAll.hasNext()){
            System.out.print("Component "+nextComponent);
            ArrayList<Node> component=(ArrayList<Node>)itrAll.next();
            Iterator itrOne=component.iterator(); // This iterates over all nodes in a component.
          
            while(itrOne.hasNext()){  
                Node n=(Node)itrOne.next();
                System.out.print(" "+n.getLabel().substring(1)); // Display all on nodes one line (should be improved!)
            }// End of inner while for one component.
             if(component.size()>1)
                 System.out.println(  " has at least one cycle.\n"); // Move to the next line.
             else
                 System.out.println();
            nextComponent++; 
           
        }// End of outer while for all components.
        
    } // End of displayAllScc()
     
     
     // The following method displays one SCC. It is useful in debugging.
      public static void displayAScc(ArrayList<Node> scc){
        
        int componentSize=scc.size();
        if(componentSize==0)
            return;
        // Loop through the ArrayList and print all SCCs.
        
            System.out.println("Strongly connected component: ");
            Iterator  itr=scc.iterator(); // This iterates over all nodes in the component
            while(itr.hasNext()){    
                    Node n=(Node)itr.next();
                    System.out.print(" "+n.getLabel().substring(1)); // Display all on one line (should be improved!)
                }// End of inner while
                System.out.println(); // Move to the next line.
            }// End of  displayAScc for one component
    
     
     // The next method displays successors of each node in the given graph.
     
     public static void displaySuccessors(Graph g, ArrayList<String> ignoreEdges){
    
         ArrayList<Node> nodes=g.getNodes();
         System.out.println("\nAll succesors.\n");
         Iterator itr=nodes.iterator();
         while(itr.hasNext()){
             Node n=(Node)itr.next();
             ArrayList<Node> succ=n.getSuccessors(ignoreEdges);
             System.out.println("Node: "+n.getLabel());
             displayNodeList(succ,null,  true, false);
         } 
     }// End of displaySuccessors()
     
     public static void displayEdge(Edge e, boolean displayWeight){
         System.out.print("\tEdge: "+e.getSource().getLabel().substring(1)+"-"+e.getLabel().substring(1)+"-"+
                          e.getDestination().getLabel().substring(1));
         if(displayWeight){
             int weightIn=e.getWeightIn();
             int weightOut=e.getWeightOut();
             
             System.out.print("\tWeight: "+e.getWeight()+" ("+weightIn+"+"+weightOut+")");
             String status=e.getDeletionStatus();
             if(!status.equals("N"))
                 System.out.println("\tDeletion status: "+e.getDeletionStatus());
             else 
                 System.out.println();
         }// End of weight display code.
     }// End of display edge.
     
     
     // The next method replaces an edge in some node in graph g
     // by e. Note that there will be exactly one edge in exactly one node
     // in g that will match e. This method is not used in Tai-daniels algorithm.
     
     public static void updateEdgeInGraph(Graph g, Edge e){
      
         String newLabel="eA"+Edge.deleteString;
         ArrayList<Node> nodes=g.getNodes();
         Iterator<Node> itr=nodes.iterator();
         for(int nextNode=0; nextNode<nodes.size(); nextNode++){
             ArrayList<Edge> edges=nodes.get(nextNode).getOutgoingEdges();
             for(int nextEdge=0; nextEdge<edges.size(); nextEdge++){
                 String labelHeadNode1 =edges.get(nextEdge).getSource().getLabel();
                 String labelTailNode1=edges.get(nextEdge).getDestination().getLabel();
                 String labelHeadNode2 =e.getSource().getLabel();
                 String labelTailNode2=e.getDestination().getLabel();
                 String label1=edges.get(nextEdge).getLabel();
                 String label2=e.getLabel();
                 if(labelHeadNode1.equals(labelHeadNode2)&&
                    labelTailNode1.equals(labelTailNode2)&&
                    label2.equals("eA"+Edge.deleteString)){
                     nodes.get(nextNode).getOutgoingEdges().get(nextEdge).delete("u");
                     Utilities.debug(Utilities.cycleBreakingDebug, "        Edge updated in graph");
                 }
     
             }// End of for to iterate over edges.      
         }// End of for to iterate over nodes
         
     }// End of updateEdges()
     
     // Method to display grammatically correct counts.
     // Edge "2 edges" or "1 edge".
     
     public static String singPlu(int n){
         
        if(n>1)
            return "s";
        else 
            return "";
     }// End of singPlu
     
     // This metod find the list of class that on hwich the inut class n depends. 
     // The method understands that there may be a cycle in the graph.
     
    
     public ArrayList<Node>  dependsOn(Node n){
         
         ArrayList<Edge> edges=n.getOutgoingEdges();
         Iterator<Edge> itr=edges.iterator();
         while(itr.hasNext()){
             Edge e=itr.next();
             Node dest=e.getDestination();
             if(dest.equals(n)){
                 succNodes.add(dest);
                 return succNodes;
             }else{
                 succNodes.add(dest);
                 dependsOn(dest);
             }  
         }
         return(succNodes);
     }// End of dependsOn()
     
     public static void debug(boolean debugFlag, String message){
         if(debugFlag)
             System.out.println(message);
     }
     
     // Check if node s is related to node t by an association edge.
     
     private static boolean relatedByAssociation(Node s, Node t){
         // Return false if source node is null.
         if(s==null)
             return false;
         
         ArrayList<Edge> edges=new ArrayList<Edge>();
         
         edges=s.getOutgoingEdges();// Get outgoing edges from the source node.
         Iterator <Edge> e=edges.iterator();
         while(e.hasNext()){
             Edge nextEdge=e.next(); // Get next outgoing edge
             if(newStubCreationDebug){
                 displayEdge(nextEdge, false);
             }
             Node destination=nextEdge.getDestination(); // Get its destination node.
             if(destination.equals(t)){ // is the destination node the target t?
                 if(nextEdge.isAssociationEdge()){ // Is it an association edge?
                     return true;
                 }
             }// End of association edge check.
         }// End of while 
         return false;
     }// End relatedByAssociation
     
      
     // Get and format date/time.
     public static String todayInfo(){
       DateFormat dateFormat = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
       Date date = new Date();
       String today=dateFormat.format(date);
       return today;
     }
     
} // End of class Utilities