/*
 * Class Edge is used in the algorithm for generating
 * sequence of classes in integration testing
 * using Tai-Daniels and other  algorithms. It is used in class Node
 * to contain edges going out of  and in to a node.
 * 
 * Original: July 7, 2011.
 * Latest update: July 21, 2011
 * January 5, 2012 (Deletion status attribute added).
 * Aditya Mathur
 * */


public class Edge{
    
    Node source; // Source node of the edge (this is redundant).
    Node destination; // Destination node of edge.
    String label; // Edge label. 
    String associationLabel="A"; // label for association edge.
    boolean visited; // Used in graph traversal, e.g., Depth First Search.
    EdgeWeight w; // Weight of an edge
    public static String deleteString="*"; // Used to mark an edge as deleted.
     boolean frond; // Used in LeTraon et al. and Briand et al. algorithms
      boolean deleted; // Used in Le Traon algorithm
    
    // This next variable is either "s", "w", or "u" depending on whether the edge 
    // was deleted when using stub minimization or by weight or for some other reason.
    
    private String deletionStatus;
    
    
    // For the Tai-Daniels algorithm an edge label can be one of the following:
    // eA: Association. eC: Collection (Aggregation) and eI  Inheritance. The prefix e indicates
    // that the label is that of an edge. A prefix of n indicates a node.
    
    // An edge is always associated with a node, its source node.
    //  Hence only a destination node is needed to define an edge. Nevertheless,
    // to make things a bit simple, we also add a source node to each edge.
    
    // The weight of an edge e is =s+d, where s is the number of incoming edges to the
    // source node of e and d is the number of edges outgoing from
    // the destination node of e. [Tai-Daniels refer to source and destination
    // as, respectively, head and tail.]
       
    
    public Edge(){ 
        source=null;
        destination=null;
        label="Undefined";   
        w=new EdgeWeight(0, 0, 0);
        deletionStatus="N";  // Default is not deleted.
    }
    
    public Edge(Node s, Node d, String l){ 
        source=s;
        destination=d;
        label=l;   
        visited=false;
        w=new EdgeWeight(0, 0, 0);
        deletionStatus="N";  // Default is not deleted.
    }
    
    public Node getSource(){ // Extract source node.
        return source;
    }
    
    public Node getDestination(){ // Extract destination node.
        return destination;
    }
    
     public EdgeWeight getEdgeWeight(){
        return w;
    }
    public int getWeight(){
        return w.getWeight();
    }
    
     public int getWeightIn(){
        return w.getWeightIn();
    }
     
      public int getWeightOut(){
       return w.getWeightOut();
    }
    
    public void setWeight(EdgeWeight w){
       this.w=w; 
    }
    
     public void setEdgeWeight(EdgeWeight w){
       this.w=w; 
    }

    // This method is used when the label
    // of an edge is changed in the Tai-Danils algorithm
    // to indicate that it has been deleted from the graph.
    // Thus, the edge remains in the graph except that an asterisk is appended to its label.
    // For example, a label "eA" would become "eA*"
    
    public void delete(String cause){
        this.label=label+deleteString;
        deletionStatus=cause;
    }
    
    public void setDestination(Node n){
        destination=n;
    }
    public String getLabel(){ // Extract label.
        return label;
    }
    
    public void markVisited(){
        visited=true;
    }
    
     public void markNotVisited(){
        visited=false;
    }
     
     public boolean getVisited(){
         return visited;
     }
     
     public String getDeletionStatus(){
         return deletionStatus;
     }
     
       public boolean isFrond(){
          return frond;
      }
     
     public boolean isAssociationEdge(){
         String l=label.substring(1,2);
         if(l.equals(associationLabel)){
             return true;
         }else
             return false;
     }// End of isAssociation()
}

