
/*
 * Class Node used in the algorithm for generating
 * sequence of classes in integration testing
 * using Tai-Daniels algorithm. It is used to create nodes in
 * a Graph.
 * 
 * July 7, 2011. July 28, 2011
 * Aditya Mathur
 * */

import java.util.*;
public class Node{
  
  ArrayList<Edge> incomingEdges=new ArrayList<Edge>(); // Contains all incoming edges to this node.
  ArrayList<Edge> outgoingEdges=new ArrayList<Edge>(); // Contains all outgoing edges from this node.
  ArrayList<Node> staticDependence=new ArrayList<Node>(); // Transtive closure of static "depends" relation.
  ArrayList<Node> dynamicDependence=new ArrayList<Node>(); // Transtive closure of dynamic "depends" relation.
  ArrayList<Node> inheritanceChain=new ArrayList<Node>(); // Transtive closure of dynamic "depends" relation.
  ArrayList<Node> stubsUsed=new ArrayList<Node>(); // List of nodes whose stubs are needed to test this node.
  ArrayList<Node> parentClass=new ArrayList<Node>(); // Holds  nodes of which this node is a parent.
  String label; // Node label (prefixed with an n, e.g. nXY, but only XY displayed). 
    
    int levelMajor; // Major level assigned by Tai-Daniels algorithm
    int levelMinor; // Minor level assigned by Tai-Daniels algorithm
    int incomingEdgeCount=0; // Number of incoming edges
    int outgoingEdgeCount=0; // Number of outgoing edges
    
    boolean visited; // Indicates whether or not this node is visited during a graph traversal.
    boolean stub;// True means a stub for this class exists, false: does not exist.
    boolean isAbstract=false; // Tue if this node represents an abstract class, false otherwise.
       int color;// This is used by the Cycle Detection algorithm
       
       // Next variable used in Le Traon et al. algorithm to denote the sequence number
    // in which this node was traversed while finding SCCs during a recursive call.
    int traversalIndex=0; 
    
    // The next index keeps track if the initial Scc to which the node belongs.
    int sccIndex=0; // 
    
    
    boolean tested; // True means this node (i.e. the corresponding class) has been tested.
    int frondWeight;// This is used in LaTraon algorithm to count the number of incoming/outgoing frond edges
    boolean rootNode; // Set to true if this the root of a strongly connected component.
    boolean frondRoot; // Set to true if this node might have an incoming frond edge.
    boolean masterRoot;// True implies that this node is the root of a tree in the forest.
    
    
    
   
   // Attributes for use in Tarjan's Strongly Connected Component finding algorithm
   int index; 
   int lowLink;
   
   public Node(){
       visited=false;
       label="Undefined";   
       levelMajor=0;
       levelMinor=0;
       incomingEdgeCount=0;
       outgoingEdgeCount=0;
       stub=false; // No stub created for this node (also a class).
       index=-1; // // Implies that this index is undefined as in Tarjan's algorithm.
 
   }
   
   public Node(String l, int levMajor, int levMinor){ // Create an empty node.
       visited=false;
       label=l;   
       levelMajor=levMajor;
       levelMinor=levMinor;
       incomingEdgeCount=0;
       outgoingEdgeCount=0;
       stub=false; // NO stub created for this node (also a class).
       index=-1; // Implies that this index is undefined as in Tarjan's algorithm.
   }
   
   public void addOutgoingEdge(Edge e){ // Add an outgoing edge to this node.
           outgoingEdges.add(e);   
           outgoingEdgeCount++;
           Node dest=e.getDestination();
           dest.addIncomingEdge(e); 
   }// End of addEdge()
   
   public void addIncomingEdge(Edge e){ // Add an outgoing edge to this node.
       incomingEdges.add(e);   
       incomingEdgeCount++; 
   }
   
   public void removeOutgoingEdge(Edge e){ // Remove an edge from the node.
       outgoingEdges.remove(e);  
       outgoingEdgeCount--;
       Node dest=e.getDestination();
       staticDependence.remove(dest);
   }
  
    
   public ArrayList<Edge>  getOutgoingEdges(){ // Return all outgoing edges    
       return(outgoingEdges);
   }
   
     // Return all outgoing edges  from this scc to nodes within this
    // scc. 
    public ArrayList<Edge>  getOutgoingEdges(ArrayList<Node> scc){ 
        ArrayList<Edge> returnEdgeList=new ArrayList<Edge>();
        Iterator<Edge> itr=outgoingEdges.iterator();
        while(itr.hasNext()){
            Edge e=itr.next();
            Node n=e.getDestination();
            if(scc.contains(n)){ // Source of this edge must be in scc.
                returnEdgeList.add(e);
            }
        }// End of edge iteration loop.
        return(returnEdgeList); // This might be an empty list if no edge exists that satisfies the condition
    }// End of getIncomingEdges()
    
     public ArrayList<Edge>  getIncomingEdges(){ // Return all incoming edges   
       return(incomingEdges);
   }
   
   
   public void addToStubsUsed(Node n){
       stubsUsed.add(n);
   }
   
   public ArrayList<Node> getStubsUsed(){
       return stubsUsed;
   }
   // Return only edges that have destination node at majorLevel=level
    public ArrayList<Edge>  getOutgoingEdges(int level){ 
        ArrayList<Edge> edgesAtLevel=new ArrayList<Edge> ();
        Iterator<Edge> itr=outgoingEdges.iterator();
        while(itr.hasNext()){
            Edge e=itr.next();
            if(e.getDestination().getLevel(0)==level){
                edgesAtLevel.add(e);
            } 
        }// End of edge loop
       return(edgesAtLevel);
   }//End of getEdges
   
   public String getLabel(){ // Extract node label.
       return label;
   }
   
    public int getLevel(int majorMinor){ // Extract node major/minor level set by the Tai-Daniel's algorithm.
       if(majorMinor==0)
           return levelMajor;
       else
           return levelMinor;
   }// End of getLevel()
    
    public String getLevel(){ // Extract node major/minor level set by the Tai-Daniel's algorithm.
           return Integer.toString(levelMajor)+"."+Integer.toString(levelMinor);
   }// End of getLevel()
    
    public int getOutgoingEdgeCount(){
        if(outgoingEdgeCount!=outgoingEdges.size()){
            System.out.println("**Program error** Outgoing edge counts do not match for node: "+
                               label.substring(1));
        }
        return outgoingEdges.size();
    }
    
    public int getIncomingEdgeCount(){
        if(incomingEdgeCount!=incomingEdges.size()){
            System.out.println("**Program error** Incoming edge counts do not match for node: "+
                               label.substring(1));
        }
        return incomingEdges.size();
    }
    public int getIndex(){ 
        return index;
    }
    public int getLowLink(){
        return lowLink;
    }
    
     public int getTraversalIndex(){
        return traversalIndex;
    }
     
       public int getSccIndex(){
        return sccIndex;
    }
    
    public void setTraversalIndex(int index){
        traversalIndex=index;
    }
    
    public void setSccIndex(int index){
        sccIndex=index;
    }
      public boolean getVisitStatus(){
        return visited;
    }
       
    public void setVisitStatus(boolean s){ // visited true implies visited, false implies not visited.
        visited=s;
    }
    
     
     public void markMasterRoot(){
        masterRoot=true;
    }
    
    public boolean isMasterRoot(){
        return masterRoot;
    }
    
    public int getFrondWeight(){
        return frondWeight;
    }
    
     public boolean getMark(){
      return visited;
    }
    
       public boolean isRoot(){
        return rootNode;
    }
    
    public boolean isFrondRoot(){
        return frondRoot;
    }
    
    public void setOutgoingEdges(ArrayList<Edge> e){
        outgoingEdges=e;
    }
    
       public void setIncomingEdges(ArrayList<Edge> e){
        incomingEdges=e;
    }
     public void setLevel(int l, int majorMinor){ // Set node major level 
       if(majorMinor==0)
           levelMajor=l;
       else
           levelMinor=l;
   }//End of setLevel()
     
     public void setIndex(int index){
         this.index=index;
     }
     
     public void setLowLink(int link){
         lowLink=link;
     }

    public void markVisited(){ // mark=true implies visited, false implies not visited.
        visited=true;
    }
    
    public void markNotVisited(){ // mark=true implies visited, false implies not visited.
        visited=false;
        Iterator<Edge> itr =outgoingEdges.iterator();
        while(itr.hasNext()){
            Edge e=itr.next();
            e.markNotVisited(); 
        }
    }
    
    public void setAbstract(){
        isAbstract=true;
    }
    
    public boolean isNodeAbstract(){ 
        return isAbstract;
    }
    
   
    
    // Find the set of (immediate) successor nodes of a given node.
    // Given that there might be a loop in the graph due to association edges,
    // we omit destination nodes of edges labeled eA (Association).
    
    // ignoreEdges contains labels of edges that are to be ignored when finding 
    // successors of a node. It is used to implement the Tai-Daniels algorithm.
    // Of course, it could empty when all edges in a graph are to be considered.
    
    public ArrayList<Node> getSuccessors(ArrayList<String> ignoreEdges){
        ArrayList <Node> successorNodes=new ArrayList<Node>();
      
        Iterator itr =outgoingEdges.iterator();
        while(itr.hasNext()){
            Edge nextEdge=(Edge)itr.next();
            Node destination=nextEdge.getDestination();
            if(ignoreEdges.indexOf(nextEdge.getLabel())==-1){
                successorNodes.add(destination);
            }
        }
        return successorNodes;
        
    } // End of successors()
    
    
    // The next method is similar to getSuccessor() except that
    // it returns a list of indices (integers) to nodes that
    // are immediate successors of this node. The indices point
    /// directly into the graph.
    
    public ArrayList<Integer> getSuccessorsIndex(){
        ArrayList <Integer> indices=new ArrayList<Integer>();
        
        Iterator itr =outgoingEdges.iterator();
        int index=0;
        while(itr.hasNext()){
            Edge nextEdge=(Edge)itr.next();
            Node destination=nextEdge.getDestination();
            if(!nextEdge.getLabel().equals("eA")){
                indices.add(index);
            }
        }
        return indices;
        
    } // End of successors()
   

    
    // This method adds a a list of statically dependent nodes 
    // to that of this node.
    
    public void setStaticDependence( ArrayList<Node> list){
        staticDependence=list;
        
    }// End of updateStaticDependence()
    
    public ArrayList<Node> getStaticDependence(){
        
        return staticDependence;
        
    }// End getStaticDependence()
    
    public void updateDynamicDependence(Node n){
        dynamicDependence.add(n);
    } // End of updateDynamicDependence()
    
    public ArrayList<Node> getDynamicDependence(){ 
        return dynamicDependence;
    }// End getDynamicDependence()
    
    public void setDynamicDependence(ArrayList<Node> list){ 
        dynamicDependence=list;
    }// End of etDynamicDependence
    
     public ArrayList<Node> getInheritanceChain(){ 
        return inheritanceChain;
    }// End getDynamicDependence()
    
    public void setInheritanceChain(ArrayList<Node> chain){ 
        inheritanceChain=chain;
    }// End of etDynamicDependence
    
    public boolean isStub(){
        return stub;
    }// End of isStub()
    
     public int getColor(){
        return color;
    }
    
    public void setColor(int c){
        color=c;
    }
    
    public void addParent(Node n){
      parentClass.add(n);
    }
    
    public boolean isParent(Node n){
      return parentClass.contains(n);
    }
    
     
    // Check if n is an ancestor of this node.
    public boolean isAncestor(Node n){
      Iterator<Edge> itr= outgoingEdges.iterator();
      while(itr.hasNext()){
        Edge nextEdge=itr.next();
        Node destination=nextEdge.getDestination();
        if(destination.equals(n)){
          return true; // Ancestor found
        }
      }
      return false;
    }// End of isAncestor()
    
    // Check if this node has any edge coming in from its parent.
    public boolean checkInheritance(){
      
      Iterator <Edge> itr=outgoingEdges.iterator();
      while(itr.hasNext()){
        Edge e=itr.next();
        if(e.getLabel().equals(Utilities.InheritanceEdgeLabel)){
          Node dest=e.getDestination();
          if(dest.isAncestor(this)){
            System.out.println("Invalid relations among Node "+label+" and node " +dest.getLabel());
            return false;
          }
        }
      }
      return true;
    }// End of checkInheritance()
} // End of Node class
        