
/*
 * Tarjan's algorithm for finding Strongly Connected Componets (SCC) in a
 * directed graph.
 * 
 * Coded by: Aditya Mathur
 * Algorithm source: http://en.wikipedia.org/wiki/Tarjan%27s_strongly_connected_components_algorithm
 * Original coding: July 13, 2011
 * Latest update: July 20, 2011 [Inadvertently introduced a bug and then removed it!!]
 * 
 * 
 * Each component is an ArrayList<Node>. Thus, the method tarjan() returns
 * an ArrayList<ArrayList<Node>> object that
 * contains one or more SCCs.
 * 
 * tarjan() ignores all edges whose label does not match
 * "eA". 
 */

import java.util.*;
public class SccTarjan{
    
    int index;
    Stack <Node> s=new Stack<Node>(); // For use in Tarjan's algorithm
    Graph g;
    ArrayList<ArrayList<Node>> scc=new ArrayList<ArrayList<Node>>(); // This is where all SCCs in g are saved.
    
    // Constructor.
    public SccTarjan(Graph g){
        this.g=g; // Save graph for future use.
    }
    
    // Find all SCCs. tarjan() is prelude to the algorithm
    // to find SCCs.
    public   ArrayList<ArrayList<Node>> tarjan(){
        index=0;  
        ArrayList<Node> nodes=g.getNodes(); // Get all nodes in g.
        int numNodes=nodes.size(); // Number of nodes in g.
        
        if(numNodes==0){
            Utilities.debug(Utilities.tarjanDebug, "--Graph is empty.");
            return scc; // Empty scc returned.
        }
        
        // Loop over all nodes, one at a time.
        Iterator itr=nodes.iterator();
        int nodeIndex=0; // Index into nodes.
        for(int i=0; i<nodes.size(); i++){  
            Node n=nodes.get(i);
            if(n.getIndex()==-1){ // Process this node if not  visited.
                Utilities.debug(Utilities.tarjanDebug, "\n     Now processing node:"+
                                i+": " +n.getLabel().substring(1));
                
                strongConnect(nodes.get(i)); // Get an scc starting at node n.
                
                Utilities.debug(Utilities.tarjanDebug,"     Node: "+n.getLabel().substring(1)+" processed.");
            }// End of if
        }// End of for over nodes.
        return scc; // scc is updated inside strongConnect() after each component is found.
        
    } // End of tarjan() starter portion of the algorithm.
    

    // Tarjan's strongConnect() method follows.
    
    private void strongConnect(Node v){
       
        v.setIndex(index);
        v.setLowLink(index);
        index++;
        s.push(v);
        ArrayList<Edge> edges=v.getOutgoingEdges();
        int edgeCount=edges.size();
 
      // Consider successors of node v.
        
        Iterator itrEdge=edges.iterator();
        
        while(itrEdge.hasNext()){
            Edge e=(Edge)itrEdge.next();
            if(e.getLabel().equals("eA")){ // Process only the association edges that have not been deleted. 
                Node w=e.getDestination(); // Get the successor in g.
                
                // Make sure the destination node is in the graph.
                // Note: This may be a subgraph of a graph and hence
                // the check is needed because an edge of a subgraph (an SCC)
                // may take us out of the subgraph.
                if(inGraph(w)){
                    if(w.getIndex()==-1){
                        strongConnect(e.getDestination());
                        v.setLowLink(min(v.getLowLink(), w.getLowLink()));
                    } else{
                        if(s.search(w)!=-1)
                            v.setLowLink(min(v.getLowLink(), w.getIndex()));
                    }// End of else part  
                }
            }// End of outer if to ignore an edge.
        }  // End of while to iterate through successors of n.
        
        // If n is the root node then pop the stack and generate an SCC.
        
        if(v.getLowLink()==v.getIndex()){
            ArrayList<Node> anScc=new ArrayList<Node>(); // Initialize an Scc
            Node w;
            do{
                w=s.pop();
                anScc.add(w);
            }while(!v.getLabel().equals(w.getLabel()));
            scc.add(anScc); // Add this SCC to the list of SCCs.
        } // End of if
    }// End of strongConnect()
    
    
    // Check if the input node exists in the graph (or subgraph) g.
    
    private boolean inGraph(Node w){
        ArrayList<Node> nodes =g.getNodes();
        Iterator itr=nodes.iterator();
        while(itr.hasNext()){
            Node n=(Node)itr.next();
            String labelN=n.getLabel();
            String labelW=w.getLabel();
            if(labelW.equals(labelN)){
                return true;
            }     
        }
        return false;
    }  // End of inGraph()

    private int min(int a, int b){
        if(a<b)
            return a;
        else
            return b;
    }// End of min()
    
    // DEbugging method
    
    public void displayIndices(Graph g){
        ArrayList<Node> nodes=g.getNodes();
        int count=nodes.size();
        for(int i=0; i<count; i++){
            System.out.print(" "+nodes.get(i).getLabel().substring(1)+" "+nodes.get(i).getIndex()); 
        }
        System.out.println();
    }
 
} // End of class SccTarjan